
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Interfaces.C;                      use Interfaces.C;
with Interfaces.C.Strings;              use Interfaces.C.Strings;

package body Allegro is

    procedure Allegro_Date( day, month, year : out Positive ) is

        function C_Allegro_Date return Integer;
        pragma Import( C, C_Allegro_Date, "get_allegro_date" );

        dt : constant Integer := C_Allegro_Date;
    begin
        year := dt / 10000;
        month := (dt / 100) mod 100;
        day := dt mod 100;
    end Allegro_Date;

    ----------------------------------------------------------------------------

    function Allegro_Error return String is

        function C_Allegro_Error return C.Strings.chars_ptr;
        pragma Import( C, C_Allegro_Error, "get_allegro_error" );

    begin
         return To_Ada( Value( C_Allegro_Error ) );
    end Allegro_Error;

    ----------------------------------------------------------------------------

    function Allegro_Id return String is

        function C_Allegro_Id return C.Strings.chars_ptr;
        pragma Import( C, C_Allegro_Id, "get_allegro_id" );

    begin
        return To_Ada( Value( C_Allegro_Id ) );
    end Allegro_Id;

    ----------------------------------------------------------------------------

    procedure Allegro_Message( message : String ) is

        procedure C_Allegro_Message( message : C.char_array );
        pragma Import( C, C_Allegro_Message, "allegro_message" );

    begin
        C_Allegro_Message( To_C( message ) );
    end Allegro_Message;

    ----------------------------------------------------------------------------

    procedure Allegro_Version( major, minor, wip : out Integer ) is

        function C_Allegro_Version return Integer;
        pragma Import( C, C_Allegro_Version, "get_allegro_version" );

        function C_Allegro_Sub_Version return Integer;
        pragma Import( C, C_Allegro_Sub_Version, "get_allegro_sub_version" );

        function C_Allegro_Wip_Version return Integer;
        pragma Import( C, C_Allegro_Wip_Version, "get_allegro_wip_version" );

    begin
        major := C_Allegro_Version;
        minor := C_Allegro_Sub_Version;
        wip := C_Allegro_Wip_Version;
    end Allegro_Version;

    ----------------------------------------------------------------------------

    function CPU_Vendor return String is

        function C_CPU_Vendor return C.Strings.chars_ptr;
        pragma Import( C, C_CPU_Vendor, "get_cpu_vendor" );

    begin
        return To_Ada( Value( C_CPU_Vendor ) );
    end CPU_Vendor;

    ----------------------------------------------------------------------------

    procedure Desktop_Resolution( width,
                                  height  : out Natural;
                                  success : out Boolean ) is

        function C_Desktop_Resolution( width, height : access Integer ) return Integer;
        pragma Import( C, C_Desktop_Resolution, "get_desktop_resolution" );

        awidth, aheight : aliased Integer;
    begin
        success := C_Desktop_Resolution( awidth'Access, aheight'Access ) = 0;
        if success then
            width := awidth;
            height := aheight;
        else
            width := 0;
            height := 0;
        end if;
    end Desktop_Resolution;

    ----------------------------------------------------------------------------

    function Is_OS_Multitasking return Boolean is

        function C_Get_OS_Multitasking return Integer;
        pragma Import( C, C_Get_OS_Multitasking, "get_os_multitasking" );

    begin
        return C_Get_OS_Multitasking /= 0;
    end Is_OS_Multitasking;

    ----------------------------------------------------------------------------

    procedure Set_Window_Title( name : String ) is

        procedure C_Set_Window_Title( name : C.char_array );
        pragma Import( C, C_Set_Window_Title, "set_window_title" );

    begin
        C_Set_Window_Title( To_C( name ) );
    end Set_Window_Title;

    ----------------------------------------------------------------------------

    function To_String( capabilities : CPU_Capabilities_Type ) return String is
        result : Unbounded_String;

        ------------------------------------------------------------------------

        procedure Check( feature : CPU_Capabilities_Type; name : String ) is
        begin
            if (capabilities and feature) > 0 then
                if Length( result ) > 0 then
                    result := result & " | ";
                end if;
                result := result & name;
            end if;
        end Check;

        ------------------------------------------------------------------------

    begin
        Check( CPU_ID,       "CPUID" );
        Check( CPU_FPU,      "FPU" );
        Check( CPU_MMX,      "MMX" );
        Check( CPU_MMXPLUS,  "MMX+" );
        Check( CPU_SSE,      "SSE" );
        Check( CPU_SSE2,     "SSE2" );
        Check( CPU_SSE3,     "SSE3" );
        Check( CPU_3DNOW,    "3DNOW" );
        Check( CPU_ENH3DNOW, "ENH_3DNOW" );
        Check( CPU_CMOV,     "CMOV" );
        Check( CPU_AMD64,    "AMD64" );
        Check( CPU_IA64,     "IA64" );
        return To_String( result );
    end To_String;

    ----------------------------------------------------------------------------

    function To_String( family : CPU_Family_Type ) return String is
    begin
        case family is
            when CPU_FAMILY_I386     => return "Intel-compatible 386 class";
            when CPU_FAMILY_I486     => return "Intel-compatible 486 class";
            when CPU_FAMILY_I586     => return "Pentium class";
            when CPU_FAMILY_I686     => return "Pentium Pro/II/III class";
            when CPU_FAMILY_ITANIUM  => return "Itanium class";
            when CPU_FAMILY_EXTENDED => return "Extended (see model)";
            when others              => return "Unknown CPU Family";
        end case;
    end To_String;

    ----------------------------------------------------------------------------

    function To_String( family : CPU_Family_Type;
                        model  : CPU_Model_Type ) return String is
    begin
        case family is
            when CPU_FAMILY_I486 =>
                case model is
                    when CPU_MODEL_I486DX   => return "Intel 486 DX";
                    when CPU_MODEL_I486DX50 => return "Intel 486 DX50";
                    when CPU_MODEL_I486SX   => return "Intel 486 SX";
                    when CPU_MODEL_I487SX   => return "Intel 487 SX";
                    when CPU_MODEL_I486SL   => return "Intel 486 SL";
                    when CPU_MODEL_I486SX2  => return "Intel 486 SX2";
                    when CPU_MODEL_I486DX2  => return "Intel 486 DX2";
                    when CPU_MODEL_I486DX4  => return "Intel 486 DX4";
                    when others             => return To_String( family );
                end case;

            when CPU_FAMILY_I586 =>
                case model is
                    when CPU_MODEL_PENTIUM             => return "Pentium";
                    when CPU_MODEL_PENTIUMP54C         => return "Pentium P54C";
                    when CPU_MODEL_PENTIUMOVERDRIVE    => return "Pentium Overdrive";
                    when CPU_MODEL_PENTIUMOVERDRIVEDX4 => return "Pentium Overdrive DX4";
                    when CPU_MODEL_CYRIX               => return "Cyrix";
                    when CPU_MODEL_K5                  => return "AMD K5";
                    when CPU_MODEL_K6                  => return "AMD K6";
                    when others                        => return To_String( family );
                end case;

            when CPU_FAMILY_I686 =>
                case model is
                    when CPU_MODEL_PENTIUMPROA          => return "Pentium Pro A";
                    when CPU_MODEL_PENTIUMPRO           => return "Pentium Pro";
                    when CPU_MODEL_PENTIUMIIKLAMATH     => return "Pentium II KLA MATH";
                    when CPU_MODEL_PENTIUMII            => return "Pentium II";
                    when CPU_MODEL_CELERON              => return "Celeron";
                    when CPU_MODEL_PENTIUMIIIKATMAI     => return "Pentium III Katmai";
                    when CPU_MODEL_PENTIUMIIICOPPERMINE => return "Pentium III Coppermine";
                    when CPU_MODEL_PENTIUMIIIMOBILE     => return "Pentium III Mobile";
                    when CPU_MODEL_ATHLON               => return "AMD Athlon";
--                    when CPU_MODEL_DURON                => return "AMD Duron";
                    when others                         => return To_String( family );
                end case;

            when CPU_FAMILY_EXTENDED =>
                case model is
                    when CPU_MODEL_PENTIUMIV => return "Pentium 4";
                    when CPU_MODEL_XEON      => return "Xeon";
                    when CPU_MODEL_ATHLON64  => return "Athlon64";
                    when CPU_MODEL_OPTERON   => return "Opteron";
                    when others              => return "Unknown Model";
                end case;

            when others =>
                return To_String( family );
        end case;
    end To_String;

    ----------------------------------------------------------------------------

    overriding
    function To_String( os : OS_Type_Type ) return String is
    begin
        if os = OSTYPE_WIN3 then
            return "Windows 3.1 or earlier";
        elsif os = OSTYPE_WIN95 then
            return "Windows 95";
        elsif os = OSTYPE_WIN98 then
            return "Windows 98";
        elsif os = OSTYPE_WINME then
            return "Windows ME";
        elsif os = OSTYPE_WINNT then
            return "Windows NT";
        elsif os = OSTYPE_WIN2000 then
            return "Windows 2000";
        elsif os = OSTYPE_WINXP then
            return "Windows XP";
        elsif os = OSTYPE_WIN2003 then
            return "Windows 2003";
        elsif os = OSTYPE_WINVISTA then
            return "Windows Vista";
        elsif os = OSTYPE_OS2 then
            return "OS/2";
        elsif os = OSTYPE_WARP then
            return "OS/2 Warp 3";
        elsif os = OSTYPE_DOSEMU then
            return "Linux DOSEMU";
        elsif os = OSTYPE_OPENDOS then
            return "Caldera OpenDOS";
        elsif os = OSTYPE_LINUX then
            return "Linux";
        elsif os = OSTYPE_SUNOS then
            return "SunOS/Solaris";
        elsif os = OSTYPE_FREEBSD then
            return "FreeBSD";
        elsif os = OSTYPE_NETBSD then
            return "NetBSD";
        elsif os = OSTYPE_OPENBSD then
            return "OpenBSD";
        elsif os = OSTYPE_IRIX then
            return "IRIX";
        elsif os = OSTYPE_DARWIN then
            return "Darwin";
        elsif os = OSTYPE_QNX then
            return "QNX";
        elsif os = OSTYPE_UNIX then
            return "Unknown Unix variant";
        elsif os = OSTYPE_BEOS then
            return "BeOS";
        elsif os = OSTYPE_MACOS then
            return "MacOS";
        elsif os = OSTYPE_MACOSX then
            return "MacOS X";
        else
            return "Unknown";
        end if;
    end To_String;

end Allegro;
