
with Ada.Streams;                       use Ada.Streams;
with Objects;                           use Objects;
with Values;                            use Values;

private with Ada.Containers.Indefinite_Ordered_Maps;
private with Ada.Strings.Less_Case_Insensitive;

package Associations is

    -- An Association is an ordered map of strings to Value objects. Names are
    -- case insensitive. Associations do not subclass the Value class to avoid
    -- the possibility of nesting.
    type Association is new Object with private;
    type A_Association is access all Association'Class;

    -- Creates a new empty association.
    function Create_Association return A_Association;
    pragma Postcondition( Create_Association'Result /= null );

    -- Returns the named value in the association as a boolean. An exception is
    -- raised if the value is not found or if it can't be converted to a
    -- boolean.
    function As_Boolean( this : access Association; name : String ) return Boolean;
    pragma Precondition( name'Length > 0 );

    -- Returns the named value in the association as an integer. An exception is
    -- raised if the value is not found or if it can't be converted to an
    -- integer.
    function As_Integer( this : access Association; name : String ) return Integer;
    pragma Precondition( name'Length > 0 );

    -- Returns the named value in the association as a string. An exception is
    -- raised if the value is not found or if it can't be converted to a string.
    function As_String( this : access Association; name : String ) return String;
    pragma Precondition( name'Length > 0 );

    -- Iterate through the name/value pairs in the association. The iteration
    -- will be in alphabetical order of names.
    procedure Iterate( this    : access Association;
                       examine : not null access procedure( name : String; value : A_Value ) );

    -- Returns the number of name/value pairs in the assocation.
    function Length( this : access Association ) return Natural;

    -- Reads an Association from a stream. This should not be called directly.
    function Object_Input( stream : access Root_Stream_Type'Class ) return Association;

    -- Removes the named value from the association.
    procedure Remove( this : access Association; name : String );
    pragma Precondition( name'Length > 0 );

    -- Sets a value in the association by name.
    procedure Set_Value( this : access Association; name : String; val : Boolean );
    pragma Precondition( name'Length > 0 );

    -- Sets a value in the association by name.
    procedure Set_Value( this : access Association; name : String; val : Integer );
    pragma Precondition( name'Length > 0 );

    -- Sets a value in the association by name.
    procedure Set_Value( this : access Association; name : String; val : String );
    pragma Precondition( name'Length > 0 );

    -- Sets a value in the association by name, consuming 'val'.
    procedure Set_Value( this : access Association; name : String; val : in out A_Value );
    pragma Precondition( name'Length > 0 );
    pragma Precondition( val /= null );
    pragma Postcondition( val = null );

    -- Returns a deep copy of the assocation.
    function Copy( src : A_Association ) return A_Association;
    pragma Postcondition( Copy'Result /= src or else src = null );

    -- Deletes the association.
    procedure Delete( this : in out A_Association );
    pragma Postcondition( this = null );

private

    package Value_Maps is new
        Ada.Containers.Indefinite_Ordered_Maps( String, A_Value,
                                                Ada.Strings.Less_Case_Insensitive,
                                                "=" );
    use Value_Maps;

    type Association is new Object with
        record
            pairs : Value_Maps.Map;
        end record;

    procedure Adjust( this : access Association );

    procedure Delete( this : in out Association );

    for Association'Input use Object_Input;

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Association );
    for Association'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Association );
    for Association'Write use Object_Write;

    function A_Association_Input( stream : access Root_Stream_Type'Class ) return A_Association;
    for A_Association'Input use A_Association_Input;

    procedure A_Association_Output( stream : access Root_Stream_Type'Class; obj : A_Association );
    for A_Association'Output use A_Association_Output;

    procedure A_Association_Read( stream : access Root_Stream_Type'Class; obj : out A_Association );
    for A_Association'Read use A_Association_Read;

    procedure A_Association_Write( stream : access Root_Stream_Type'Class; obj : A_Association );
    for A_Association'Write use A_Association_Write;

end Associations;
