
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Entities;                   use Events.Entities;
with GNAT.Regexp;                       use GNAT.Regexp;
with Preferences;                       use Preferences;
with Resources;                         use Resources;
with Support;                           use Support;
with Worlds;                            use Worlds;

package body Entities is

    protected ID_Registry is
        procedure Assign( id : out Entity_Id );
    private
        next : Entity_Id := INVALID_ID + 1;
    end ID_Registry;

    ----------------------------------------------------------------------------

    protected body ID_Registry is

        procedure Assign( id : out Entity_Id ) is
        begin
            id := next;
            next := next + 1;
        end Assign;

    end ID_Registry;

    --==========================================================================

    procedure Activate( this : not null access Entity'Class; activator : not null A_Entity ) is
    begin
        this.On_Activate( activator );
    end Activate;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Entity ) is
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;

    ----------------------------------------------------------------------------

    procedure Collided( this : not null access Entity'Class; e : not null A_Entity ) is
    begin
        pragma Debug( this.touching.Contains( e ),
                      Dbg( "Entity " & To_String( this ) & " collided with " &
                           To_String( e ) & " without separation" ) );
        this.touching.Append( e );
        pragma Debug( Dbg( "Entity " & To_String( this ) &
                           " collided with " & To_String( e ), D_PHYSICS, INFO ) );
        this.On_Collide( e );
    end Collided;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Entity;
                         width,
                         height  : Natural;
                         libName : String ) is
    begin
        Object(this.all).Construct;
        ID_Registry.Assign( this.id );
        this.width := width;
        this.height := height;
        this.attrs := Create_Association;
        this.editingMode := Get_Pref( "application", "editing_mode" );
        if libName'Length > 0 then
            this.lib := Load_Library( libName );
            if this.lib = null then
                raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
            end if;
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Entity ) is
        crsr : Entity_Lists.Cursor := this.touching.First;
    begin
        while Has_Element( crsr ) loop
            Element( crsr ).Separated( this'Unrestricted_Access );
            Next( crsr );
        end loop;
        this.touching.Clear;
        Unload_Library( this.lib );
        Delete( this.attrs );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Face( this : access Entity; dir : Direction_Type ) is
        pragma Unreferenced( this, dir );
    begin
        null;
    end Face;

    ----------------------------------------------------------------------------

    function Get_Attributes( this : not null access Entity'Class ) return A_Association is
    begin
        return this.attrs;
    end Get_Attributes;

    ----------------------------------------------------------------------------

    function Get_Direction( this : not null access Entity'Class ) return Direction_Type is
    begin
        return this.dir;
    end Get_Direction;

    ----------------------------------------------------------------------------

    function Get_Frame( this : not null access Entity'Class ) return Integer is
    begin
        return this.frame;
    end Get_Frame;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access Entity'Class ) return Integer is
    begin
        return this.height;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Entity'Class ) return Entity_Id is
    begin
        return this.id;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Lib_Name( this : not null access Entity'Class ) return String is
    begin
        if this.lib /= null then
            return this.lib.Get_Name;
        else
            return "";
        end if;
    end Get_Lib_Name;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access Entity'Class ) return Integer is
    begin
        return this.width;
    end Get_Width;

    ----------------------------------------------------------------------------

    function Get_X( this : not null access Entity'Class ) return Float is
    begin
        return this.x;
    end Get_X;

    ----------------------------------------------------------------------------

    function Get_XV( this : not null access Entity'Class ) return Float is
    begin
        return this.xv;
    end Get_XV;

    ----------------------------------------------------------------------------

    function Get_Y( this : not null access Entity'Class ) return Float is
    begin
        return this.y;
    end Get_Y;

    ----------------------------------------------------------------------------

    function Get_YV( this : not null access Entity'Class ) return Float is
    begin
        return this.yv;
    end Get_YV;

    ----------------------------------------------------------------------------

    procedure Hit_Wall( this : not null access Entity'Class; dir : Cardinal_Direction ) is
    begin
        this.On_Hit_Wall( dir );
    end Hit_Wall;

    ----------------------------------------------------------------------------

    function Is_Clipped( this : not null access Entity'Class ) return Boolean is
    begin
        return this.clipped;
    end Is_Clipped;

    ----------------------------------------------------------------------------

    function Is_Metaphysical( this : not null access Entity'Class ) return Boolean is
    begin
        return this.metaphysical;
    end Is_Metaphysical;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Permanent( this : access Entity ) return Boolean is
        pragma Unreferenced( this );
    begin
        return False;
    end Is_Permanent;

    ----------------------------------------------------------------------------

    function Is_Physical( this : not null access Entity'Class ) return Boolean is
    begin
        return this.physical;
    end Is_Physical;

    ----------------------------------------------------------------------------

    procedure Iterate_Touching( this    : not null access Entity'Class;
                                examine : not null access procedure( e : A_Entity ) ) is
        procedure Local_Examine( pos : Entity_Lists.Cursor ) is
        begin
            examine( Element( pos ) );
        end Local_Examine;
    begin
        this.touching.Iterate( Local_Examine'Access );
    end Iterate_Touching;

    ----------------------------------------------------------------------------

    procedure Separated( this : not null access Entity'Class; e : not null A_Entity ) is
        crsr : Entity_Lists.Cursor := this.touching.Find( e );
    begin
        if Entity_Lists.Has_Element( crsr ) then
            pragma Debug( Dbg( "Entity " & To_String( this ) &
                               " separated from " & To_String( e ),
                               D_PHYSICS, INFO ) );
            this.touching.Delete( crsr );
            this.On_Separate( e );
        else
            null;
            pragma Debug( Dbg( "Entity " & To_String( this ) &
                               " separated from " & To_String( e ) &
                               " without collision",
                               D_PHYSICS, INFO ) );
        end if;
    end Separated;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Attribute( this : access Entity;
                             name : String;
                             val  : in out A_Value ) is
        valCopy : A_Value;
        lname   : constant String := To_Lower( name );
    begin
        if lname = "physical" then
            if val /= null then
                this.physical := val.As_Boolean;
            end if;
        elsif lname = "clipped" then
            if val /= null then
                this.clipped := val.As_Boolean;
            end if;
        else
            if val = null then
                this.attrs.Remove( name );
            else
                valCopy := Copy( val );
                this.attrs.Set_Value( name, valCopy );         -- consumes valCopy
            end if;
        end if;
        Queue_Entity_Attribute_Changed( this.id, name, val );  -- consumes val
    end Set_Attribute;

    ----------------------------------------------------------------------------

    procedure Set_Frame( this   : not null access Entity'Class;
                         frame  : Natural;
                         notify : Boolean := True ) is
    begin
        if this.frame /= frame then
            this.frame := frame;
            if notify and then frame > 0 then
                Queue_Frame_Changed( this.id, frame );
            end if;
        end if;
    end Set_Frame;

    ----------------------------------------------------------------------------

    procedure Set_Grounded( this : not null access Entity'Class; grounded : Boolean ) is
    begin
        this.grounded := grounded;
        this.Update_Frame;
    end Set_Grounded;

    ----------------------------------------------------------------------------

    procedure Set_Location( this : not null access Entity'Class; x, y : Float ) is
    begin
        this.x := x;
        this.y := y;
    end Set_Location;

    ----------------------------------------------------------------------------

    procedure Set_Size( this   : not null access Entity'Class;
                        width,
                        height : Natural ) is
    begin
        this.width := width;
        this.height := height;
    end Set_Size;

    ----------------------------------------------------------------------------

    procedure Set_Velocity_X( this : not null access Entity'Class; xv : Float ) is
    begin
        this.xv := xv;
    end Set_Velocity_X;

    ----------------------------------------------------------------------------

    procedure Set_Velocity_Y( this : not null access Entity'Class; yv : Float ) is
    begin
        this.yv := yv;
    end Set_Velocity_Y;

    ----------------------------------------------------------------------------

    procedure Set_World( this  : not null access Entity'Class;
                         world : not null access Worlds.World_Object'Class ) is
    begin
        if this.world /= null then
            raise Constraint_Error with "Entity " & To_String( this ) &
                                        " already has a world reference!";
        end if;
        this.world := world;
    end Set_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure Tick( this : access Entity; upTime, dt : Time_Span ) is
        pragma Unreferenced( upTime );
    begin
        this.age := this.age + dt;
    end Tick;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Entity ) return String is
    begin
        return '<' & this.Get_Class_Name & " id=" & Image( Unsigned_32(this.id) ) & '>';
    end To_String;

    ----------------------------------------------------------------------------

    not overriding
    procedure Update_Frame( this : access Entity; notify : Boolean := True ) is
        pragma Unreferenced( this, notify );
    begin
        null;
    end Update_Frame;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Entity ) is
        attrs : A_Association;
    begin
        Object'Read( stream, Object(obj) );
        obj.age := Time_Span'Input( stream );
        obj.width := Integer'Input( stream );
        obj.height := Integer'Input( stream );
        obj.x := Float'Input( stream );
        obj.y := Float'Input( stream );
        obj.xv := Float'Input( stream );
        obj.yv := Float'Input( stream );
        obj.dir := Direction_Type'Input( stream );
        obj.frame := Natural'Input( stream );
        attrs := A_Association'Input( stream );
        if attrs /= null then
            Delete( obj.attrs );
            obj.attrs := attrs;
        end if;
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Entity ) is
    begin
        Object'Write( stream, Object(obj) );
        Time_Span'Output( stream, obj.age );
        Integer'Output( stream, obj.width );
        Integer'Output( stream, obj.height );
        Float'Output( stream, obj.x );
        Float'Output( stream, obj.y );
        Float'Output( stream, obj.xv );
        Float'Output( stream, obj.yv );
        Direction_Type'Output( stream, obj.dir );
        Natural'Output( stream, obj.frame );
        A_Association'Output( stream, obj.attrs );
    end Object_Write;

    --==========================================================================

    function A_Entity_Input( stream : access Root_Stream_Type'Class ) return A_Entity is
        obj : A_Entity := null;
    begin
        if Boolean'Input( stream ) then
            A_Entity'Read( stream, obj );
        end if;
        return obj;
    end A_Entity_Input;

    ----------------------------------------------------------------------------

    procedure A_Entity_Output( stream : access Root_Stream_Type'Class; obj : A_Entity ) is
    begin
        Boolean'Output( stream, obj /= null );
        if obj /= null then
            A_Entity'Write( stream, obj );
        end if;
    end A_Entity_Output;

    ----------------------------------------------------------------------------

    procedure A_Entity_Read( stream : access Root_Stream_Type'Class; obj : out A_Entity ) is
    begin
        obj := new Entity'Class'(Entity'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown entity class " &
                      "<" & To_Lower( msg((Index( msg, ".", Backward ) + 1)..msg'Last) ) & ">";
            end;
    end A_Entity_Read;

    ----------------------------------------------------------------------------

    procedure A_Entity_Write( stream : access Root_Stream_Type'Class; obj : A_Entity ) is
    begin
        Entity'Class'Output( stream, obj.all );
    end A_Entity_Write;

    ----------------------------------------------------------------------------

    function Allocate( id : String ) return A_Entity is
        e : A_Entity;
    begin
        e := Factory.Allocate( id );
        pragma Debug( e = null,
                      Dbg( "Attempted to create unrecognized entity " &
                           "class <" & id & ">", D_ENTITIES, ERROR ) );
        return e;
    exception
        when e : RESOURCE_ERROR =>
            pragma Debug( Dbg( "Error creating entity class <" & id & ">: " &
                               Exception_Message( e ), D_ENTITIES, ERROR ) );
            return null;
    end Allocate;

    ----------------------------------------------------------------------------

    function Copy( src : A_Entity ) return A_Entity is
    begin
        return A_Entity(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Entity ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    procedure Iterate_Classes( pattern : String := "";
                               examine : access procedure( id : String ) ) is
        exp : Regexp;

        procedure Filtered_Examine( id : String ) is
        begin
            if Match( id, exp ) then
                examine( id );
            end if;
        end Filtered_Examine;

    begin
        if pattern'Length > 0 then
            exp := Compile( pattern, True, False );
            Factory.Iterate_Classes( Filtered_Examine'Access );
        else
            Factory.Iterate_Classes( examine );
        end if;
    end Iterate_Classes;

    ----------------------------------------------------------------------------

    function Template( id : String ) return A_Entity renames Factory.Template;

    ----------------------------------------------------------------------------

    procedure Initialize is
    begin
        Factory.Initialize( "entity" );    -- may raise an exception
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize renames Factory.Finalize;

end Entities;
