
with Ada.Strings.Hash_Case_Insensitive;
with Events.Listeners;                  use Events.Listeners;
with Processes;                         use Processes;

private with Ada.Containers.Doubly_Linked_Lists;
private with Ada.Containers.Indefinite_Hashed_Maps;
private with Mutable_Lists;

package Events.Corrals is

    type Corral is new Limited_Object and Process with private;
    type A_Corral is access all Corral'Class;

    -- Creates a new corral to receive events from the global event manager.
    function Create_Corral( name : String ) return A_Corral;
    pragma Precondition( name'Length > 0 );
    pragma Postcondition( Create_Corral'Result /= null );

    -- Registers an object as a listener for a specific type of event. Listeners
    -- will receive all events sent to this corral unless the event has been
    -- consumed by an earlier registered handler that wants to prevent an event
    -- from propagating. Event listeners are notified in the order in which they
    -- register with the corral.
    procedure Add_Listener( this     : access Corral;
                            listener : not null A_Event_Listener;
                            evtName  : String );
    pragma Precondition( evtName'Length > 0 );

    -- Returns the name of the corral.
    function Get_Name( this : access Corral ) return String;
    pragma Postcondition( Get_Name'Result'Length > 0 );

    -- Queue an event to be dispatched to all the registered listener objects.
    -- The event will be consumed.
    procedure Queue_Event( this : access Corral; evt : in out A_Event );
    pragma Precondition( evt /= null );
    pragma Postcondition( evt = null );

    -- Unregisters an object as a listener for a specific type of event. If the
    -- listener was not previously registered, this has no effect.
    procedure Remove_Listener( this     : access Corral;
                               listener : not null A_Event_Listener;
                               evtName  : String );
    pragma Precondition( evtName'Length > 0 );

    -- Dispatches an event immediately, separate from the queue. The event will
    -- be consumed.
    procedure Trigger_Event( this     : access Corral;
                             evt      : in out A_Event;
                             response : out Response_Type );
    pragma Precondition( evt /= null );
    pragma Postcondition( evt = null );

    -- Deletes the corral.
    procedure Delete( this : in out A_Corral );
    pragma Postcondition( this = null );

private

    -- A broker acts as a queueable go-between object that brokers the exchange
    -- between a calling thread and an event dispatching thread. The calling
    -- thread puts an event into the broker, inserts it into the corral's event
    -- thread, and waits for the corral's dispatching thread to get around to
    -- dispatching the event and returning a response via the broker.
    protected type Broker_Type is

        -- Consumes the event. However, if an event has already been set, the
        -- first event will be preserved and the second event pass will be
        -- ignored and deleted.
        procedure Put_Event( evt : in out A_Event );
        pragma Postcondition( evt = null );

        -- This can only be called once because the internal pointer is cleared
        -- when the event is returned. The returned event belongs to the caller.
        procedure Take_Event( evt : in out A_Event );
        pragma Precondition( evt = null );

        -- Sets the response to be returned by Get_Response and allows any
        -- callers waiting on Get_Response to proceed.
        procedure Set_Response( response : Response_Type );

        -- Returns the event's response. This will block until Set_Response has
        -- been called.
        entry Get_Response( response : out Response_Type );

    private
        my_event    : A_Event := null;
        my_response : Response_Type := No_Response;
        initialized : Boolean := False;
        complete    : Boolean := False;
    end Broker_Type;
    type A_Broker is access all Broker_Type;

    procedure Create_Broker( evt : in out A_Event; broker : out A_Broker );
    pragma Precondition( evt /= null );
    pragma Postcondition( evt = null );
    pragma Postcondition( broker /= null );

    procedure Delete( broker : in out A_Broker );
    pragma Postcondition( broker = null );

    ----------------------------------------------------------------------------

    package Broker_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Broker, "=" );
    use Broker_Lists;

    package Event_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Event, "=" );
    use Event_Lists;

    protected type Event_Queue is

        procedure Add( broker : A_Broker );

        procedure Add( evt : in out A_Event );
        pragma Precondition( evt /= null );
        pragma Postcondition( evt = null );

        procedure Mark_Frame;

        -- Returns null when queue is being shut down.
        procedure Remove( broker : out A_Broker );

        -- Returns null when queue is being shut down.
        procedure Remove( evt : out A_Event );

        procedure Shutdown;

    private
        queue     : Event_Lists.List;
        brokers   : Broker_Lists.List;
        remaining : Integer := 0;       -- queued events remaining in the frame
        stopped   : Boolean := False;
    end Event_Queue;

    ----------------------------------------------------------------------------

    package Listener_Lists is new Mutable_Lists( A_Event_Listener, "=" );
    use Listener_Lists;

    type Listener_List is tagged
        record
            corral    : A_Corral := null;
            name      : Unbounded_String;
            listeners : access Listener_Lists.List := new Listener_Lists.List;
        end record;
    type A_Listener_List is access all Listener_List;

    -- Adds the listener to the list and registers the list's corral with
    -- the event manager if the list was empty.
    procedure Add( this     : access Listener_List;
                   listener : not null A_Event_Listener );

    -- Synchronously notifies each listener in the list until the event is
    -- consumed. The response is returned from the last event listener to
    -- be notified, or contains error information if an exception was
    -- raised by the last listener to be notified.
    procedure Call( this     : access Listener_List;
                    evt      : in out A_Event;
                    response : out Response_Type );

    -- Clears the listener list and unregisters the list's corral with the
    -- event manager.
    procedure Clear( this : access Listener_List );

    -- Call this with the listener list's owning corral and the event name that
    -- the list is associated with at time of object creation.
    procedure Construct( this    : access Listener_List;
                         owner   : not null A_Corral;
                         evtName : String );

    -- Removes the listener from the list and unregisters the list's corral
    -- with the event manager if the list is now empty.
    procedure Remove( this     : access Listener_List;
                      listener : not null A_Event_Listener );

    ----------------------------------------------------------------------------

    package Type_Map is new
        Ada.Containers.Indefinite_Hashed_Maps( String,
                                               A_Listener_List,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               "=", "=" );
    use Type_Map;

    protected type Listener_Registry is

        -- Clears all event listener lists in the registry.
        procedure Clear;

        -- Returns a reference to the list of event listeners for the given
        -- event name. If no listeners have been registered for the name yet, an
        -- empty list is returned. The reference to the listener list that is
        -- returned is owned by the listener registry. Do not delete it.
        procedure Get_List( evtName : String; list : out A_Listener_List );

        -- Call this with the listener registry's owning corral at time of
        -- object creation.
        procedure Init( owner : not null A_Corral );

    private
        corral   : A_Corral := null;
        registry : Type_Map.Map;
    end Listener_Registry;

    ----------------------------------------------------------------------------

    type Corral is new Limited_Object and Process with
        record
            name      : Unbounded_String;
            listeners : access Listener_Registry := null;
            queue     : access Event_Queue := null;
        end record;

    procedure Construct( this : access Corral; name : String );

    procedure Delete( this : in out Corral );

    function Get_Process_Name( this : access Corral ) return String;

    procedure Tick( this : access Corral; upTime, dt : Time_Span );

    function To_String( this : access Corral ) return String;

end Events.Corrals;
