
with Ada.Containers;                    use Ada.Containers;
with Applications;                      use Applications;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Application;                use Events.Application;
with Events.Game;                       use Events.Game;
with Events.Input;                      use Events.Input;
with Game_Views.Window_Hooks;           use Game_Views.Window_Hooks;
with Preferences;                       use Preferences;
with Support;                           use Support;

package body Game_Views is

    allocate : Allocator := null;

    ----------------------------------------------------------------------------

    function Create_Game_View( xres, yres : Positive;
                               scale      : Positive ) return A_Game_View is
    begin
        if allocate /= null then
            return allocate.all( xres, yres, scale );
        else
            raise Constraint_Error with "Game_View allocator not registered";
        end if;
    end Create_Game_View;

    ----------------------------------------------------------------------------

    procedure Attach( this : access Game_View; process : not null A_Process ) is
    begin
        this.pman.Attach( process );
    end Attach;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Game_View ) is
    begin
        Limited_Object(this.all).Construct;
        this.corral := Create_Corral( "game view" );
        this.pman := Create_Process_Manager( "view", Constrain( Get_Pref( "application", "max_render_fps" ), 0, 120 ) );
        this.audioPlayer := Create_Audio_Player( this.corral );
        this.inhandler := Create_Input_Handler( Constrain( Get_Pref( "application", "max_input_fps" ), 1, 120 ) );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game_View ) is
    begin
        Delete( this.renderer );
        Delete( this.win );
        Delete( this.inhandler );
        Delete( this.audioPlayer );
        Delete( this.pman );
        Delete( this.corral );
        pragma Debug( this.widgets.Length > 0,
                      Dbg( "Game_Views.Delete Warning: " &
                           Image( Integer(this.widgets.Length) ) &
                           " widgets were abandoned",
                           D_VIEW, WARNING ) );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Detach( this : access Game_View; process : not null A_Process ) is
    begin
        this.pman.Detach( process );
    end Detach;

    ----------------------------------------------------------------------------

    function Get_Corral( this : not null access Game_View'Class ) return A_Corral is
    begin
        return this.corral;
    end Get_Corral;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Game_View ) return String is
        pragma Unreferenced( this );
    begin
        return "Game View";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Widget( this : access Game_View; id : String ) return A_Widget is
        pos : Widget_Registry.Cursor;
    begin
        pos := this.widgets.Find( id );
        if Has_Element( pos ) then
            return Element( pos );
        else
            pragma Debug( Dbg( "Warning: Widget id '" & id & "' not found", D_GUI, WARNING ) );
            raise ID_NOT_FOUND with "Widget not found, id: " & id;
        end if;
    end Get_Widget;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Window( this : access Game_View ) return A_Window is
    begin
        return this.win;
    end Get_Window;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Close_Request( this : access Game_View ) is
        pragma Unreferenced( this );
    begin
        Queue_Close_Window;
    end Handle_Close_Request;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
        id      : constant Event_Id := evt.Get_Id;
        handled : Boolean := False;
    begin
        if id = KEY_HELD_ID then
            handled := this.win.Dispatch_Key_Held( A_Key_Event(evt) );

        elsif id = KEY_PRESS_ID then
            handled := this.win.Dispatch_Key_Press( A_Key_Event(evt) );

        elsif id = KEY_RELEASE_ID then
            handled := this.win.Dispatch_Key_Release( A_Key_Event(evt) );

        elsif id = MOUSE_HELD_ID then
            this.win.Dispatch_Mouse_Held( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_MOVE_ID then
            this.win.Dispatch_Mouse_Move( A_Mouse_Event(evt) );
            handled := True;

        elsif id = MOUSE_PRESS_ID then
            this.win.Dispatch_Mouse_Press( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_RELEASE_ID then
            this.win.Dispatch_Mouse_Release( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_CLICK_ID then
            this.win.Dispatch_Click( A_Mouse_Button_Event(evt) );
            handled := True;

        elsif id = MOUSE_SCROLL_ID then
            handled := this.win.Dispatch_Mouse_Scroll( A_Mouse_Scroll_Event(evt) );

        elsif id = GAME_PAUSED_ID then
            if this.paused /= A_Game_Paused_Event(evt).Is_Paused then
                this.paused := A_Game_Paused_Event(evt).Is_Paused;
                A_Game_View(this).Handle_Paused( this.paused );
            end if;

        elsif id = LOADING_ID then
            A_Game_View(this).Handle_Loading( A_Loading_Event(evt).Is_Loading );

        elsif id = CLOSE_REQUEST_ID then
            A_Game_View(this).Handle_Close_Request;

        elsif id = CLOSE_WINDOW_ID then
            Get_Application.Stop;
        end if;

        if handled then
            Delete( evt );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Loading( this : access Game_View; loading : Boolean ) is
        pragma Unreferenced( this, loading );
    begin
        null;
    end Handle_Loading;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Paused( this : access Game_View; paused : Boolean ) is
        pragma Unreferenced( this, paused );
    begin
        null;
    end Handle_Paused;

    ----------------------------------------------------------------------------

    not overriding
    procedure Register( this : access Game_View; widget : not null A_Widget ) is
        id       : constant String := widget.Get_Id;
        pos      : Widget_Registry.Cursor;
        inserted : Boolean;
    begin
        this.widgets.Insert( widget.Get_Id, widget, pos, inserted );
        if not inserted then
            raise DUPLICATE_ID with "Duplicate widget id: " & id ;
        end if;
    end Register;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Window( this : access Game_View; window : in out A_Window ) is
    begin
        if not this.started then
            if this.win /= null then
                Delete( this.renderer );
                Delete( this.win );
            end if;
            if window /= null then
                this.renderer := Create_Renderer( window,
                                                  Constrain( Get_Pref( "application",
                                                                       "max_render_fps" ),
                                                             1, 60 ) );
            end if;
            this.win := window;
            window := null;
        else
            Delete( window );
            pragma Debug( Dbg( "Error: Game view Window can't be changed while running!",
                               D_GUI, ERROR ) );
        end if;
    end Set_Window;

    ----------------------------------------------------------------------------

    not overriding
    procedure Start( this : access Game_View ) is
    begin
        if not this.started and then not this.stopped then
            pragma Debug( Dbg( "Starting game view", D_VIEW, INFO ) );
            this.corral.Add_Listener( A_Event_Listener(this), "Close_Request" );
            this.corral.Add_Listener( A_Event_Listener(this), "Close_Window" );
            this.corral.Add_Listener( A_Event_Listener(this), "Game_Paused" );
            this.corral.Add_Listener( A_Event_Listener(this), "Key_Held" );
            this.corral.Add_Listener( A_Event_Listener(this), "Key_Press" );
            this.corral.Add_Listener( A_Event_Listener(this), "Key_Release" );
            this.corral.Add_Listener( A_Event_Listener(this), "Loading" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Held" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Move" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Press" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Release" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Click" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Doubleclick" );
            this.corral.Add_Listener( A_Event_Listener(this), "Mouse_Scroll" );
            this.pman.Attach( A_Process(this) );
            this.pman.Attach( A_Process(this.corral) );
            this.pman.Attach( A_Process(this.renderer) );
            this.pman.Attach( A_Process(this.audioPlayer) );
            this.audioPlayer.Start;
            this.inhandler.Start;
            this.pman.Start;
            this.started := True;
        end if;
    end Start;

    ----------------------------------------------------------------------------

    not overriding
    procedure Stop( this : access Game_View ) is
    begin
        if this.started and then not this.stopped then
            pragma Debug( Dbg( "Stopping game view", D_VIEW, INFO ) );
            this.pman.Stop;
            this.inhandler.Stop;
            this.audioPlayer.Stop;
            this.pman.Detach( A_Process(this.audioPlayer) );
            this.pman.Detach( A_Process(this.renderer) );
            this.pman.Detach( A_Process(this.corral) );
            this.pman.Detach( A_Process(this) );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Scroll" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Doubleclick" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Click" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Release" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Press" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Move" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Mouse_Held" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Loading" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Key_Release" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Key_Press" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Key_Held" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Game_Paused" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Close_Window" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Close_Request" );
            this.stopped := True;
        end if;
    end Stop;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Game_View; upTime, dt : Time_Span ) is
        pragma Unreferenced( this, upTime, dt );
    begin
        if Window_Hooks.Close_Window_Clicked then
            Queue_Close_Request;
        end if;
    end Tick;

    ----------------------------------------------------------------------------

    not overriding
    procedure Unregister( this : access Game_View; id : String ) is
        pos : Widget_Registry.Cursor := this.widgets.Find( id );
    begin
        if Has_Element( pos ) then
            this.widgets.Delete( pos );
        end if;
    end Unregister;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game_View ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Game_Views.allocate := allocate;
    end Register_Allocator;

end Game_Views;
