
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Tags;                          use Ada.Tags;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Game;                       use Events.Game;
with Events.World;                      use Events.World;
with Preferences;                       use Preferences;
with Statuses;                          use Statuses;
with Support;                           use Support;
with Values;                            use Values;

package body Games is

    use Actor_Lists;
    use View_Lists;

    allocate : Allocator := null;

    ----------------------------------------------------------------------------

    function Create_Game return A_Game is
    begin
        if allocate /= null then
            return allocate.all;
        else
            raise Constraint_Error with "Game allocator not registered";
        end if;
    end Create_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Event_Listeners( this : access Game ) is
    begin
        this.corral.Add_Listener( A_Event_Listener(this), "End_Game" );
        this.corral.Add_Listener( A_Event_Listener(this), "Load_World" );
        this.corral.Add_Listener( A_Event_Listener(this), "New_Game" );
        this.corral.Add_Listener( A_Event_Listener(this), "Pause_Game" );
        this.corral.Add_Listener( A_Event_Listener(this), "View_Ready" );
    end Add_Event_Listeners;

    ----------------------------------------------------------------------------

    procedure Add_View( this : not null access Game'Class;
                        view : in out A_Game_View ) is
    begin
        this.views.Append( view );
        view := null;
    end Add_View;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Game ) is
    begin
        Limited_Object(this.all).Construct;
        this.corral := Create_Corral( "game logic" );
        this.physics := Create_Physics( this.corral );
        this.pman := Create_Process_Manager( "game",
                                             Constrain( Get_Pref( "application",
                                                                  "max_logic_fps" ),
                                                        0, 120 ), 15 );
        this.sessionVars := Create_Association;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game ) is
    begin
        -- Delete actors
        declare
            crsr  : Actor_Lists.Cursor := this.actors.First;
            actor : A_Actor;
        begin
            while Has_Element( crsr ) loop
                actor := Element( crsr );
                this.actors.Delete( crsr );
                Delete( actor );
                crsr := this.actors.First;
            end loop;
        end;

        -- Delete world
        Delete( this.world );

        -- Delete game views
        declare
            crsr : View_Lists.Cursor := this.views.First;
            view : A_Game_View;
        begin
            while Has_Element( crsr ) loop
                view := Element( crsr );
                this.views.Delete( crsr );
                Delete( view );
                crsr := this.views.First;
            end loop;
        end;

        -- Delete other owned objects
        Delete( this.sessionVars );
        Delete( this.pman );
        Delete( this.physics );
        Delete( this.corral );

        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure End_Game( this : access Game ) is
    begin
        if this.playing then
            pragma Debug( Dbg( "Game: Game session ended", D_GAME, INFO ) );
            this.playing := False;

            this.pman.Detach( A_Process(this.physics) );
            this.pman.Detach( A_Process(this) );
            this.Unload_World;

            Queue_Game_State( playing => False );
        else
            null;
            pragma Debug( Dbg( "Warning: Can't end game; no game in progress.",
                               D_GAME, WARNING ) );
        end if;
    end End_Game;

    ----------------------------------------------------------------------------

    procedure Game_Var_Add( this : not null access Game'Class;
                            var  : String;
                            val  : Integer ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, this.sessionVars.As_Integer( var ) + val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Game_Var_Add;

    ----------------------------------------------------------------------------

    function Get_Corral( this : not null access Game'Class ) return A_Corral is
    begin
        return this.corral;
    end Get_Corral;

    ----------------------------------------------------------------------------

    function Get_Game_Var( this : not null access Game'Class;
                           var  : String ) return Integer is
    begin
        return this.sessionVars.As_Integer( var );
    end Get_Game_Var;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Game ) return String is
        pragma Unreferenced( this );
    begin
        return "Game Logic";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is

        ------------------------------------------------------------------------

        procedure Do_Load_World( evt : not null A_Load_World_Event ) is
        begin
            -- Loads the world from disk and attaches it to the game framework.
            -- This involves starting the world as a process and sending events
            -- to notify the system that a new world has been loaded.
            this.Load_World( evt.Get_Filename );

            -- Detach the game and physics processes until the view has notified
            -- that it is ready. These will be reattached by receiving the
            -- View_Ready event.
            this.pman.Detach( A_Process(this.physics) );
            this.pman.Detach( A_Process(this) );

            Set_Status( resp, ST_SUCCESS );
        exception
            when e : others =>
                -- Loading the world from disk failed. The current state of the
                -- game world remains unchanged.
                Set_Status( resp, ST_FAILED );
                Set_Message( resp, Exception_Message( e ) );
        end Do_Load_World;

        ------------------------------------------------------------------------

        procedure Do_Pause( evt : not null A_Pause_Game_Event ) is
        begin
            if evt.Is_Paused = not this.paused then
                A_Game(this).Pause( not this.paused );
            end if;
        end Do_Pause;

        ------------------------------------------------------------------------

        procedure Do_View_Ready( evt : not null A_View_Ready_Event ) is
        begin
            if evt.Is_Ready then
                this.pman.Attach( A_Process(this) );
                this.pman.Attach( A_Process(this.physics) );
            end if;
        end Do_View_Ready;

        ------------------------------------------------------------------------

    begin
        if evt.Get_Id = END_GAME_ID then
            A_Game(this).End_Game;
        elsif evt.Get_Id = LOAD_WORLD_ID then
            Do_Load_World( A_Load_World_Event(evt) );
        elsif evt.Get_Id = NEW_GAME_ID then
            A_Game(this).New_Game;
        elsif evt.Get_Id = PAUSE_GAME_ID then
            Do_Pause( A_Pause_Game_Event(evt) );
        elsif evt.Get_Id = VIEW_READY_ID then
            Do_View_Ready( A_View_Ready_Event(evt) );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    procedure Load_World( this : not null access Game'Class; name : String ) is
        world : A_World;
    begin
        if this.views.Is_Empty then
            pragma Debug( Dbg( "Game.Load_World Error: Game has no registered views",
                               D_GAME, ERROR ) );
            return;
        end if;

        -- Load_World will raise an exception on error and then we won't unload
        -- the current world.
        pragma Debug( Dbg( "Game.Load_World: Loading <" & name & ">", D_GAME, INFO ) );
        world := Load_World( name );
        this.Load_World( world );
    end Load_World;

    ----------------------------------------------------------------------------

    procedure Load_World( this  : not null access Game'Class;
                          world : in out A_World ) is

        ------------------------------------------------------------------------

        procedure Connect_Actors_To_Entities is

            -- Checks for entities to connect to an existing actor (ie: players)
            procedure Connect_Actors( pos : Actor_Lists.Cursor ) is
                actor : constant A_Actor := Element( pos );

                procedure Examine_Entity( e : not null A_Entity ) is
                begin
                    if actor.Set_Entity( e ) then
--Dbg( "Session actor " & actor.To_String & " controlling entity " & e.To_String );
                        this.actorIndex.Insert( e.Get_Id, actor );
                    end if;
                end Examine_Entity;
            begin
                this.world.Examine_Entities( Examine_Entity'Unrestricted_Access );
            end Connect_Actors;

            --------------------------------------------------------------------

            -- Creates and connects an actor to control an entity, if possible.
            procedure Create_Actors( e : not null A_Entity ) is
                actor : A_Actor;
            begin
                actor := Actors.Create_Actor( External_Tag( e.all'Tag ) );
                if actor /= null then
                    -- we just created this actor with the entities class
                    -- so we can assume that it can control the entity.
                    actor.Set_Entity( e );
--Dbg( "Temporal actor " & actor.To_String & " controlling entity " & e.To_String );
                    actor.Set_Game( A_Game(this) );
                    this.actorIndex.Insert( e.Get_Id, actor );
                    this.actors.Append( actor );
                end if;
            end Create_Actors;

       begin
            -- Examine each existing (non-temporal) Actor and associate it
            -- with an entity in the world as necessary.
            this.actors.Iterate( Connect_Actors'Unrestricted_Access );

            -- Examine each entity in the world and create a temporal
            -- controlling Actor as necessary.
            this.world.Examine_Entities( Create_Actors'Unrestricted_Access );
        end Connect_Actors_To_Entities;

        ------------------------------------------------------------------------

    begin
        this.Unload_World;

        this.world := world;
        world := null;
        if this.world /= null then
            this.world.Attach_To_Framework( this.corral, this.pman );
            Connect_Actors_To_Entities;
        end if;
    end Load_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure New_Game( this : access Game ) is
    begin
        if not this.playing then
            pragma Debug( Dbg( "Game.New_Game: Game session begun", D_GAME, INFO ) );
            Queue_Game_State( playing => True );
            this.paused := False;
            this.playing := True;
        else
            null;
            pragma Debug( Dbg( "Game.New_Game: Can't begin game; " &
                               "game already in progress", D_GAME, ERROR ) );
        end if;
    end New_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pause( this : access Game; enabled : Boolean ) is
    begin
        if this.playing and then enabled /= this.paused then
            this.paused := enabled;

            -- Change the running state of all the game processes
            this.pman.Pause( A_Process(this), this.paused );
            this.pman.Pause( A_Process(this.physics), this.paused );
            this.pman.Pause( A_Process(this.world), this.paused );

            -- Queue an event notifying the system of the new paused state
            Queue_Game_Paused( this.paused );
        end if;
    end Pause;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Event_Listeners( this : access Game ) is
    begin
        this.corral.Remove_Listener( A_Event_Listener(this), "View_Ready" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Pause_Game" );
        this.corral.Remove_Listener( A_Event_Listener(this), "New_Game" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Load_World" );
        this.corral.Remove_Listener( A_Event_Listener(this), "End_Game" );
    end Remove_Event_Listeners;

    ----------------------------------------------------------------------------

    procedure Set_Game_Var( this : not null access Game'Class;
                            var  : String;
                            val  : Integer ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    procedure Set_Game_Var( this : not null access Game'Class;
                            var  : String;
                            val  : Boolean ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    not overriding
    procedure Start( this : access Game ) is

        procedure Examine( pos : View_Lists.Cursor ) is
        begin
            Element( pos ).Start;
        end Examine;

    begin
        if not this.started and then not this.stopped then
            this.started := True;
            A_Game(this).Add_Event_Listeners;
            this.pman.Attach( A_Process(this.corral) );

            this.views.Iterate( Examine'Access );
            this.pman.Start;
        end if;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Game'Class ) is

        procedure Stop_View( pos : View_Lists.Cursor ) is
        begin
            Element( pos ).Stop;
        end Stop_View;

        procedure Delete_Actor( pos : Actor_Lists.Cursor ) is
            actor : A_Actor := Element( pos );
        begin
            Delete( actor );
        end Delete_Actor;

    begin
        if not this.stopped then
            this.stopped := True;
            this.pman.Stop;
            this.views.Iterate( Stop_View'Access );

            this.Unload_World;
            this.actors.Iterate( Delete_Actor'Access );
            this.actors.Clear;
            this.pman.Detach( A_Process(this.physics) );
            this.pman.Detach( A_Process(this) );
            this.pman.Detach( A_Process(this.corral) );
            A_Game(this).Remove_Event_Listeners;
        end if;
    end Stop;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Game; upTime, dt : Time_Span ) is
        pragma Unreferenced( this, upTime, dt );
    begin
        null;
    end Tick;

    ----------------------------------------------------------------------------

    procedure Unload_World( this : not null access Game'Class ) is
    begin
        if this.world /= null then
            this.world.Detach_From_Framework;

            -- delete temporal actors controlling entities in the world
            declare
                crsr  : Actor_Lists.Cursor := this.actors.First;
                nxt   : Actor_Lists.Cursor;
                actor : A_Actor;
            begin
               this.actorIndex.Clear;
                while Has_Element( crsr ) loop
                    nxt := Next( crsr );
                    actor := Element( crsr );
                    if actor.Is_Temporal then
                        this.actors.Delete( crsr );
                        Delete( actor );
                    else
                        actor.Set_Entity( null );
                    end if;
                    crsr := nxt;
                end loop;
            end;

            -- Do not delete this.world here because it is still owned by the
            -- process manager. The world will be destroyed by a destructor
            -- that it gave to the process manager when it registered itself as
            -- a process.
            this.world := null;
        end if;
    end Unload_World;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Games.allocate := allocate;
    end Register_Allocator;

end Games;
