
with Ada.Unchecked_Deallocation;
with Allegro.Bitmaps.Direct_Access;     use Allegro.Bitmaps.Direct_Access;
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Truecolor;                 use Allegro.Truecolor;
--with Debugging;                         use Debugging;
with Interfaces;                        use Interfaces;

package body Image_Filters is

    type Integer_Array is array (Integer range <>) of Integer;
    type A_Integer_Array is access all Integer_Array;

    procedure Free is new Ada.Unchecked_Deallocation( Integer_Array, A_Integer_Array );

    ----------------------------------------------------------------------------

    -- Originally by Sik; adapted by Oog; adapted by Kevin Wellwood.
    procedure Fast_Blur( bmp    : not null A_Bitmap;
                         x1, y1 : Natural;
                         x2, y2 : Natural;
                         passes : Positive := 1 ) is

        ------------------------------------------------------------------------

        procedure Fast_Blur_16( bmp    : not null A_Bitmap;
                                x1, y1 : Natural;
                                x2, y2 : Natural;
                                passes : Positive ) is
            c1, c2, c3 : Color_Type;
        begin
            for p in 1..passes loop
                for y in y1..y2 loop
                    c1 := Memory_Getpixel_15_or_16_bpp( bmp, x1, y );
                    for x in x1..x2-1 loop
                        c2 := Memory_Getpixel_15_or_16_bpp( bmp, x + 1, y );
                        c3 := ((((c1 and 16#F800#) + (c2 and 16#F800#)) / 2 ) and 16#F800#) or
                              ((((c1 and 16#07E0#) + (c2 and 16#07E0#)) / 2 ) and 16#07E0#) or
                               (((c1 and 16#001F#) + (c2 and 16#001F#)) / 2 );
                        Memory_Putpixel_15_or_16_bpp( bmp, x, y, c3 );
                        c1 := c2;
                    end loop;
                end loop;
            end loop;
        end Fast_Blur_16;
        pragma Inline_Always( Fast_Blur_16 );

        ------------------------------------------------------------------------

        procedure Fast_Blur_32( bmp    : not null A_Bitmap;
                                x1, y1 : Natural;
                                x2, y2 : Natural;
                                passes : Positive ) is
            c1, c2, c3 : Color_Type;
        begin
            for p in 1..passes loop
                for y in y1..y2 loop
                    c1 := Memory_Getpixel_32( bmp, x1, y );
                    for x in x1..x2-1 loop
                        c2 := Memory_Getpixel_32( bmp, x + 1, y );
                        c3 := ((((c1 and 16#FF0000#) + (c2 and 16#FF0000#)) / 2 ) and 16#FF0000#) or
                              ((((c1 and 16#00FF00#) + (c2 and 16#00FF00#)) / 2 ) and 16#00FF00#) or
                               (((c1 and 16#0000FF#) + (c2 and 16#0000FF#)) / 2 );
                        Memory_Putpixel_32( bmp, x, y, c3 );
                        c1 := c2;
                    end loop;
                end loop;
            end loop;
        end Fast_Blur_32;
        pragma Inline_Always( Fast_Blur_32 );

        ------------------------------------------------------------------------

        depth : constant Integer := Get_Color_Depth;
    begin
        if Is_Memory_Bitmap( bmp ) /= 0 then
            if depth = 32 then
                Fast_Blur_32( bmp, x1, y1, x2, y2, passes );
            elsif depth = 16 then
                Fast_Blur_16( bmp, x1, y1, x2, y2, passes );
            end if;
        end if;
    end Fast_Blur;

    ----------------------------------------------------------------------------

    procedure Klingemann_Fast_Blur( bmp    : not null A_Bitmap;
                                    radius : Positive;
                                    passes : Positive ) is
        depth : constant Integer := Get_Color_Depth;
        w     : constant Integer := Get_Width( bmp );
        h     : constant Integer := Get_Height( bmp );
        wm    : constant Integer := w - 1;
        hm    : constant Integer := h - 1;
        wh    : constant Integer := w * h;
        div   : constant Integer := radius + radius + 1;

        r : A_Integer_Array := new Integer_Array(0..wh-1);
        g : A_Integer_Array := new Integer_Array(r'Range);
        b : A_Integer_Array := new Integer_Array(r'Range);

        rSum, gSum, bSum : Integer;

        i1, i2 : Integer;
        c1, c2 : Color_Type;

        vMin : A_Integer_Array := new Integer_Array(0..Integer'Max( w, h )-1);
        vMax : A_Integer_Array := new Integer_Array(vMin'Range);
        dv   : A_Integer_Array := new Integer_Array(0..div*256-1);
    begin
        if Is_Memory_Bitmap( bmp ) = 0 or else depth /= 32 then
            return;
        end if;

        for i in dv'Range loop
            dv(i) := i / div;
        end loop;

        for z in 1..passes loop
            i1 := 0;

            for y in 0..hm loop
                rSum := 0;
                gSum := 0;
                bSum := 0;

                for i in -radius..radius loop
                    c1 := Memory_Getpixel_32( bmp, Integer'Min( wm, Integer'Max( i, 0 ) ), y );
                    rSum := rSum + Integer(Shift_Right( c1 and 16#FF0000#, 16 ));
                    gSum := gSum + Integer(Shift_Right( c1 and 16#00FF00#, 8 ));
                    bSum := bSum +             Integer( c1 and 16#0000FF# );
                end loop;

                for x in 0..wm loop
                    r(i1) := dv(rSum);
                    g(i1) := dv(gSum);
                    b(i1) := dv(bSum);

                    if y = 0 then
                        vMin(x) := Integer'Min( x + radius + 1, wm );
                        vMax(x) := Integer'Max( x - radius, 0 );
                    end if;

                    c1 := Memory_Getpixel_32( bmp, vMin(x), y );
                    c2 := Memory_Getpixel_32( bmp, vMax(x), y );

                    rSum := rSum + Integer(Shift_Right( c1 and 16#FF_00_00#, 16 )) -
                                   Integer(Shift_Right( c2 and 16#FF_00_00#, 16 ));
                    gSum := gSum + Integer(Shift_Right( c1 and 16#00_FF_00#, 8  )) -
                                   Integer(Shift_Right( c2 and 16#00_FF_00#, 8  ));
                    bSum := bSum + Integer(             c1 and 16#00_00_FF# ) -
                                   Integer(             c2 and 16#00_00_FF# );

                    i1 := i1 + 1;
                end loop;
            end loop;

            for x in 0..wm loop
                rSum := 0;
                gSum := 0;
                bSum := 0;

                i1 := (-radius) * w;
                for i in -radius..radius loop
                    i2 := Integer'Max( 0, i1 ) + x;
                    rSum := rSum + r(i2);
                    gSum := gSum + g(i2);
                    bSum := bSum + b(i2);
                    i1 := i1 + w;
                end loop;

                for y in 0..hm loop
                    c1 := Color_Type(Shift_Left( Unsigned_32(dv(rSum)), 16 ) or
                                     Shift_Left( Unsigned_32(dv(gSum)), 8 ) or
                                                 Unsigned_32(dv(bSum)));
                    Memory_Putpixel_32( bmp, x, y, c1 );

                    if x = 0 then
                        vMin(y) := Integer'Min( y + radius + 1, hm ) * w;
                        vMax(y) := Integer'Max( y - radius, 0 ) * w;
                    end if;

                    i1 := x + vMin(y);
                    i2 := x + vMax(y);

                    rSum := rSum + r(i1) - r(i2);
                    gSum := gSum + g(i1) - g(i2);
                    bSum := bSum + b(i1) - b(i2);
                end loop;
            end loop;
        end loop;

        Free( dv );
        Free( vMax );
        Free( vMin );
        Free( b );
        Free( g );
        Free( r );
    end Klingemann_Fast_Blur;

    ----------------------------------------------------------------------------

    procedure Klingemann_Stack_Blur( bmp    : not null A_Bitmap;
                                     radius : Positive;
                                     passes : Positive ) is
        depth : constant Integer := Get_Color_Depth;
        w     : constant Integer := Get_Width( bmp );
        h     : constant Integer := Get_Height( bmp );
        wm    : constant Integer := w - 1;
        hm    : constant Integer := h - 1;
        wh    : constant Integer := w * h;
        div   : constant Integer := radius + radius + 1;
        r1    : constant Integer := radius + 1;

        r : A_Integer_Array := new Integer_Array(0..wh-1);
        g : A_Integer_Array := new Integer_Array(r'Range);
        b : A_Integer_Array := new Integer_Array(r'Range);

        rSum, gSum, bSum : Integer;
        c                : Color_Type;

        vMin   : A_Integer_Array := new Integer_Array(0..Integer'Max( w, h )-1);
        dv     : A_Integer_Array;
        divSum : Integer;

        yi, yp, p : Integer;

        rbs : Integer;

        rOutSum, gOutSum, bOutSum : Integer;
        rInSum, gInSum, bInSum    : Integer;

        type Integer_Array_Array is array (Integer range <>) of A_Integer_Array;

        stack        : Integer_Array_Array(0..div-1) := (others => new Integer_Array(0..2));
        stackPointer : Integer;
        stackStart   : Integer;
        sir          : A_Integer_Array;
    begin
        if Is_Memory_Bitmap( bmp ) = 0 or else depth /= 32 then
            return;
        end if;

        divSum := Integer(Shift_Right( Unsigned_32(div) + 1, 1 ));
        divSum := divSum * divSum;

        dv := new Integer_Array(0..divSum*256-1);
        for i in dv'Range loop
            dv(i) := i / divSum;
        end loop;

        for z in 1..passes loop
            yi := 0;

            for y in 0..hm loop
                rInSum := 0;
                gInSum := 0;
                bInSum := 0;
                rOutSum := 0;
                gOutSum := 0;
                bOutSum := 0;
                rSum := 0;
                gSum := 0;
                bSum := 0;

                for i in -radius..radius loop
                    c := Memory_Getpixel_32( bmp, Integer'Max( i, 0 ), y );
                    sir := stack(i + radius);
                    sir(0) := Integer(Shift_Right( c and 16#FF0000#, 16 ));
                    sir(1) := Integer(Shift_Right( c and 16#00FF00#, 8 ));
                    sir(2) := Integer(             c and 16#0000FF# );

                    rbs := r1 - (abs i);
                    rSum := rSum + sir(0) * rbs;
                    gSum := gSum + sir(1) * rbs;
                    bSum := bSum + sir(2) * rbs;

                    if i > 0 then
                        rInSum := rInSum + sir(0);
                        gInSum := gInSum + sir(1);
                        bInSum := bInSum + sir(2);
                    else
                        rOutSum := rOutSum + sir(0);
                        gOutSum := gOutSum + sir(1);
                        bOutSum := bOutSum + sir(2);
                    end if;
                end loop;

                stackPointer := radius;

                for x in 0..wm loop
                    r(yi) := dv(rSum);
                    g(yi) := dv(gSum);
                    b(yi) := dv(bSum);

                    rSum := rSum - rOutSum;
                    gSum := gSum - gOutSum;
                    bSum := bSum - bOutSum;

                    stackStart := stackPointer - radius + div;
                    sir := stack(stackStart mod div);

                    rOutSum := rOutSum - sir(0);
                    gOutSum := gOutSum - sir(1);
                    bOutSum := bOutSum - sir(2);

                    if y = 0 then
                        vMin(x) := Integer'Min( x + radius + 1, wm );
                    end if;

                    c := Memory_Getpixel_32( bmp, vMin(x), y );
                    sir(0) := Integer(Shift_Right( c and 16#FF0000#, 16 ));
                    sir(1) := Integer(Shift_Right( c and 16#00FF00#, 8 ));
                    sir(2) := Integer(             c and 16#0000FF# );

                    rInSum := rInSum + sir(0);
                    gInSum := gInSum + sir(1);
                    bInSum := bInSum + sir(2);

                    rSum := rSum + rInSum;
                    gSum := gSum + gInSum;
                    bSum := bSum + bInSum;

                    stackPointer := (stackPointer + 1) mod div;
                    sir := stack(stackPointer);

                    rOutSum := rOutSum + sir(0);
                    gOutSum := gOutSum + sir(1);
                    bOutSum := bOutSum + sir(2);

                    rInSum := rInSum - sir(0);
                    gInSum := gInSum - sir(1);
                    bInSum := bInSum - sir(2);

                    yi := yi + 1;
                end loop;
            end loop;

            for x in 0..wm loop
                rInSum := 0;
                gInSum := 0;
                bInSum := 0;
                rOutSum := 0;
                gOutSum := 0;
                bOutSum := 0;
                rSum := 0;
                gSum := 0;
                bSum := 0;

                yp := (-radius) * w;
                for i in -radius..radius loop
                    yi := Integer'Max( 0, yp ) + x;

                    sir := stack(i + radius);

                    sir(0) := r(yi);
                    sir(1) := g(yi);
                    sir(2) := b(yi);

                    rbs := r1 - abs i;
                    rSum := rSum + r(yi) * rbs;
                    gSum := gSum + g(yi) * rbs;
                    bSum := bSum + b(yi) * rbs;

                    if i > 0 then
                        rInSum := rInSum + sir(0);
                        gInSum := gInSum + sir(1);
                        bInSum := bInSum + sir(2);
                    else
                        rOutSum := rOutSum + sir(0);
                        gOutSum := gOutSum + sir(1);
                        bOutSum := bOutSum + sir(2);
                    end if;

                    if i < hm then
                        yp := yp + w;
                    end if;
                end loop;

                yi := x;
                stackPointer := radius;
                for y in 0..hm loop
                    c := Color_Type(Shift_Left( Unsigned_32(dv(rSum)), 16 ) or
                                    Shift_Left( Unsigned_32(dv(gSum)), 8 ) or
                                                Unsigned_32(dv(bSum)));
                    Memory_Putpixel_32( bmp, x, y, c );

                    rSum := rSum - rOutSum;
                    gSum := gSum - gOutSum;
                    bSum := bSum - bOutSum;

                    stackStart := stackPointer - radius + div;
                    sir := stack(stackStart mod div);

                    rOutSum := rOutSum - sir(0);
                    gOutSum := gOutSum - sir(1);
                    bOutSum := bOutSum - sir(2);

                    if x = 0 then
                        vMin(y) := Integer'Min( y + r1, hm ) * w;
                    end if;
                    p := x + vMin(y);

                    sir(0) := r(p);
                    sir(1) := g(p);
                    sir(2) := b(p);

                    rInSum := rInSum + sir(0);
                    gInSum := gInSum + sir(1);
                    bInSum := bInSum + sir(2);

                    rSum := rSum + rInSum;
                    gSum := gSum + gInSum;
                    bSum := bSum + bInSum;

                    stackPointer := (stackPointer + 1) mod div;
                    sir := stack(stackPointer);

                    rOutSum := rOutSum + sir(0);
                    gOutSum := gOutSum + sir(1);
                    bOutSum := bOutSum + sir(2);

                    rInSum := rInSum - sir(0);
                    gInSum := gInSum - sir(1);
                    bInSum := bInSum - sir(2);

                    yi := yi + w;
                end loop;
            end loop;

        end loop;

        for i in stack'Range loop
            Free( stack(i) );
        end loop;
        Free( dv );
        Free( vMin );
        Free( b );
        Free( g );
        Free( r );
    end Klingemann_Stack_Blur;

end Image_Filters;




--Convolver c;
--BImage a;
--BImage b;
--
--
--void setup()
--{
--
--  a=loadImage("cecelia2.jpg");
--  size(a.width, a.height);
--  b=new BImage(a.width, a.height);
--  image(b, 0, 0);
--  noStroke();
--  framerate(30);
--  c=new Convolver(1);
--}
--
--void loop()
--{
--  c.setRadius(mouseY>>4);
--  System.arraycopy(a.pixels,0,b.pixels,0,a.pixels.length);
--  c.blur(b,mouseX-30,0,60,height);
--  image(b, 0, 0);
--
--}
--
--class Convolver{
--  int radius;
--  int kernelSize;
--  int[] kernel;
--  int[][] mult;
--
--  Convolver(int sz){
--    this.setRadius(sz);
--  }
--
--  void setRadius(int sz){
--
--    int i,j;
--    sz=min(max(1,sz),248);
--    if (radius==sz) return;
--    kernelSize=1+sz*2;
--    radius=sz;
--    kernel= new int[1+sz*2];
--    mult=new int[1+sz*2][256];
--
--    int sum=0;
--    for (i=1;i<sz;i++){
--      int szi=sz-i;
--      kernel[sz+i]=kernel[szi]=szi*szi;
--      sum+=kernel[szi]+kernel[szi];
--      for (j=0;j<256;j++){
--        mult[sz+i][j]=mult[szi][j]=kernel[szi]*j;
--      }
--    }
--    kernel[sz]=sz*sz;
--    sum+=kernel[sz];
--    for (j=0;j<256;j++){
--      mult[sz][j]=kernel[sz]*j;
--    }
--  }
--
--  void blur(BImage img,int x, int y,int w,int h){
--
--    int sum,cr,cg,cb,k;
--    int pixel,read,i,ri,xl,yl,yi,ym,riw;
--    int[] pix=img.pixels;
--    int iw=img.width;
--
--    int wh=iw*img.height;
--
--    int r[]=new int[wh];
--    int g[]=new int[wh];
--    int b[]=new int[wh];
--
--    for (i=0;i<wh;i++){
--      ri=pix[i];
--      r[i]=(ri&0xff0000)>>16;
--      g[i]=(ri&0x00ff00)>>8;
--      b[i]=(ri&0x0000ff);
--    }
--
--    int r2[]=new int[wh];
--    int g2[]=new int[wh];
--    int b2[]=new int[wh];
--
--    x=max(0,x);
--    y=max(0,y);
--    w=x+w-max(0,(x+w)-iw);
--    h=y+h-max(0,(y+h)-img.height);
--    yi=y*iw;
--
--    for (yl=y;yl<h;yl++){
--      for (xl=x;xl<w;xl++){
--        cb=cg=cr=sum=0;
--        ri=xl-radius;
--        for (i=0;i<kernelSize;i++){
--          read=ri+i;
--          if (read>=x && read<w){
--            read+=yi;
--            cr+=mult[i][r[read]];
--            cg+=mult[i][g[read]];
--            cb+=mult[i][b[read]];
--            sum+=kernel[i];
--          }
--        }
--        ri=yi+xl;
--        r2[ri]=cr/sum;
--        g2[ri]=cg/sum;
--        b2[ri]=cb/sum;
--      }
--      yi+=iw;
--    }
--    yi=y*iw;
--
--    for (yl=y;yl<h;yl++){
--      ym=yl-radius;
--      riw=ym*iw;
--      for (xl=x;xl<w;xl++){
--        cb=cg=cr=sum=0;
--        ri=ym;
--        read=xl+riw;
--        for (i=0;i<kernelSize;i++){
--          if (ri<h && ri>=y){
--            cr+=mult[i][r2[read]];
--            cg+=mult[i][g2[read]];
--            cb+=mult[i][b2[read]];
--            sum+=kernel[i];
--          }
--          ri++;
--          read+=iw;
--        }
--        pix[xl+yi]=0xff000000 | (cr/sum)<<16 | (cg/sum)<<8 | (cb/sum);
--      }
--      yi+=iw;
--    }
--  }
--}
