
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Tags;                          use Ada.Tags;
with Ada.Unchecked_Deallocation;
--with Debugging;                         use Debugging;

package body Maps is

    allocate : Allocator := null;

    ----------------------------------------------------------------------------

    function Create_Map( width, height : Positive ) return A_Map is
    begin
        if allocate /= null then
            return allocate.all( width, height );
        else
            raise Constraint_Error with "Map allocator not registered";
        end if;
    end Create_Map;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Map_Object ) is
    begin
        Object(this.all).Adjust;
        this.layers := Copy( this.layers );
    end Adjust;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this   : access Map_Object;
                         width,
                         height,
                         layers : Positive ) is
    begin
        Object(this.all).Construct;

        if width > MAX_WIDTH or else height > MAX_HEIGHT then
            raise Constraint_Error with "Map size too large";
        end if;

        this.width := width;
        this.height := height;
        this.layers := new Layer_Array(1..layers);
        for i in this.layers'Range loop
            this.layers(i) := new Layer(width*height);
            for j in this.layers(i).data'Range loop
                this.layers(i).data(j) := 0;
            end loop;
        end loop;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Map_Object ) is
    begin
        Delete( this.layers );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access Map_Object'Class ) return Positive is
    begin
        return this.height;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Layers( this : not null access Map_Object'Class ) return Positive is
    begin
        return this.layers'Length;
    end Get_Layers;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Layers_Data( this : access Map_Object ) return A_Layer_Array is
    begin
        return this.layers;
    end Get_Layers_Data;

    ----------------------------------------------------------------------------

    function Get_Tile_Id( this  : not null access Map_Object'Class;
                          layer : Positive;
                          x, y  : Integer ) return Natural is
    begin
        if layer <= this.layers'Last and then
           x >= 0 and then x < this.width and then
           y >= 0 and then y < this.height
        then
            return this.layers(layer).data(this.layers(layer).data'First + y * this.width + x);
        end if;
        return 0;
    end Get_Tile_Id;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access Map_Object'Class ) return Positive is
    begin
        return this.width;
    end Get_Width;

    ----------------------------------------------------------------------------

    procedure Resize( this   : not null access Map_Object'Class;
                      width,
                      height : Positive ) is
        newMap : A_Map;
    begin
        newMap := Create_Map( width, height );
        for l in this.layers'Range loop
            for y in 0..this.height-1 loop
                for x in 0..this.width-1 loop
                    newMap.Set_Tile( l, x, y, this.Get_Tile_Id( l, x, y ) );
                end loop;
            end loop;
        end loop;
        this.width := width;
        this.height := height;
        Move( this.layers, newMap.layers );
        Delete( newMap );
    end Resize;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Tile( this  : access Map_Object;
                        layer : Positive;
                        x, y  : Integer;
                        tile  : Natural ) is
    begin
        if layer <= this.layers'Last and then
           x >= 0 and then x < this.width and then
           y >= 0 and then y < this.height
        then
            this.layers(layer).data(this.layers(layer).data'First + y * this.width + x) := tile;
        end if;
    end Set_Tile;

    ----------------------------------------------------------------------------

    function Tile_Width( this : not null access Map_Object'Class ) return Positive is
        pragma Unreferenced( this );
    begin
        return 16;
    end Tile_Width;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Map_Object ) is
    begin
        Object'Read( stream, Object(obj) );
        obj.width := Positive'Input( stream );
        obj.height := Positive'Input( stream );

        if obj.width > MAX_WIDTH or else obj.height > MAX_HEIGHT then
            raise Constraint_Error with "Map size too large";
        end if;

        obj.layers := A_Layer_Array'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Map_Object ) is
    begin
        Object'Write( stream, Object(obj) );
        Positive'Output( stream, obj.width );
        Positive'Output( stream, obj.height );
        A_Layer_Array'Output( stream, obj.layers );
    end Object_Write;

    --==========================================================================

    function A_Map_Input( stream : access Root_Stream_Type'Class ) return A_Map is
        map : A_Map;
    begin
        if Boolean'Input( stream ) then
            A_Map'Read( stream, map );
        end if;
        return map;
    end A_Map_Input;

    ----------------------------------------------------------------------------

    procedure A_Map_Output( stream : access Root_Stream_Type'Class; map : A_Map ) is
    begin
        Boolean'Output( stream, map /= null );
        if map /= null then
            A_Map'Write( stream, map );
        end if;
    end A_Map_Output;

    ----------------------------------------------------------------------------

    procedure A_Map_Read( stream : access Root_Stream_Type'Class; map : out A_Map ) is
    begin
        map := new Map_Object'Class'(Map_Object'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown map class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end A_Map_Read;

    ----------------------------------------------------------------------------

    procedure A_Map_Write( stream : access Root_Stream_Type'Class; map : A_Map ) is
    begin
        Map_Object'Class'Output( stream, map.all );
    end A_Map_Write;

    ----------------------------------------------------------------------------

    function Copy( src : A_Map ) return A_Map is
    begin
        return A_Map(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Map ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    --==========================================================================

    function Copy( src : A_Layer_Array ) return A_Layer_Array is
        dest : constant A_Layer_Array := new Layer_Array(src'Range);
    begin
        for i in src'Range loop
            dest(i) := new Layer'(src(i).all);
        end loop;
        return dest;
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( la : in out A_Layer_Array ) is
        procedure Free is new Ada.Unchecked_Deallocation( Layer_Array, A_Layer_Array );
    begin
        if la /= null then
            for i in la'Range loop
                Delete( la(i) );
            end loop;
            Free( la );
        end if;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Move( dst, src : in out A_Layer_Array ) is
    begin
        Delete( dst );
        dst := src;
        src := null;
    end Move;

    ----------------------------------------------------------------------------

    procedure Delete( lyr : in out A_Layer ) is
        procedure Free is new Ada.Unchecked_Deallocation( Layer, A_Layer );
    begin
        if lyr /= null then
            Free( lyr );
        end if;
    end Delete;

    ----------------------------------------------------------------------------

    function A_Layer_Input( stream : access Root_Stream_Type'Class ) return A_Layer is
        lyr      : A_Layer := null;
        size     : Positive;
        offset   : Positive;
        val      : Integer;
        tile     : Natural := 0;
        needTile : Boolean := True;
    begin
        if Boolean'Input( stream ) then
            size := Integer'Max( 1, Integer'Input( stream ) );
            lyr := new Layer(size);
            offset := lyr.data'First;

            loop
                val := Integer'Input( stream );

                -- always accept tiles
                if val >= 0 then
                    tile := val;
                    lyr.data(offset) := tile;
                    offset := offset + 1;
                    needTile := False;
                -- only accept a length if we don't need a tile
                elsif not needTile then
                    -- val is negative of length
                    for i in offset..Integer'Min( offset - val - 2, lyr.data'Last ) loop
                        lyr.data(offset) := tile;
                        offset := offset + 1;
                    end loop;
                    needTile := True;
                else
                    -- we needed a tile and didn't find one
                    raise Constraint_Error with "Invalid map data encoding";
                end if;
                exit when offset > lyr.data'Last;
            end loop;
            lyr.physical := Boolean'Input( stream );
        end if;
        return lyr;
    exception
        when others =>
            Delete( lyr );
            raise;
    end A_Layer_Input;

    ----------------------------------------------------------------------------

    procedure A_Layer_Output( stream : access Root_Stream_Type'Class; lyr : A_Layer ) is
        offset : Integer := lyr.data'First;
        run    : Integer := 1;
    begin
        Boolean'Output( stream, lyr /= null );
        if lyr /= null then
            Integer'Output( stream, lyr.data'Length );
            while offset <= lyr.data'Last loop
                if run = 1 then
                    -- write out the tile id at the start of a run
                    Integer'Output( stream, lyr.data(offset) );
                end if;
                if offset < lyr.data'Last and then lyr.data(offset) = lyr.data(offset + 1) then
                    -- the run will continue
                    if run < Integer'Last - 1 then
                        -- just increment length
                        run := run + 1;
                    else
                        -- we hit max run length. this will never happen.
                        -- to handle this case anyway, we just write out the
                        -- current run length and start a new run that includes
                        -- the current tile.
                        Integer'Output( stream, -run );
                        Integer'Output( stream, lyr.data(offset) );
                        run := 2;
                    end if;
                elsif run > 1 then
                    -- the run has ended, write out the length
                    Integer'Output( stream, -run );
                    run := 1;
                end if;
                offset := offset + 1;
            end loop;
            Boolean'Output( stream, lyr.physical );
        end if;
    end A_Layer_Output;

    ----------------------------------------------------------------------------

    function A_Layer_Array_Input( stream : access Root_Stream_Type'Class ) return A_Layer_Array is
        la  : A_Layer_Array := null;
        len : Integer;
    begin
        if Boolean'Input( stream ) then
            len := Integer'Input( stream );
            la := new Layer_Array(1..len);
            for i in la'Range loop
                la(i) := A_Layer'Input( stream );
            end loop;
        end if;
        return la;
    exception
        when others =>
            Delete( la );
            raise;
    end A_Layer_Array_Input;

    ----------------------------------------------------------------------------

    procedure A_Layer_Array_Output( stream : access Root_Stream_Type'Class; la : A_Layer_Array ) is
    begin
        Boolean'Output( stream, la /= null );
        if la /= null then
            Integer'Output( stream, la'Length );
            for i in la'Range loop
                A_Layer'Output( stream, la(i) );
            end loop;
        end if;
    end A_Layer_Array_Output;

    --==========================================================================

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Maps.allocate := allocate;
    end Register_Allocator;

end Maps;
