
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Equal_Case_Insensitive;
with Ada.Strings.Hash_Case_Insensitive;
with Ada.Streams;                       use Ada.Streams;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Ada.Unchecked_Deallocation;
with Allegro.Files.Memory;              use Allegro.Files.Memory;
with Archives;                          use Archives;
with Content_Signing;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Game;                       use Events.Game;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Resources is

    -- File extension for resolving archive files when searching for a resource.
    ARCHIVE_EXTENSION : constant String := "zip";

    ----------------------------------------------------------------------------

    protected body Async_Operation is

        procedure Complete is
        begin
            c := True;
        end Complete;

        ------------------------------------------------------------------------

        function Is_Complete return Boolean is
        begin
            return c;
        end Is_Complete;

        ------------------------------------------------------------------------

        entry Wait when c is
        begin
            null;
        end Wait;

    end Async_Operation;

    ----------------------------------------------------------------------------

    procedure Delete( op : in out A_Async_Operation ) is
        procedure Free is new Ada.Unchecked_Deallocation( Async_Operation, A_Async_Operation );
    begin
        Free( op );
    end Delete;

    --==========================================================================

    protected Load_Count is
        procedure Increment;
        procedure Decrement;
    private
        count : Natural := 0;
    end Load_Count;

    ----------------------------------------------------------------------------

    protected body Load_Count is

        procedure Increment is
        begin
            count := count + 1;
            if count = 1 then
                Queue_Loading( True );
            end if;
        end Increment;

        ------------------------------------------------------------------------

        procedure Decrement is
        begin
            if count > 0 then
                count := count - 1;
                if count = 0 then
                    Queue_Loading( False );
                end if;
            end if;
        end Decrement;

    end Load_Count;

    ----------------------------------------------------------------------------

    procedure Loading_Begins is
    begin
        Load_Count.Increment;
    end Loading_Begins;

    ----------------------------------------------------------------------------

    procedure Loading_Ends is
    begin
        Load_Count.Decrement;
    end Loading_Ends;

    --==========================================================================

    procedure Create_Resource( this     : out A_Resource_File;
                               filepath : String;
                               group    : String;
                               data     : in out A_SEA ) is
    begin
        this := new Resource_File;
        this.filepath := To_Unbounded_String( filepath );
        this.group := To_Unbounded_String( group );
        this.data := data;
        data := null;
    end Create_Resource;

    ----------------------------------------------------------------------------

    function Get_Address( this : access Resource_File ) return Address is
    begin
        if this.data'Length > 0 then
            return this.data(this.data'First)'Address;
        end if;
        return Null_Address;
    end Get_Address;

    ----------------------------------------------------------------------------

    function Get_Filename( this : access Resource_File ) return String is
    begin
        return Get_Filename( To_String( this.filepath ) );
    end Get_Filename;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Id( this : access Resource_File ) return String is
    begin
        return To_String( this.group ) & ':' & To_String( this.filepath );
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Packfile( this : access Resource_File ) return A_Packfile is
    begin
        return Pack_FOpen( this.data, consume => False );
    end Get_Packfile;

    ----------------------------------------------------------------------------

    not overriding
    procedure Get_Stream( this : access Resource_File;
                          strm : out A_Buffer_Stream ) is
    begin
        strm := Stream( this.data );
    end Get_Stream;

    ----------------------------------------------------------------------------

    not overriding
    procedure Reference( this : access Resource_File ) is
    begin
        this.refs := this.refs + 1;
        this.lastUse := Clock;
    end Reference;

    ----------------------------------------------------------------------------

    not overriding
    function Size( this : access Resource_File ) return Unsigned_32 is
    begin
        return Unsigned_32(this.data'Length);
    end Size;

    ----------------------------------------------------------------------------

    not overriding
    procedure Unreference( this : access Resource_File ) is
    begin
        this.refs := this.refs - 1;
    end Unreference;

    --==========================================================================

    function Load_From_Disk( filepath : String; group : String ) return A_SEA is

        ------------------------------------------------------------------------

        function Read_File( path : String ) return A_SEA is
            len    : constant Long_Integer := File_Length( path );
            ft     : File_Type;
            buffer : A_SEA := null;
        begin
            if len >= 0 then
                Open( ft, In_File, path );
                if Is_Open( ft ) then
                    pragma Debug( Dbg( group & ':' & filepath & " -> " & path,
                                       D_RES, INFO ) );
                    buffer := new Stream_Element_Array(0..Stream_Element_Offset(len-1));
                    Stream_Element_Array'Read( Stream( ft ), buffer.all );
                    Close( ft );
                    if not Content_Signing.Is_Valid( path, buffer.all ) then
                        Delete( buffer );
                        pragma Debug( Dbg( path & " signature does not match",
                                           D_RES, ERROR ) );
                    end if;
                end if;
            end if;
            return buffer;
        exception
            when others =>
                Delete( buffer );
                if Is_Open( ft ) then
                    Close( ft );
                end if;
                return null;
        end Read_File;

        ------------------------------------------------------------------------

        -- 'path' is the path of the archive and 'filename' is the name of the
        -- file within the archive to read.
        function Read_From_Archive( path     : String;
                                    filename : String ) return A_SEA is
            archive : A_Archive;
            buffer  : A_SEA := null;
        begin
            archive := Load_Archive( path );
            buffer := A_SEA(archive.Open_Chunk( filename ));   -- may raise exception
            pragma Debug( Dbg( group & ':' & filepath & " -> " &
                               path & "/" & filename, D_RES, INFO ) );
            Delete( archive );
            return buffer;
        exception
            when others =>
                Delete( buffer );
                Delete( archive );
                return null;
        end Read_From_Archive;

        ------------------------------------------------------------------------

        media : constant String := Get_Pref( "application", "media" );
        data  : A_SEA;
    begin
        -- 1. Look at absolute filepath
        if Is_Absolute_Path( filepath ) and then Is_Regular_File( filepath ) then
            data := Read_File( filepath );
        end if;

        -- 2. Look at working_dir/filename
        if data = null and then Is_Regular_File( Get_Filename( filepath ) ) then
            data := Read_File( Get_Filename( filepath ) );
        end if;

        -- 3. Look at working_dir/group.archive/filename
        if data = null and then
           group'Length > 0 and then
           Is_Regular_File( group & '.' & ARCHIVE_EXTENSION )
       then
            data := Read_From_Archive( group & '.' & ARCHIVE_EXTENSION,
                                       Get_Filename( filepath ) );
        end if;

        -- 4. Look at media_dir/filename
        if data = null and then Is_Regular_File( media & Get_Filename( filepath ) ) then
            data := Read_File( media & Get_Filename( filepath ) );
        end if;

        -- 5. Look at media_dir/group/filename
        if data = null and then Is_Regular_File( media & group & Slash & Get_Filename( filepath ) ) then
            data := Read_File( media & group & Slash & Get_Filename( filepath ) );
        end if;

        -- 6. Look at media_dir/group.archive/filename
        if data = null and then
           group'Length > 0 and then
           Is_Regular_File( media & group & '.' & ARCHIVE_EXTENSION )
        then
            data := Read_From_Archive( media & group & '.' & ARCHIVE_EXTENSION,
                                       Get_Filename( filepath ) );
        end if;

        return data;
    end Load_From_Disk;

    ----------------------------------------------------------------------------

    -- Finds the physical file containing the specified resource, or the empty
    -- string if it doesn't exist. The search follows the same algorithm as
    -- Load_From_Disk. If the resource is in an archive, the path of the
    -- archive will be returned.
    function Find_On_Disk( filepath : String; group : String ) return String is

        ------------------------------------------------------------------------

        -- 'path' is the path of the archive to test and 'filename' is the name
        -- of the file within the archive to test.
        function Exists_In_Archive( path     : String;
                                    filename : String ) return Boolean is
            archive : A_Archive;
            exists  : Boolean := False;
        begin
            archive := Load_Archive( path );
            exists := archive.Exists( filename );
            Delete( archive );
            return exists;
        exception
            when others =>
                Delete( archive );
                return False;
        end Exists_In_Archive;

        ------------------------------------------------------------------------

        media : constant String := Get_Pref( "application", "media" );
    begin
        -- 1. Look at filepath (if it is absolute)
        if Is_Absolute_Path( filepath ) and then Is_Regular_File( filepath ) then
            return filepath;
        end if;

        -- 2. Look at working_dir/filename
        if Is_Regular_File( Get_Filename( filepath ) ) then
            return Get_Filename( filepath );
        end if;

        -- 3. Look at working_dir/group.archive/filename
        if group'Length > 0 and then
           Is_Regular_File( group & '.' & ARCHIVE_EXTENSION ) and then
           Exists_In_Archive( group & '.' & ARCHIVE_EXTENSION, Get_Filename( filepath ) )
       then
           return group & '.' & ARCHIVE_EXTENSION;
        end if;

        -- 4. Look at media_dir/filename
        if Is_Regular_File( media & Get_Filename( filepath ) ) then
            return media & Get_Filename( filepath );
        end if;

        -- 5. Look at media_dir/group/filename
        if Is_Regular_File( media & group & Slash & Get_Filename( filepath ) ) then
            return media & group & Slash & Get_Filename( filepath );
        end if;

        -- 6. Look at media_dir/group.archive/filename
        if group'Length > 0 and then
           Is_Regular_File( media & group & '.' & ARCHIVE_EXTENSION ) and then
           Exists_In_Archive( media & group & '.' & ARCHIVE_EXTENSION, Get_Filename( filepath ) )
        then
            return media & group & '.' & ARCHIVE_EXTENSION;
        end if;

        -- Resource not found
        return "";
    end Find_On_Disk;

    ----------------------------------------------------------------------------

    package Maps is new Ada.Containers.Indefinite_Hashed_Maps( String,
                                                               A_Resource_File,
                                                               Ada.Strings.Hash_Case_Insensitive,
                                                               Ada.Strings.Equal_Case_Insensitive,
                                                               "=" );

    protected File_Cache is

        -- Returns a Resource_File in the cache, or loads it if it hasn't
        -- been loaded. 'resource' is returned as null on error.
        procedure Load( filepath : String;
                        group    : String;
                        resource : out A_Resource_File );
        pragma Precondition( filepath'Length > 0 );

        procedure Unload( resource : in out A_Resource_File );
        pragma Precondition( resource /= null );
        pragma Postcondition( resource = null );

    private
        map  : Maps.Map;
        size : Unsigned_32 := 0;
    end File_Cache;

    ----------------------------------------------------------------------------

    protected body File_Cache is

        procedure Load( filepath : String;
                        group    : String;
                        resource : out A_Resource_File ) is
            pos  : Maps.Cursor := map.Find( group & ':' & filepath );
            tmp  : Boolean;
            data : A_SEA;
        begin
            if Maps.Has_Element( pos ) then
                resource := Maps.Element( pos );
                resource.Reference;
            else
                data := Load_From_Disk( filepath, group );
                if data /= null then
                    Create_Resource( resource, filepath, group, data );
                    map.Insert( group & ':' & filepath, resource, pos, tmp );
                    size := size + resource.Size;
                    pragma Debug( Dbg( "Resources: Cache size:" &
                                       Unsigned_32'Image( size / 1024 ) & "KB",
                                       D_RES, INFO ) );
                else
                    resource := null;
                end if;
            end if;
        end Load;

        ------------------------------------------------------------------------

        procedure Unload( resource : in out A_Resource_File ) is
            pos  : constant Maps.Cursor := map.Find( resource.Get_Id );
        begin
            if Maps.Has_Element( pos ) then
                if resource.refs > 0 then
                    resource.refs := resource.refs - 1;
                else
                    null;
                    pragma Debug( Dbg( "Resources.File_Cache: " &
                                       "Reference count error for <" &
                                       resource.Get_Id & '>', D_RES, ERROR ) );
                end if;
            else
                null;
                pragma Debug( Dbg( "Resources.File_Cache: " &
                                   "Unloaded unknown resource <" &
                                   resource.Get_Id & '>', D_RES, ERROR ) );
            end if;
            resource := null;
        end Unload;

    end File_Cache;

    ----------------------------------------------------------------------------

    function Load_Resource( filepath : String; group : String ) return A_Resource_File is
        resource : A_Resource_File;
    begin
        File_Cache.Load( filepath, group, resource );
        return resource;
    end Load_Resource;

    ----------------------------------------------------------------------------

    procedure Unload_Resource( resource : in out A_Resource_File ) is
    begin
        if resource /= null then
            File_Cache.Unload( resource );
        end if;
    end Unload_Resource;

    ----------------------------------------------------------------------------

    function Write_Path( filepath : String; group : String ) return String is
        found : constant String := Find_On_Disk( filepath, group );
    begin
        if found'Length > 0 then
            if Get_Filename( found ) /= Get_Filename( filepath ) then
                -- found it in an archive, return the file as though it were
                -- in the same directory as the archive
                return Normalize_Pathname( Get_Directory( found ) & Get_Filename( filepath ) );
            end if;
            return Normalize_Pathname( found );
        end if;

        -- didn't find the file, return a path to the working directory
        return Normalize_Pathname( Get_Filename( filepath ) );
    end Write_Path;

end Resources;
