
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Ada.Unchecked_Deallocation;
with Allegro.Files;                     use Allegro.Files;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Streams.Packfiles;                 use Streams.Packfiles;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

package body Tiles.Indices is

    protected body Tile_Queue is

        procedure Add_Back( tile : not null A_Tile ) is
        begin
            if not stopped then
                queue.Append( tile );
                total := total + 1;
            end if;
        end Add_Back;

        ------------------------------------------------------------------------

        function Get_Progress return Natural is
        begin
            return (loaded * 100) / total;
        end Get_Progress;

        ------------------------------------------------------------------------

        procedure Prioritize( tile : not null A_Tile ) is
        begin
            if not stopped then
                queue.Prepend( tile );
            end if;
        end Prioritize;

        ------------------------------------------------------------------------

        procedure Stop is
        begin
            if not stopped then
                queue.Clear;
                stopped := True;
            end if;
        end Stop;

        ------------------------------------------------------------------------

        procedure Remove( tile : out A_Tile ) is
        begin
            if not stopped then
                loop
                    if not queue.Is_Empty then
                        tile := queue.First_Element;
                        queue.Delete_First;
                        if not tile.loadState.Is_Complete then
                            -- this one hasn't been loaded yet, return it
                            loaded := loaded + 1;
                            exit;
                        end if;
                    else
                        -- no more tiles on the queue
                        tile := null;
                        exit;
                    end if;
                end loop;
            else
                tile := null;
            end if;
        end Remove;

    end Tile_Queue;

    ----------------------------------------------------------------------------

    procedure Delete( queue : in out A_Tile_Queue ) is
        procedure Free is new Ada.Unchecked_Deallocation( Tile_Queue, A_Tile_Queue );
    begin
        if queue /= null then
            queue.Stop;
            Free( queue );
        end if;
    end Delete;

    --==========================================================================

    function Create_Tile_Index return A_Tile_Index is
        index : constant A_Tile_Index := new Tile_Index;
    begin
        index.loadlist := new Tile_Queue;
        return index;
    end Create_Tile_Index;

    ----------------------------------------------------------------------------

    function Load_Index( archive  : not null A_Archive;
                         filename : String ) return A_Tile_Index is
        index    : A_Tile_Index;
        packfile : A_Packfile;
        pfStream : A_Packfile_Stream;
    begin
        packfile := archive.Open( filename );
        pfStream := Stream( packfile );
        index := A_Tile_Index'Input( pfStream );
        Close( pfStream );
        Pack_FClose0( packfile );
        return index;
    exception
        when others =>
            Close( pfStream );
            Pack_FClose0( packfile );
            return null;
    end Load_Index;

    ----------------------------------------------------------------------------

    procedure Delete( index : in out A_Tile_Index ) is

        ------------------------------------------------------------------------

        procedure Delete_Tile( pos : Tile_Maps.Cursor ) is
            tile : A_Tile := Tile_Maps.Element( pos );
        begin
            Delete( tile );
            index.idmap.Replace_Element( pos, null );
        end Delete_Tile;

        procedure Free is new Ada.Unchecked_Deallocation( Tile_Index, A_Tile_Index );

        ------------------------------------------------------------------------

        procedure Delete_Matrix( pos : Matrix_Vectors.Cursor ) is
            matrix : A_Tile_Matrix := Matrix_Vectors.Element( pos );
        begin
            Delete( matrix );
        end Delete_Matrix;

        ------------------------------------------------------------------------

    begin
        if index /= null then
            index.idmap.Iterate( Delete_Tile'Access );
            index.list.Clear;
            index.matlist.Iterate( Delete_Matrix'Access );
            index.matlist.Clear;
            Delete( index.loadlist );
            Free( index );
        end if;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Add_Matrix( index : not null A_Tile_Index; matrix : in out A_Tile_Matrix ) is
    begin
        index.matlist.Append( matrix );
        matrix := null;
    end Add_Matrix;

    ----------------------------------------------------------------------------

    procedure Add_Tile( index : not null A_Tile_Index; tile : in out A_Tile ) is
    begin
        if tile /= null then
            if tile.Get_Id /= 0 then
                if not index.idmap.Contains( tile.Get_Id ) then
                    index.idmap.Insert( tile.Get_Id, tile );
                    index.loadlist.Add_Back( tile );
                else
                    raise DUPLICATE_TILE with
                          "Duplicate tile ID:" & Natural'Image( tile.Get_Id );
                end if;
            end if;
            index.list.Append( tile.Get_Id );
            tile := null;
        else
            index.list.Append( 0 );
        end if;
    end Add_Tile;

    ----------------------------------------------------------------------------

    function Get_Matrix( index : not null A_Tile_Index; num : Natural ) return A_Tile_Matrix is
    begin
        if num > 0 and then num <= Integer(index.matlist.Length) then
            return index.matlist.Element( num );
        end if;
        return null;
    end Get_Matrix;

    ----------------------------------------------------------------------------

    function Get_Matrix_Count( index : not null A_Tile_Index ) return Natural is
    begin
        return Natural(index.matlist.Length);
    end Get_Matrix_Count;

    ----------------------------------------------------------------------------

    function Get_Progress( index : not null A_Tile_Index ) return Natural is
    begin
        return index.loadlist.Get_Progress;
    end Get_Progress;

    ----------------------------------------------------------------------------

    function Get_Slot( index : not null A_Tile_Index; id : Natural ) return Natural is
    begin
        if id < 1 then
            return 0;
        end if;
        for slot in 1..index.list.Length loop
            if index.list.Element( Natural(slot) ) = id then
                return Natural(slot);
            end if;
        end loop;
        return 0;
    end Get_Slot;

    ----------------------------------------------------------------------------

    function Get_Tile( index : not null A_Tile_Index; id : Natural ) return A_Tile is
    begin
        return index.idmap.Element( id );
    exception
        when Constraint_Error =>
            return null;
    end Get_Tile;

    ----------------------------------------------------------------------------

    function Get_Tile( index : not null A_Tile_Index; name : String ) return A_Tile is

        ------------------------------------------------------------------------

        function Find_Tile( name : String ) return A_Tile is
            lname : constant String := To_Lower( name );
            crsr  : Tile_Maps.Cursor := index.idmap.First;
            tile  : A_Tile;
        begin
            while Tile_Maps.Has_Element( crsr ) loop
                tile := Tile_Maps.Element( crsr );
                if lname = To_Lower( tile.Get_Name ) then
                    return tile;
                end if;
                crsr := Tile_Maps.Next( crsr );
            end loop;
            return null;
        end Find_Tile;

        ------------------------------------------------------------------------

        tile : A_Tile;
    begin
        if Get_Extension( name )'Length = 0 then
            tile := Find_Tile( Add_Extension( name, "png" ) );
            if tile = null then
                tile := Find_Tile( Add_Extension( name, "bmp" ) );
            end if;
        else
            tile := Find_Tile( name );
        end if;
        return tile;
    end Get_Tile;

    ----------------------------------------------------------------------------

    function Get_Tile_At_Slot( index : not null A_Tile_Index;
                               slot  : Natural ) return A_Tile is
    begin
        return index.idmap.Element( index.list.Element( slot ) );
    exception
        when Constraint_Error =>
            return null;
    end Get_Tile_At_Slot;

    ----------------------------------------------------------------------------

    function Get_Tile_Count( index : not null A_Tile_Index ) return Natural is
    begin
        return Natural(index.list.Length);
    end Get_Tile_Count;

    ----------------------------------------------------------------------------

    procedure Iterate_Tiles( index   : not null A_Tile_Index;
                             examine : not null access procedure( tile : A_Tile ) ) is

        procedure Local_Examine( pos : Tile_Maps.Cursor ) is
        begin
            examine.all( Tile_Maps.Element( pos ) );
        end Local_Examine;

    begin
        index.idmap.Iterate( Local_Examine'Access );
    end Iterate_Tiles;

    ----------------------------------------------------------------------------

    procedure Load_Complete( index : not null A_Tile_Index ) is
    begin
        -- clear the load list and prevent any further additions
        index.loadlist.Stop;
    end Load_Complete;

    ----------------------------------------------------------------------------

    procedure Load_Images( index   : not null A_Tile_Index;
                           archive : not null A_Archive ) is
        tile : A_Tile;
    begin
        loop
            index.loadlist.Remove( tile );
            exit when tile = null;

            tile.Load_Bitmap( archive );
        end loop;
        index.loadlist.Stop;
    end Load_Images;

    ----------------------------------------------------------------------------

    procedure Prioritize_Tile( index : not null A_Tile_Index; tile : not null A_Tile ) is
    begin
        tile.lock.Lock;
        if tile.priority = 0 then
            tile.priority := 1;
            index.loadlist.Prioritize( tile );
        end if;
        tile.lock.Unlock;
    end Prioritize_Tile;

    ----------------------------------------------------------------------------

    function Write_Index( index    : not null A_Tile_Index;
                          filename : String ) return Boolean is
        file : Ada.Streams.Stream_IO.File_Type;
    begin
        Create( file, Out_File, Add_Extension( filename, Index_Extension ) );
        if Is_Open( file ) then
            A_Tile_Index'Output( Stream( file ), index );
            Close( file );
            return True;
        else
            return False;
        end if;
    exception
        when others =>
            if Is_Open( file ) then
                Close( file );
            end if;
            return False;
    end Write_Index;

    ----------------------------------------------------------------------------

    function A_Tile_Index_Input( stream : access Root_Stream_Type'Class ) return A_Tile_Index is
        index   : A_Tile_Index;
        ident1  : constant String := Index_Identifier;
        ident2  : String(1..ident1'Length);
        version : Integer := 0;
    begin
        begin
            -- verify the header info
            String'Read( stream, ident2 );
            if ident2 /= ident1 then
                -- invalid identifier
                raise Constraint_Error with "Unrecognized file format";
            end if;
            version := Integer'Input( stream );
            if version /= Tile_Version then
                -- unrecognized version
                raise Constraint_Error with "Unrecognized index file version: " &
                                            Image( version );
            end if;
        exception
            when Constraint_Error =>
                raise;
            when e : others =>
                -- file too short
                pragma Debug( Dbg( "Tiles.Indices Exception: " &
                                   Exception_Name( e ) & " - " & Exception_Message( e ),
                                   D_TILES, ERROR ) );
                raise Constraint_Error with "Unrecognized file format";
        end;

        index := Create_Tile_Index;

        -- read the tile list
        declare
            id    : Integer;
            tile  : A_Tile;
            count : constant Natural := Natural'Input( stream );
        begin
            for i in 1..count loop
                id := Integer'Input( stream );
                if id > 0 then
                    tile := A_Tile'Input( stream );
                else
                    tile := null;
                end if;
                Add_Tile( index, tile );
            end loop;
        end;

        -- read the matrix list
        declare
            matrix : A_Tile_Matrix;
            count  : constant Natural := Natural'Input( stream );
        begin
            for i in 1..count loop
                matrix := A_Tile_Matrix'Input( stream );
                Add_Matrix( index, matrix );
            end loop;
        end;

        return index;
    exception
        when others =>
            Delete( index );
            raise;
    end A_Tile_Index_Input;

    ----------------------------------------------------------------------------

    procedure A_Tile_Index_Output( stream : access Root_Stream_Type'Class;
                                   index  : A_Tile_Index ) is
        id : Natural;
    begin
        -- write the header info
        String'Write( stream, Index_Identifier );
        Integer'Output( stream, Tile_Version );

        -- write the tile list
        Natural'Output( stream, Natural(index.list.Length) );
        for i in 1..index.list.Length loop
            id := index.list.Element( Integer(i) );
            Natural'Output( stream, id );
            if id > 0 then
                A_Tile'Output( stream, index.idmap.Element( Integer(id) ) );
            end if;
        end loop;

        -- write the matrix list
        Natural'Output( stream, Natural(index.matlist.Length) );
        for i in 1..index.matlist.Length loop
            A_Tile_Matrix'Output( stream, index.matlist.Element( Integer(i) ) );
        end loop;
    end A_Tile_Index_Output;

    --==========================================================================

    function Index_Extension return String is
    begin
        return "idx";
    end Index_Extension;

    ----------------------------------------------------------------------------

    function Index_Identifier return String is
    begin
        return Tile_Identifier & ":" & Index_Extension;
    end Index_Identifier;

end Tiles.Indices;
