
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Hash_Case_Insensitive;
--with Debugging;                         use Debugging;
with Tiles.Libraries.Loading;           use Tiles.Libraries.Loading;

package body Tiles.Libraries is

    type Cache_Element is
        record
            lib  : A_Tile_Library := null;
            refs : Integer := 1;
        end record;

    package Maps is new
        Ada.Containers.Indefinite_Hashed_Maps( String,
                                               Cache_Element,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               "=" );

    ----------------------------------------------------------------------------

    protected Cache is
        procedure Load_Library( name : String; lib : out A_Tile_Library );
        procedure Unload_Library( name : String );
        procedure Unload_Library( lib : in out A_Tile_Library );
    private
        map : Maps.Map;
    end Cache;

    protected body Cache is

        -- Loads a library into memory or just returns a reference if it's
        -- already in the cache.
        procedure Load_Library( name : String; lib : out A_Tile_Library ) is
            crsr : Maps.Cursor;
            elem : Cache_Element;
        begin
            if map.Contains( name ) then
                crsr := map.Find( name );
                elem := Maps.Element( crsr );
                elem.refs := elem.refs + 1;
                lib := elem.lib;
                map.Replace_Element( crsr, elem );
            else
                elem.lib := Load_From_Disk( name );
                elem.refs := 2;    -- one ref for requestor, one ref for loader
                lib := elem.lib;
                if elem.lib /= null then
                    map.Insert( name, elem );
                end if;
            end if;
        end Load_Library;

        ------------------------------------------------------------------------

        -- Removes a library from memory if no more references remain.
        procedure Unload_Library( name : String ) is
            crsr : Maps.Cursor;
            elem : Cache_Element;
        begin
            crsr := map.Find( name );
            if Maps.Has_Element( crsr ) then
                elem := Maps.Element( crsr );
                elem.refs := elem.refs - 1;
                -- The following code has been commented out to prevent libraries
                -- from being unloaded, saving time when they are needed again.
                -- I could add some more advanced cache purging rules, but
                -- currently the amount of memory required to load all game
                -- assets simultaneously isn't enough to require it.
                --if elem.refs <= 0 then
                --    Delete( elem.lib );
                --    map.Delete( crsr );
                --else
                    map.Replace_Element( crsr, elem );
                --end if;
            end if;
        end Unload_Library;

        ------------------------------------------------------------------------

        -- Removes a library from memory if no more references remain.
        procedure Unload_Library( lib : in out A_Tile_Library ) is
            crsr : Maps.Cursor;
            elem : Cache_Element;
        begin
            if lib /= null then
                crsr := map.Find( To_String( lib.name ) );
                if Maps.Has_Element( crsr ) then
                    elem := Maps.Element( crsr );
                    elem.refs := elem.refs - 1;
                    --if elem.refs <= 0 then
                    --    Delete( elem.lib );
                    --    map.Delete( crsr );
                    --else
                        map.Replace_Element( crsr, elem );
                    --end if;
                end if;
                lib := null;
            end if;
        end Unload_Library;

    end Cache;

    --==========================================================================

    procedure Initialize renames Loading.Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize renames Loading.Finalize;

    ----------------------------------------------------------------------------

    function Create_Tile_Library return A_Tile_Library is
        this : constant A_Tile_Library := new Tile_Library;
    begin
        this.Construct;
        return this;
    end Create_Tile_Library;

    ----------------------------------------------------------------------------

    function Load_Library( name : String ) return A_Tile_Library is
        lib : A_Tile_Library;
    begin
        Cache.Load_Library( name, lib );
        return lib;
    end Load_Library;

    ----------------------------------------------------------------------------

    function Load_Library_Sync( name : String ) return A_Tile_Library is
        lib : A_Tile_Library;
    begin
        Cache.Load_Library( name, lib );
        if lib /= null then
            lib.Wait_For_Load;
        end if;
        return lib;
    end Load_Library_Sync;

    ----------------------------------------------------------------------------

    procedure Unload_Library( name : String ) is
    begin
        Cache.Unload_Library( name );
    end Unload_Library;

    ----------------------------------------------------------------------------

    procedure Unload_Library( lib : in out A_Tile_Library ) is
    begin
        Cache.Unload_Library( lib );
    end Unload_Library;

    --==========================================================================

    function Create_Library( name : String;
                             dir  : String ) return A_Tile_Library is
        this : constant A_Tile_Library := Create_Tile_Library;
    begin
        this.name := To_Unbounded_String( name );
        this.dir := To_Unbounded_String( dir );
        this.index := Create_Tile_Index;
        return this;
    end Create_Library;

    ----------------------------------------------------------------------------

    procedure Add_Matrix( this   : not null access Tile_Library'Class;
                          matrix : in out A_Tile_Matrix ) is
    begin
        if matrix /= null then
            Add_Matrix( this.index, matrix );
            matrix := null;
        end if;
    end Add_Matrix;

    ----------------------------------------------------------------------------

    procedure Add_Tile( this : not null access Tile_Library'Class;
                        tile : in out A_Tile ) is
    begin
        if tile /= null then
            Add_Tile( this.index, tile );
            tile := null;
        end if;
    end Add_Tile;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Tile_Library ) is
    begin
        Limited_Object(this.all).Construct;
        this.loadState := new Async_Operation;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Tile_Library ) is
    begin
        Delete( this.loadState );
        Delete( this.index );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Exists( this : not null access Tile_Library'Class;
                     id   : Natural ) return Boolean is
    begin
        if id > 0 then
            return Get_Tile( this.index, id ) /= null;
        end if;
        return False;
    end Exists;

    ----------------------------------------------------------------------------

    function Get_Bitmap( this : not null access Tile_Library'Class;
                         id   : Natural ) return A_Bitmap is
        tile : A_Tile;
    begin
        if id > 0 then
            tile := Get_Tile( this.index, id );
            if tile /= null then
                if tile.Is_Loaded then
                    return tile.Get_Bitmap;
                else
                    Prioritize_Tile( this.index, tile );
                end if;
            end if;
        end if;
        return null;
    end Get_Bitmap;

    ----------------------------------------------------------------------------

    function Get_Bitmap_Sync( this : not null access Tile_Library'Class;
                              id   : Natural ) return A_Bitmap is
        tile : A_Tile;
    begin
        if id > 0 then
            tile := Get_Tile( this.index, id );
            if tile /= null then
                if not tile.Is_Loaded then
                    Prioritize_Tile( this.index, tile );
                    tile.Wait_For_Load;
                end if;
                return tile.Get_Bitmap;
            end if;
        end if;
        return null;
    end Get_Bitmap_Sync;

    ----------------------------------------------------------------------------

    function Get_Bitmap_At_Slot( this : not null access Tile_Library'Class;
                                 slot : Natural ) return A_Bitmap is
        tile : A_Tile;
    begin
        if slot > 0 then
            tile := Get_Tile_At_Slot( this.index, slot );
            if tile /= null then
                if tile.Is_Loaded then
                    return tile.Get_Bitmap;
                else
                    Prioritize_Tile( this.index, tile );
                end if;
            end if;
        end if;
        return null;
    end Get_Bitmap_At_Slot;

    ----------------------------------------------------------------------------

    function Get_Clip( this : not null access Tile_Library'Class;
                       id   : Natural ) return Clip_Type is
        tile : A_Tile;
    begin
        if id > 0 then
            tile := Get_Tile( this.index, id );
            if tile /= null then
                return tile.Get_Clipping;
            end if;
        end if;
        return Passive;
    end Get_Clip;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Tile_Library'Class;
                     name : String ) return Natural is
        tile : constant A_Tile := Get_Tile( this.index, name );
    begin
        if tile /= null then
            return tile.Get_Id;
        end if;
        return 0;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Id_At_Slot( this : not null access Tile_Library'Class;
                             slot : Natural ) return Natural is
        tile : A_Tile;
    begin
        if slot > 0 then
            tile := Get_Tile_At_Slot( this.index, slot );
            if tile /= null then
                return tile.Get_Id;
            end if;
        end if;
        return 0;
    end Get_Id_At_Slot;

    ----------------------------------------------------------------------------

    function Get_Matrix( this : not null access Tile_Library'Class;
                         num  : Natural ) return A_Tile_Matrix is
    begin
        return Get_Matrix( this.index, num );
    end Get_Matrix;

    ----------------------------------------------------------------------------

    function Get_Matrix_Count( this : not null access Tile_Library'Class ) return Natural is
    begin
        return Get_Matrix_Count( this.index );
    end Get_Matrix_Count;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Tile_Library'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Get_Progress( this : not null access Tile_Library'Class ) return Natural is
    begin
        return Get_Progress( this.index );
    end Get_Progress;

    ----------------------------------------------------------------------------

    function Get_Slot_For_ID( this : not null access Tile_Library'Class;
                              id   : Natural ) return Natural is
    begin
        if this.Exists( id ) then
            return Get_Slot( this.index, id );
        end if;
        return 0;
    end Get_Slot_For_ID;

    ----------------------------------------------------------------------------

    function Get_Tile( this : not null access Tile_Library'Class;
                       id   : Natural ) return A_Tile is
    begin
        if id > 0 then
            return Get_Tile( this.index, id );
        end if;
        return null;
    end Get_Tile;

    ----------------------------------------------------------------------------

    function Get_Tile_Count( this : not null access Tile_Library'Class ) return Natural is
    begin
        return Get_Tile_Count( this.index );
    end Get_Tile_Count;

    ----------------------------------------------------------------------------

    function Is_Loaded( this : not null access Tile_Library'Class ) return Boolean is
    begin
        return this.loadState.Is_Complete;
    end Is_Loaded;

    ----------------------------------------------------------------------------

    procedure Wait_For_Load( this : not null access Tile_Library'Class ) is
    begin
        this.loadState.Wait;
    end Wait_For_Load;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Tile_Library ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    function Library_Extension return String is
    begin
        return "gfx";
    end Library_Extension;

end Tiles.Libraries;
