
with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Tiles.Libraries;                   use Tiles.Libraries;

package body Tiles is

    allocate    : Allocator := null;
    version     : Natural := 0;
    identifier  : Unbounded_String;

    initialized : Boolean := False;
    finalized   : Boolean := False;

    ----------------------------------------------------------------------------

    -- Raises ATTRIBUTE_ERROR on error.
    procedure Parse_Id_List( str : String; list : in out A_Tile_Id_Array ) is

        ------------------------------------------------------------------------

        function Get_List_Length( str : String ) return Natural is
            idx : Integer := str'First;

            --------------------------------------------------------------------

            procedure Expect( c : Character ) is
            begin
                if idx > str'Last or else str(idx) /= c then
                    raise ATTRIBUTE_ERROR with "Expected '" & c & "'";
                end if;
                idx := idx + 1;
            end Expect;

            --------------------------------------------------------------------

            id_count    : Natural := 0;  -- number of ids seen
            digit_count : Natural := 0;  -- number of digits seen in an id
        begin
            list := null;
            Expect( '(' );
            while idx <= str'Last and then str(idx) /= ')' loop
                if str(idx) /= ',' and then (str(idx) < '0' or else str(idx) > '9') then
                    -- found an invalid character in the list
                    -- (only digits and commas are allowed)
                    raise ATTRIBUTE_ERROR with "Invalid character '" & str(idx) & "'";
                elsif str(idx) = ',' then
                    if digit_count = 0 then
                        -- found a comma following the opening parenthesis or
                        -- another comma. we expected to find an id.
                        raise ATTRIBUTE_ERROR with "Expected tile id";
                    end if;
                    digit_count := 0;
                    id_count := id_count + 1;
                else
                    --
                    digit_count := digit_count + 1;
                    if digit_count > 9 then
                        -- way too many digits in this tile id number
                        raise ATTRIBUTE_ERROR with "Tile id too large";
                    end if;
                end if;
                idx := idx + 1;
            end loop;
            if digit_count = 0 then
                raise ATTRIBUTE_ERROR with "Expected tile id";
            else
                -- count the id number we ended on
                id_count := id_count + 1;
            end if;
            Expect( ')' );
            if idx <= str'Last then
                raise ATTRIBUTE_ERROR with "Unexpected characters after list";
            end if;
            return id_count;
        end Get_List_Length;

        ------------------------------------------------------------------------

        len       : Natural;
        elemStart : Integer := str'First + 1;
        curElem   : Positive := 1;
    begin
        Delete( list );
        len := Get_List_Length( str );

        -- we can now be sure the value is a properly formatted tile id list
        -- containing at least one tile id.

        list := new Tile_Id_Array(1..len);
        for i in elemStart..str'Last-1 loop
            if str(i) = ',' then
                -- record next tile id
                list(curElem) := Natural'Max( 0, Integer'Value( str(elemStart..i-1) ) );
                elemStart := i + 1;
                curElem := curElem + 1;
            end if;
        end loop;
        -- record last tile id
        list(curElem) := Natural'Max( 0, Integer'Value( str(elemStart..str'Last-1) ) );
    end Parse_Id_List;

    ----------------------------------------------------------------------------

    procedure Parse_Natural( str     : String;
                             val     : in out Natural;
                             success : out Boolean ) is
        tmp : Integer;
    begin
        success := False;
        tmp := Integer'Value( str );
        if tmp'Valid and then tmp >= 0 then
            val := tmp;
            success := True;
        end if;
    end Parse_Natural;

    --==========================================================================

    procedure Initialize is
    begin
        if not initialized then
            pragma Debug( Dbg( "Initializing tile system", D_INIT, INFO ) );
            Libraries.Initialize;
            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is
    begin
        if initialized and then not finalized then
            pragma Debug( Dbg( "Finalizing tile system", D_INIT, INFO ) );
            Libraries.Finalize;
            finalized := True;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Create_Tile return A_Tile is
    begin
        if allocate /= null then
            return allocate.all;
        else
            raise Constraint_Error with "Tile allocator not registered";
        end if;
    end Create_Tile;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Tile_Object ) is
        pragma Unreferenced( this );
    begin
        raise COPY_NOT_ALLOWED;
    end Adjust;
    pragma No_Return( Adjust );

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Tile_Object ) is
    begin
        Object(this.all).Construct;
        this.lock := new Locking_Object;
        this.loadState := new Async_Operation;
        this.clipping := Passive;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Tile_Object ) is
    begin
        this.lock.Lock;
        Destroy_Bitmap( this.bmp );
        Delete( this.loadState );
        Delete( this.lock );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Anm_Delay( this : not null access Tile_Object'Class ) return Time_Span is
    begin
        return this.anmDelay;
    end Get_Anm_Delay;

    ----------------------------------------------------------------------------

    function Get_Bitmap( this : not null access Tile_Object'Class ) return A_Bitmap is
        bmp : A_Bitmap := null;
    begin
        this.lock.Lock;
        if this.loadState.Is_Complete then
            bmp := this.bmp;
        end if;
        this.lock.Unlock;
        return bmp;
    end Get_Bitmap;

    ----------------------------------------------------------------------------

    function Get_Clipping( this : not null access Tile_Object'Class ) return Clip_Type is
    begin
        return this.clipping;
    end Get_Clipping;

    ----------------------------------------------------------------------------

    function Get_Frame_List( this : not null access Tile_Object'Class ) return A_Tile_Id_Array is
    begin
        return this.frames;
    end Get_Frame_List;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Tile_Object'Class ) return Natural is
    begin
        return this.id;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Tile_Object'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Get_Next_Frame( this : not null access Tile_Object'Class ) return Natural is
    begin
        return this.nextFrame;
    end Get_Next_Frame;

    ----------------------------------------------------------------------------

    function Is_Animated( this : not null access Tile_Object'Class ) return Boolean is
    begin
        return this.anmDelay > Time_Span_Zero;
    end Is_Animated;

    ----------------------------------------------------------------------------

    function Is_Loaded( this : not null access Tile_Object'Class ) return Boolean is
    begin
        return this.loadState.Is_Complete;
    end Is_Loaded;

    ----------------------------------------------------------------------------

    procedure Load_Bitmap( this    : not null access Tile_Object'Class;
                           archive : not null A_Archive ) is
    begin
        this.lock.Lock;
        if not this.loadState.Is_Complete then
            begin
                this.bmp := archive.Load_Bitmap( To_String( this.name ) );
            exception
                when e : others =>
                    null;
                    pragma Debug( Dbg( "Tiles.Load_Bitmap Exception: ...", D_TILES, ERROR ) );
                    pragma Debug( Dbg( e, D_TILES, ERROR ) );
                    pragma Debug( Dbg( "Unable to find <" & To_String( this.name ) &
                                       "> in tile library", D_TILES, ERROR ) );
            end;
            this.loadState.Complete;
        end if;
        this.lock.Unlock;
    end Load_Bitmap;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Attribute( this  : in out Tile_Object;
                             found : out Boolean;
                             name  : String;
                             val   : String := "" ) is
        delayms : Natural := 0;
        success : Boolean;
    begin
        found := True;

        -- clip=<Physics.Clip_Type enum values>
        if name = "clip" then
            To_Clip_Type( val, this.clipping, success );
            if not success then
                raise ATTRIBUTE_ERROR with "Invalid value for 'clip'";
            end if;

        -- delay=<milliseconds>
        elsif name = "delay" then
            Parse_Natural( val, delayms, success );
            if not success then
                raise ATTRIBUTE_ERROR with "Invalid value for 'delay'";
            end if;
            this.anmDelay := Milliseconds( delayms );

        -- frames=(<tile_id>[,<tile_id>]*)
        elsif name = "frames" then
            if this.nextFrame > 0 then
                raise ATTRIBUTE_ERROR with "Cannot specify 'frames' after 'next'";
            end if;
            begin
                Parse_Id_List( val, this.frames );
            exception
                when e : ATTRIBUTE_ERROR =>
                    raise ATTRIBUTE_ERROR with "Invalid value for 'frames': " &
                                               Exception_Message( e );
            end;

        -- next=<tile_id>
        elsif name = "next" then
            if this.frames /= null then
                raise ATTRIBUTE_ERROR with "Cannot specify 'next' after 'frames'";
            end if;
            Parse_Natural( val, this.nextFrame, success );
            if not success then
                raise ATTRIBUTE_ERROR with "Invalid value for 'next'";
            end if;

        -- didn't recognize the attribute name
        else
            found := False;
        end if;
    end Set_Attribute;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Id( this : access Tile_Object; id : Natural ) is
    begin
        this.id := id;
    end Set_Id;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Name( this : access Tile_Object; name : String ) is
    begin
        this.name := To_Unbounded_String( name );
    end Set_Name;

    ----------------------------------------------------------------------------

    procedure Wait_For_Load( this : not null access Tile_Object'Class ) is
    begin
        this.loadState.Wait;
    end Wait_For_Load;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Tile ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Tile_Object ) is
    begin
        Object'Read( stream, Object(obj) );
        obj.id := Integer'Input( stream );
        obj.name := To_Unbounded_String( String'Input( stream ) );
        obj.clipping := Clip_Type'Input( stream );
        obj.anmDelay := Milliseconds( Integer'Input( stream ) );
        obj.nextFrame := Integer'Input( stream );
        obj.frames := A_Tile_Id_Array'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Tile_Object ) is
    begin
        Object'Write( stream, Object(obj) );
        Integer'Output( stream, obj.id );
        String'Output( stream, To_String( obj.name ) );
        Clip_Type'Output( stream, obj.clipping );
        Integer'Output( stream, obj.anmDelay / Milliseconds( 1 ) );
        Integer'Output( stream, obj.nextFrame );
        A_Tile_Id_Array'Output( stream, obj.frames );
    end Object_Write;

    ----------------------------------------------------------------------------

    function A_Tile_Input( stream : access Root_Stream_Type'Class ) return A_Tile is
        tile : A_Tile := null;
    begin
        if Boolean'Input( stream ) then
            A_Tile'Read( stream, tile );
        end if;
        return tile;
    end A_Tile_Input;

    ----------------------------------------------------------------------------

    procedure A_Tile_Output( stream : access Root_Stream_Type'Class; tile : A_Tile ) is
    begin
        Boolean'Output( stream, tile /= null );
        if tile /= null then
            A_Tile'Write( stream, tile );
        end if;
    end A_Tile_Output;

    ----------------------------------------------------------------------------

    procedure A_Tile_Read( stream : access Root_Stream_Type'Class; tile : out A_Tile ) is
    begin
        tile := new Tile_Object'Class'(Tile_Object'Class'Input( stream ));
    end A_Tile_Read;

    ----------------------------------------------------------------------------

    procedure A_Tile_Write( stream : access Root_Stream_Type'Class; tile : A_Tile ) is
    begin
        Tile_Object'Class'Output( stream, tile.all );
    end A_Tile_Write;

    --==========================================================================

    function A_Tile_Id_Array_Input( stream : access Root_Stream_Type'Class ) return A_Tile_Id_Array is
        tia : A_Tile_Id_Array := null;
        len : Natural;
    begin
        if Boolean'Input( stream ) then
            len := Natural'Input( stream );
            tia := new Tile_Id_Array(1..len);
            for i in tia'Range loop
                tia(i) := Natural'Input( stream );
            end loop;
        end if;
        return tia;
    exception
        when others =>
            Delete( tia );
            raise;
    end A_Tile_Id_Array_Input;

    ----------------------------------------------------------------------------

    procedure A_Tile_Id_Array_Output( stream : access Root_Stream_Type'Class; tia : A_Tile_Id_Array ) is
    begin
        Boolean'Output( stream, tia /= null );
        if tia /= null then
            Natural'Output( stream, tia'Length );
            for i in tia'Range loop
                Natural'Output( stream, tia(i) );
            end loop;
        end if;
    end A_Tile_Id_Array_Output;

    ----------------------------------------------------------------------------

    function Copy( src : A_Tile_Id_Array ) return A_Tile_Id_Array is
    begin
        if src /= null then
            return new Tile_Id_Array'(src.all);
        end if;
        return null;
    end Copy;

    --==========================================================================

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Tiles.allocate := allocate;
    end Register_Allocator;

    ----------------------------------------------------------------------------

    procedure Register_Identifier( identifier : String ) is
    begin
        Tiles.identifier := To_Unbounded_String( identifier );
    end Register_Identifier;

    ----------------------------------------------------------------------------

    procedure Register_Version( version : Positive ) is
    begin
        Tiles.version := version;
    end Register_Version;

    ----------------------------------------------------------------------------

    function Tile_Identifier return String is
    begin
        return To_String( identifier );
    end Tile_Identifier;

    ----------------------------------------------------------------------------

    function Tile_Version return Natural is
    begin
        return version;
    end Tile_Version;

end Tiles;
