
with Allegro;
with Allegro.Sprites;                   use Allegro.Sprites;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events;                            use Events;
with Game_Views;                        use Game_Views;
with Widgets.Containers;                use Widgets.Containers;

package body Widgets.Containers.Windows is

    use Widget_Lists;

    ----------------------------------------------------------------------------

    function Create_Window( view   : access Game_Views.Game_View'Class;
                            id     : String;
                            width,
                            height : Natural;
                            scale  : Positive := 1;
                            filter : Filter_Type := Filter_Nearest
                          ) return A_Window is
        this : A_Window := new Window;
    begin
        this.Construct( view, id, width, height, scale, filter );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Window;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this   : access Window;
                         view   : access Game_Views.Game_View'Class;
                         id     : String;
                         width,
                         height : Natural;
                         scale  : Positive;
                         filter : Filter_Type ) is
    begin
        Container(this.all).Construct( view, id );
        this.cwidth := width;
        this.cheight := height;
        this.scale := scale;
        this.filter := filter;
        this.x1 := 0;
        this.y1 := 0;
        this.x2 := this.cwidth - 1;
        this.y2 := this.cheight - 1;
        this.focused := True;
        this.focus := A_Widget(this);
        this.modal := A_Widget(this);
        if this.scale > 1 then
            this.bmp := Create_Bitmap( width, height );       -- for drawing contents
            this.scalebmp := Create_Bitmap( width * scale,    -- for scaled contents
                                            height * scale );
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Window ) is
    begin
        while not this.popups.Is_Empty loop
            this.Pop_Popup;
        end loop;
        Destroy_Bitmap( this.scalebmp );
        Delete( A_Limited_Object(this.menu) );
        Container(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Click( this : access Window;
                              evt  : not null A_Mouse_Button_Event ) is
        btn  : constant Mouse_Button := evt.Get_Button;
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
        cx,
        cy   : Integer;
    begin
        -- only send the click to the widget if a press and release occurred
        -- over the same widget
        if this.mouseClickedOn(btn) /= null then
            this.mouseClickedOn(btn).Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mouseClickedOn(btn).Handle_Click( evt );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Click;

    ----------------------------------------------------------------------------

    not overriding
    function Dispatch_Key_Held( this : access Window;
                                evt  : not null A_Key_Event ) return Boolean is
        key : constant Integer := evt.Get_Key;
    begin
        if this.keyPressSentTo(key) /= null then
            return this.keyPressSentTo(key).Handle_Key_Held( evt );
        end if;
        return False;
    end Dispatch_Key_Held;

    ----------------------------------------------------------------------------

    not overriding
    function Dispatch_Key_Press( this : access Window;
                                 evt  : not null A_Key_Event ) return Boolean is
        key : constant Integer := evt.Get_Key;
    begin
        this.keyPressSentTo(key) := this.focus;
        return this.keyPressSentTo(key).Handle_Key_Press( evt );
    end Dispatch_Key_Press;

    ----------------------------------------------------------------------------

    not overriding
    function Dispatch_Key_Release( this : access Window;
                                   evt  : not null A_Key_Event ) return Boolean is
        key     : constant Integer := evt.Get_Key;
        handled : Boolean := False;
    begin
        if this.keyPressSentTo(key) /= null then
            handled := this.keyPressSentTo(key).Handle_Key_Release( evt );
            this.keyPressSentTo(key) := null;
        end if;
        return handled;
    end Dispatch_Key_Release;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Mouse_Held( this : access Window;
                                   evt  : not null A_Mouse_Button_Event ) is
        btn  : constant Mouse_Button := evt.Get_Button;
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
        cx,
        cy   : Integer;
    begin
        -- if the mouse button was pressed on a widget, then we send a held
        -- event to that widget, even if the mouse isn't currently over it.
        if this.mousePressedOn(btn) /= null then
            this.mousePressedOn(btn).Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mousePressedOn(btn).Handle_Mouse_Held( evt );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Mouse_Held;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Mouse_Move( this : access Window;
                                   evt  : not null A_Mouse_Event ) is
        evtX    : constant Integer := evt.Get_X;
        evtY    : constant Integer := evt.Get_Y;
        current : A_Widget;
        widgetX,              -- translated location of mouse on the found
        widgetY : Integer;    --   widget, in widget content coordinates
    begin
        this.Find_Widget( evtX, evtY, widgetX, widgetY, current );
        if this.mouseOver /= current then
            -- handle the exit and enter events
            if this.mouseOver /= null then
                this.mouseOver.Handle_Exit;
            end if;
            if current /= null then
                current.Handle_Enter;
            end if;
            this.mouseOver := current;
        end if;

        declare
            target : A_Widget := this.mousePressedOn(Mouse_Left);
            cx, cy : Integer;
        begin
            -- if the left mouse button was pressed on a widget, then we
            -- send a move event to that widget, even if the mouse isn't
            -- currently over it.
            if target /= null and then target.Is_Visible then
                target.Translate_To_Content( evtX, evtY, cx, cy );
                evt.Set_XY( cx, cy );
                target.Handle_Mouse_Move( evt );
                evt.Set_XY( evtX, evtY );
            end if;

            -- if the mouse is over a different wiget that the one we just
            -- sent the mouse move to, then also send a mouse move to the widget
            -- the mouse is currently over.
            target := this.mouseOver;
            if target /= null and then
               target /= this.mousePressedOn(Mouse_Left)
            then
                target.Translate_To_Content( evtX, evtY, cx, cy );
                evt.Set_XY( widgetX, widgetY );
                target.Handle_Mouse_Move( evt );
                evt.Set_XY( evtX, evtY );
            end if;
        end;
    end Dispatch_Mouse_Move;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Mouse_Press( this : access Window;
                                    evt  : not null A_Mouse_Button_Event ) is
        btn  : constant Mouse_Button := evt.Get_Button;
        evtX : constant Integer := evt.Get_X;
        evtY : constant Integer := evt.Get_Y;
        cx,
        cy   : Integer;
    begin
        if this.mouseOver = null then
            this.Find_Widget( evtX, evtY, cx, cy, this.mouseOver );
        end if;
        this.mousePressedOn(btn) := this.mouseOver;

        if this.mousePressedOn(btn) /= null then
            -- iterate through the popup stack, popping each off until we find
            -- a widget that is an ancestor of mousePressedOn(btn).
            while not this.popups.Is_Empty loop
                exit when this.mousePressedOn(btn).Is_Descendant_Of( this.popups.Last_Element );
                this.Pop_Popup;
            end loop;

            -- if the widget the mouse was pressed on is not within the current
            -- modal widget then don't dispatch the press event
            if not this.mousePressedOn(btn).Is_Descendant_Of( this.modal ) then
                this.mousePressedOn(btn) := null;
                this.mouseOver := null;
                return;
            end if;
        end if;

        -- send the press to the widget the mouse is currently over
        if this.mousePressedOn(btn) /= null then
            -- change the focused widget
            this.Give_Focus( this.mousePressedOn(btn) );
            this.mousePressedOn(btn).Translate_To_Content( evtX, evtY, cx, cy );
            evt.Set_XY( cx, cy );
            this.mousePressedOn(btn).Handle_Mouse_Press( evt );
            evt.Set_XY( evtX, evtY );
        end if;
    end Dispatch_Mouse_Press;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Mouse_Release( this : access Window;
                                      evt  : not null A_Mouse_Button_Event ) is
    begin
        A_Window(this).Dispatch_Mouse_Release( evt.Get_X, evt.Get_Y, evt.Get_Button );
    end Dispatch_Mouse_Release;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Mouse_Release( this : access Window;
                                      x, y : Integer;
                                      btn  : Mouse_Button ) is
        evt    : A_Mouse_Button_Event;
        cx, cy : Integer;   -- widget content coordinates
    begin
        -- send the release event to the same widget that received the press
        if this.mousePressedOn(btn) /= null then
            this.mousePressedOn(btn).Translate_To_Content( x, y, cx, cy );
            evt := Create_Mouse_Release( cx, cy, btn );
            this.mousePressedOn(btn).Handle_Mouse_Release( evt );
            Delete( A_Event(evt) );
        end if;

        -- indicate that the last press and release were over the same widget.
        -- this does not mean an actual click occurred because the release
        -- could have delayed too long.
        if this.mousePressedOn(btn) = this.mouseOver then
            this.mouseClickedOn(btn) := this.mouseOver;
        else
            this.mouseClickedOn(btn) := null;
        end if;

        this.mousePressedOn(btn) := null;
    end Dispatch_Mouse_Release;

    ----------------------------------------------------------------------------

    not overriding
    function Dispatch_Mouse_Scroll( this : access Window;
                                    evt  : not null A_Mouse_Scroll_Event ) return Boolean is
    begin
        if this.mouseOver /= null then
            return this.mouseOver.Handle_Mouse_Scroll( evt );
        end if;
        return False;
    end Dispatch_Mouse_Scroll;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw( this : access Window; bmp : not null A_Bitmap ) is
    begin
        if this.scale > 1 then
            this.Draw( this.bmp, 0, 0 );
            Scale( this.bmp, this.scalebmp,
                   Get_Width( this.bmp ), Get_Height( this.bmp ),
                   Float(this.scale),
                   this.filter );
            Blit( this.scalebmp, bmp, 0, 0, this.x1, this.y1,
                  Get_Width( this.scalebmp ), Get_Height( this.scalebmp ) );
        else
            this.Draw( bmp, this.x1, this.y1 );
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw( this : access Window; bmp : not null A_Bitmap; x, y : Integer ) is

        ------------------------------------------------------------------------

        procedure Draw_Child( pos : Widget_Lists.Cursor ) is
            child : constant A_Widget := Element( pos );
        begin
            if child.Is_Visible and then not this.popups.Contains( child ) then
                child.Draw( bmp, x + child.x1, y + child.y1 );
            end if;
        end Draw_Child;

        ------------------------------------------------------------------------

        procedure Draw_Popup( pos : Widget_Lists.Cursor ) is
            popup : constant A_Widget := Element( pos );
        begin
            popup.Draw( bmp, x + popup.x1, y + popup.y1 );
        end Draw_Popup;

        ------------------------------------------------------------------------

    begin
        if this.visible then
            this.children.Iterate( Draw_Child'Access );
            if this.menu /= null then
                A_Widget(this.menu).Draw( bmp, x, y );
            end if;
            this.popups.Iterate( Draw_Popup'Access );
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Window; dc : Drawing_Context ) is
        pragma Unreferenced( this, dc );
    begin
        -- windows don't draw anything
        null;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    procedure Find_Widget( this   : access Window;
                           sx, sy : Integer;
                           wx, wy : out Integer;
                           found  : out A_Widget ) is
        winX : constant Integer := (sx - this.x1) / this.scale;
        winY : constant Integer := (sy - this.y1) / this.scale;
    begin
        found := null;
        if this.menu /= null then
            A_Widget(this.menu).Find_Widget( winX + A_Widget(this.menu).vx1,
                                             winY + A_Widget(this.menu).vy1,
                                             wx, wy, found );
        end if;
        if found = null then
            Container(this.all).Find_Widget( winX, winY, wx, wy, found );
        end if;
        if found /= null and then not found.Is_Descendant_Of( this.modal ) then
            -- if we have a modal widget and the widget we found is not a
            -- descendant of it, then forget that we found anything.
            found := null;
        end if;
    end Find_Widget;

    ----------------------------------------------------------------------------

    function Get_Scale( this : not null access Window'Class ) return Positive is
    begin
        return this.scale;
    end Get_Scale;

    ----------------------------------------------------------------------------

    overriding
    function Get_Window( this : access Window ) return access Window'Class is
    begin
        return A_Window(this);
    end Get_Window;

    ----------------------------------------------------------------------------

    not overriding
    procedure Give_Focus( this : access Window; target : not null A_Widget ) is
        tryFocus : A_Widget := null;
    begin
        if target /= this.focus then
            tryFocus := target;
            loop
                -- check if tryFocus accepts focus. if it can't accept and it
                -- doesn't have a parent (because it's the window) then it will
                -- get focus anyway.
                if tryFocus.Accepts_Focus or else tryFocus.parent = null then
                    -- check if the new focus widget is actually different
                    if this.focus /= tryFocus then
                        -- find any keys pressed on the old focus widget and
                        -- clear their hooks so that future held and release
                        -- events of those keys will not be sent to any widget
                        -- without another key press event.
                        for i in this.keyPressSentTo'Range loop
                            if this.keyPressSentTo(i) = this.focus then
                                this.keyPressSentTo(i) := null;
                            end if;
                        end loop;

                        -- blur the old widget and focus the new widget
                        this.focus.Set_Focused( False );
                        this.focus.Handle_Blur;
                        this.focus := tryFocus;
                        pragma Debug( Dbg( "Focused " & this.focus.To_String, D_GUI, INFO ) );
                        this.focus.Set_Focused( True );
                        this.focus.Handle_Focus;
                    end if;
                    exit;
                end if;

                -- This widget doesn't accept focus, try to give it to its "next"
                -- (if it has one that will accept focus), otherwise try giving
                -- focus to its parent. if we run out of parents, the window
                -- will have to take focus.
                if Length( tryFocus.nextId ) > 0 then
                    declare
                        next : A_Widget;
                    begin
                        next := this.Get_View.Get_Widget( To_String( tryFocus.nextId ) );
                        if next.Accepts_Focus then
                            tryFocus := next;
                        else
                            -- 'next' widget doesn't accept focus
                            tryFocus := tryFocus.parent;
                        end if;
                    exception
                        when others =>
                            -- 'next' widget doesn't exist
                            tryFocus := tryFocus.parent;
                    end;
                else
                    -- no 'next' widget specified
                    tryFocus := tryFocus.parent;
                end if;
            end loop;
        end if;
    end Give_Focus;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Descendant_Hidden( this       : access Window;
                                        descendant : not null A_Widget ) is
    begin
        if this.mouseOver /= null and then
           this.mouseOver.Is_Descendant_Of( descendant )
        then
            -- mouse was over the widget that was hidden, or a descendant
            -- of the widget that was hidden
            this.mouseOver.Handle_Exit;
            this.mouseOver := null;
        end if;
        for i in this.mousePressedOn'Range loop
            if this.mousePressedOn(i) /= null and then
               this.mousePressedOn(i).Is_Descendant_Of( descendant )
            then
                -- a mouse button was pressed on the widget that was hidden, or
                -- on a descendant of the widget that was hidden. fake a mouse
                -- button release event. handling the event will cause
                -- this.mousePressedOn(i) to be set to null.
                this.Dispatch_Mouse_Release( -1, -1, i );
            end if;
        end loop;
        for i in this.mouseClickedOn'Range loop
            if this.mouseClickedOn(i) /= null and then
               this.mouseClickedOn(i).Is_Descendant_Of( descendant )
            then
                -- a mouse button was clicked on the widget that was hidden, or
                -- on a descendant of the widget that was hidden. we don't do
                -- anything special here because widgets don't have any special
                -- state for this case.
                this.mouseClickedOn(i) := null;
            end if;
        end loop;

        if this.modal = descendant then
            this.Set_Modal( null );
        elsif this.focus /= null and then
              this.focus.Is_Descendant_Of( descendant )
        then
            -- the widget that was hidden had focus, or a descendant of it had
            -- focus, so we give focus to the parent of the widget that was
            -- hidden because all widgets between the focused and the descendant
            -- (if any) are being hidden as well.
            this.Give_Focus( descendant.parent );
        end if;
    end Handle_Descendant_Hidden;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Descendant_Unhidden( this       : access Window;
                                          descendant : not null A_Widget ) is
        pragma Unreferenced( descendant );
    begin
        -- clear the mouseOver reference. this will cause us to find it again
        -- when the mouse is moved or a mouse button is pressed.
        this.mouseOver := null;
    end Handle_Descendant_Unhidden;

    ----------------------------------------------------------------------------

    overriding
    function Handle_Key_Press( this : access Window;
                               evt  : not null A_Key_Event ) return Boolean is
    begin
        if evt.Get_Key = KEY_ESC then
            -- if nothing else has handled the ESC key, default behavior of the
            -- window is to pop the top popup widget.
            if not this.popups.Is_Empty then
                this.Pop_Popup;
                return True;
            end if;
        end if;
        return Container(this.all).Handle_Key_Press( evt );
    end Handle_Key_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure Pack( this : access Window ) is
    begin
        Container(this.all).Pack;
        if this.menu /= null then
            A_Widget(this.menu).Pack;
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pop_Popup( this : access Window ) is
        popup : A_Widget;
    begin
        if not this.popups.Is_Empty then
            -- hide then delete the top popup
            popup := this.popups.Last_Element;
            this.popups.Delete_Last;
            popup.Set_Visible( False );
            this.Remove( popup );
            -- todo: decide what to do with the popup
            --Delete( A_Object(popup) );

            -- give focus to the next popup, the modal widget, or the window
            if not this.popups.Is_Empty then
                this.Give_Focus( this.popups.Last_Element );
            elsif this.modal /= null then
                this.Give_Focus( this.modal );
            else
                this.Give_Focus( A_Widget(this) );
            end if;
        end if;
    end Pop_Popup;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pop_Popup( this : access Window; popup : not null A_Widget ) is
    begin
        if this.popups.Contains( popup ) then
            loop
                if this.popups.Last_Element = popup then
                    this.Pop_Popup;
                    exit;
                end if;
                this.Pop_Popup;
            end loop;
        end if;
    end Pop_Popup;

    ----------------------------------------------------------------------------

    not overriding
    procedure Push_Popup( this : access Window; popup : not null A_Widget ) is
        popup2 : A_Widget := popup;
    begin
        popup.Set_Visible( True );
        this.popups.Append( popup );
        A_Window(this).Add( popup2 );

        this.Give_Focus( popup );
    end Push_Popup;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Menubar( this : access Window; menu : in out A_Menubar ) is
    begin
        Delete( A_Limited_Object(this.menu) );
        if menu.all in Menubar'Class then
            this.menu := menu;
            A_Widget(this.menu).Set_Parent( A_Widget(this) );
        end if;
        menu := null;
    end Set_Menubar;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Modal( this : access Window; modal : A_Widget ) is
    begin
        if this.modal /= A_Widget(this) and then this.modal.Is_Visible then
            -- if we already have a modal widget and it's visible then hide it
            -- before setting the new modal.
            this.modal.Set_Visible( False );
        end if;

        -- set the new widget modal, make it visible and give it focus
        if modal /= null then
            this.modal := modal;
            this.modal.Bring_To_Front;
            this.modal.Set_Visible( True );
        else
            this.modal := A_Widget(this);
        end if;
        this.Give_Focus( this.modal );

        -- clear the whole mouse state
        if this.mouseOver /= null then
            this.mouseOver.Handle_Exit;
            this.mouseOver := null;
        end if;
        for i in Mouse_Widget_Array'Range loop
            this.mousePressedOn(i) := null;
            this.mouseClickedOn(i) := null;
        end loop;
    end Set_Modal;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Offset( this : access Window; x, y : Integer ) is
    begin
        this.x2 := x + (this.x2 - this.x1);
        this.y2 := y + (this.y2 - this.y1);
        this.x1 := x;
        this.y1 := y;
    end Set_Offset;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Title( this : access Window; title : String ) is
        pragma Unreferenced( this );
    begin
        Allegro.Set_Window_Title( title );
    end Set_Title;

    ----------------------------------------------------------------------------

    overriding
    procedure Translate_To_Content( this   : access Window;
                                    sx, sy : Integer;
                                    cx, cy : out Integer ) is
    begin
        cx := Integer((sx - this.x1) / this.scale);
        cy := Integer((sy - this.y1) / this.scale);
    end Translate_To_Content;

    ----------------------------------------------------------------------------

    overriding
    procedure Translate_To_Window( this   : access Window;
                                   cx, cy : Integer;
                                   wx, wy : out Integer ) is
        pragma Unreferenced( this );
    begin
        -- no further translation necessary
        wx := cx;
        wy := cy;
    end Translate_To_Window;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Window ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Widgets.Containers.Windows;
