
with Ada.Unchecked_Conversion;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Sprites;                   use Allegro.Sprites;
with Applications;                      use Applications;
--with Debugging;                         use Debugging;
with Game_Views;                        use Game_Views;
with Mouse;                             use Mouse;
with Scaling;                           use Scaling;
with Support;                           use Support;
with Widgets.Containers;                use Widgets.Containers;
with Widgets.Layouts;                   use Widgets.Layouts;

package body Widgets is

    function Accepts_Focus( this : not null access Widget'Class ) return Boolean is
    begin
        return this.acceptFocus and then this.enabled and then this.visible;
    end Accepts_Focus;

    ----------------------------------------------------------------------------

    not overriding
    procedure Activate_Popup( this   : access Widget;
                              popup  : not null A_Widget;
                              corner : Widget_Corner_Type;
                              x, y   : Integer;
                              width  : Natural := 0;
                              height : Natural := 0;
                              scale  : Boolean := False ) is
        pragma Unreferenced( scale );
        winX, winY : Integer;           -- window content coordinates for the corner
        realWidth  : Integer := width;  --
        realHeight : Integer := height;
    begin
        -- determine the location of the specified corner of the popup
        A_Widget(this).Translate_To_Window( x, y, winX, winY );

        -- determine the size of the popup
        if width = 0 then
            realWidth := popup.Get_Min_Width;
        end if;
        if height = 0 then
            realHeight := popup.Get_Min_Height;
        end if;
        -- todo: apply size scaling if scale is true

        -- apply the calculated layout to the popup
        case corner is
            when Widget_Top_Left =>
                popup.Set_Layout_LTWH( winX, winY, realWidth, realHeight );
            when Widget_Top_Right =>
                popup.Set_Layout_LTWH( winX - realWidth, winY, realWidth, realHeight );
            when Widget_Bottom_Left =>
                popup.Set_Layout_LTWH( winX, winY - realHeight, realWidth, realHeight );
            when Widget_Bottom_Right =>
                popup.Set_Layout_LTWH( winX - realWidth, winY - realHeight, realWidth, realHeight );
        end case;

        -- push the popup widget to the window
        this.Get_Window.Push_Popup( popup );
    end Activate_Popup;
    ----------------------------------------------------------------------------

    procedure Add_Listener( this      : not null access Widget'Class;
                            key       : Integer;
                            modifiers : Modifiers_Pattern;
                            handler   : not null A_Key_Handler ) is
        listener : constant A_Key_Listener := Create_Listener( key, modifiers, handler );
    begin
        this.Add_Listener( listener );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this    : not null access Widget'Class;
                            key     : Integer;
                            handler : not null A_Key_Handler ) is
    begin
        this.Add_Listener( key, MODIFIERS_ANY, handler );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Widget'Class;
                            listener : not null A_Key_Listener ) is
    begin
        this.Add_Listener( "Key_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Widget'Class;
                            listener : not null A_Resize_Listener ) is
    begin
        this.Add_Listener( "Resize_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this     : not null access Widget'Class;
                            listener : not null A_Visibility_Listener ) is
    begin
        this.Add_Listener( "Visibility_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    procedure Add_Listener( this         : not null access Widget'Class;
                            listenerType : String;
                            listener     : not null A_Action_Listener ) is

        --------------------------------------------------------------------

        procedure Update( key     : String;
                          element : in out Action_Listeners.List ) is
            pragma Unreferenced( key );
        begin
            -- Verify the listener hasn't already been added for this type
            if not element.Contains( listener ) then
                element.Append( listener );
            end if;
        end Update;

        --------------------------------------------------------------------

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        if Listener_Maps.Has_Element( crsr ) then
            this.listeners.Update_Element( crsr, Update'Access );
        else
            -- Insert a new listener list of the given type with the listener
            -- as the first element.
            declare
                list : Action_Listeners.List;
            begin
                list.Append( listener );
                this.listeners.Insert( listenerType, list );
            end;
        end if;
    end Add_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Bring_To_Front( this : access Widget ) is
    begin
        if this.parent /= null then
            A_Container(this.parent).Bring_To_Front( A_Widget(this) );
        end if;
    end Bring_To_Front;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Widget;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is

        function To_GV2 is new Ada.Unchecked_Conversion( A_Game_View, A_Game_View2 );

        theme : constant A_Theme := Get_Application.Get_Theme;
    begin
        Limited_Object(this.all).Construct;
        this.view := To_GV2(A_Game_View(view));
        this.id := To_Unbounded_String( id );
        this.attrs := Create_Association;
        this.layout := Create_Layout_LTRB( 0, 0, 0, 0 );
        for c in this.colors'Range loop
            this.colors(c) := theme.Color( c );
        end loop;
        this.visible := True;
        view.Register( A_Widget(this) );   -- raises exception on duplicate id
        this.registered := True;
    end Construct;

    ----------------------------------------------------------------------------

    procedure Deactivate_Popup( this : access Widget; popup : not null A_Widget ) is
    begin
        this.Get_Window.Pop_Popup( popup );
    end Deactivate_Popup;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Widget ) is
    begin
        Destroy_Bitmap( this.bmp );
        Destroy_Bitmap( this.drawbmp );

        -- delete the simple key listeners because this widget owns them. other
        -- listener classes are not owned by this widget and should not be
        -- deleted.
        declare

            --------------------------------------------------------------------

            procedure Delete_Listeners( key  : String;
                                        elem : in out Action_Listeners.List ) is
                pragma Unreferenced( key );

                ----------------------------------------------------------------

                procedure Delete_Listener( elem : in out A_Action_Listener ) is
                begin
                    -- if the listener is a simple listener then delete it. this
                    -- will leave a null in the list but that should be ok.
                    if elem.all in Simple_Action_Listener'Class then
                        Delete( A_Object(elem) );
                    end if;
                end Delete_Listener;

                ----------------------------------------------------------------

                crsr : Action_Listeners.Cursor := elem.First;
            begin
                -- loop through each listener in the list
                while Has_Element( crsr ) loop
                    if Element( crsr ) /= null then
                        elem.Update_Element( crsr, Delete_Listener'Access );
                    end if;
                    crsr := Next( crsr );
                end loop;
            end Delete_Listeners;

            --------------------------------------------------------------------

            pos : Listener_Maps.Cursor := this.listeners.First;
        begin
            --Debug( "Deleting Simple_Action_Listeners of widget '" & To_String( this.id ) & "'" );
            -- loop through all listener types
            while Has_Element( pos ) loop
                this.listeners.Update_Element( pos, Delete_Listeners'Access );
                pos := Next( pos );
            end loop;
        end;

        if this.registered then
            this.Get_View.Unregister( this.Get_Id );
        end if;
        Delete( this.layout );
        Delete( this.attrs );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Dispatch_Key_Action( this      : not null access Widget'Class;
                                  id        : Action_Id;
                                  key       : Integer;
                                  modifiers : Modifiers_Array ) return Boolean is
        action  : A_Key_Action;
        handled : Boolean := False;

        ------------------------------------------------------------------------

        procedure Examine( listener : A_Action_Listener ) is
            hndld : Boolean;
        begin
            if listener.all in Key_Listener'Class then
                A_Key_Listener(listener).Handle_Action( action, hndld );
                if hndld then
                    handled := True;
                end if;
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        action := new Key_Action;
        action.Construct( id, A_Widget(this), key, modifiers );
        this.Iterate_Listeners( "Key_Listener", Examine'Access );
        Delete( action );
        return handled;
    end Dispatch_Key_Action;

    ----------------------------------------------------------------------------

    procedure Dispatch_Resize_Action( this : not null access Widget'Class ) is
        action  : A_Resize_Action;

        ------------------------------------------------------------------------

        procedure Examine( listener : A_Action_Listener ) is
        begin
            if listener.all in Resize_Listener'Class then
                A_Resize_Listener(listener).Handle_Action( action );
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        action := new Resize_Action;
        action.Construct( Widget_Resized, A_Widget(this) );
        this.Iterate_Listeners( "Resize_Listener", Examine'Access );
        Delete( action );
    end Dispatch_Resize_Action;

    ----------------------------------------------------------------------------

    procedure Dispatch_Visibility_Action( this    : not null access Widget'Class;
                                          visible : Boolean ) is
        action  : A_Visibility_Action;

        ------------------------------------------------------------------------

        procedure Examine( listener : A_Action_Listener ) is
        begin
            if listener.all in Visibility_Listener'Class then
                A_Visibility_Listener(listener).Handle_Action( action );
            end if;
        end Examine;

        ------------------------------------------------------------------------

    begin
        action := new Visibility_Action;
        action.Construct( Change_Visibility, A_Widget(this), visible );
        this.Iterate_Listeners( "Visibility_Listener", Examine'Access );
        Delete( action );
    end Dispatch_Visibility_Action;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw( this : access Widget; bmp : not null A_Bitmap; x, y : Integer ) is
    begin
        if this.visible and then this.bmp /= null then
            if this.dirty then
                this.dirty := False;
                if this.drawbmp /= null then
                    if this.transparent then
                        Clear_To_Transparent( this.drawbmp );
                    end if;
                    A_Widget(this).Draw_Content( Create_Drawing_Context( this.drawbmp, this.vx1, this.vy1, Background ) );
                    Scale( this.drawbmp, this.bmp,
                           Get_Width( this.drawbmp ), Get_Height( this.drawbmp ),
                           this.zoom,
                           Get_Application.Get_Theme.Get_Scaling_Filter );
                else
                    if this.transparent then
                        Clear_To_Transparent( this.bmp );
                    end if;
                    A_Widget(this).Draw_Content( Create_Drawing_Context( this.bmp, this.vx1, this.vy1, Background ) );
                end if;
            end if;
            if this.transparent then
                Draw_Sprite( bmp, this.bmp, x, y );
            else
                Blit( this.bmp, bmp, 0, 0, x, y, this.x2 - this.x1 + 1, this.y2 - this.y1 + 1 );
            end if;
        end if;
    end Draw;

    ----------------------------------------------------------------------------

    not overriding
    procedure Find_Widget( this   : access Widget;
                           x, y   : Integer;
                           wx, wy : out Integer;
                           found  : out A_Widget ) is
    begin
        -- verify x, y is within the viewport
        if this.visible and then this.enabled and then
           x >= 0 and then x <= this.x2 - this.x1 and then
           y >= 0 and then y <= this.y2 - this.y1
        then
            -- translate the viewport coordinates into content coordinates
            wx := this.vx1 + Integer(x / this.zoom);
            wy := this.vy1 + Integer(y / this.zoom);
            found := A_Widget(this);
        else
           wx := 0;
           wy := 0;
           found := null;
        end if;
    end Find_Widget;

    ----------------------------------------------------------------------------

    function Get_Attribute( this : not null access Widget'Class;
                            name : String ) return Integer is
    begin
        return this.attrs.As_Integer( name );
    end Get_Attribute;

    ----------------------------------------------------------------------------

    function Get_Attribute( this : not null access Widget'Class;
                            name : String ) return String is
    begin
        return this.attrs.As_String( name );
    end Get_Attribute;

    ----------------------------------------------------------------------------

    function Get_Border( this : not null access Widget'Class ) return Border_Type is
    begin
        return this.border;
    end Get_Border;

    ----------------------------------------------------------------------------

    function Get_Color( this : access Widget; purpose : Color_Purpose ) return Color_Type is
    begin
        return this.colors(purpose);
    end Get_Color;

    ----------------------------------------------------------------------------

    function Get_Content_Height( this : not null access Widget'Class ) return Natural is
    begin
        return this.cheight;
    end Get_Content_Height;

    ----------------------------------------------------------------------------

    function Get_Content_Width( this : not null access Widget'Class ) return Natural is
    begin
        return this.cwidth;
    end Get_Content_Width;

    ----------------------------------------------------------------------------

    function Get_Font_Name( this : not null access Widget'Class ) return String is
    begin
        return To_String( this.fontName );
    end Get_Font_Name;

    ----------------------------------------------------------------------------

    function Get_Font_Size( this : not null access Widget'Class ) return Positive is
    begin
        return this.fontSize;
    end Get_Font_Size;

    ----------------------------------------------------------------------------

    function Get_Height( this : not null access Widget'Class ) return Integer is
    begin
        return this.y2 - this.y1 + 1;
    end Get_Height;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Widget'Class ) return String is
    begin
        return To_String( this.id );
    end Get_Id;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Min_Height( this : access Widget ) return Natural is
    begin
        return this.minHeight;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Min_Width( this : access Widget ) return Natural is
    begin
        return this.minWidth;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    not overriding
    function Get_View( this : access Widget ) return access Game_Views.Game_View'Class is
        function To_GV is new Ada.Unchecked_Conversion( A_Game_View2, A_Game_View );
    begin
        return To_GV(this.view);
    end Get_View;

    ----------------------------------------------------------------------------

    procedure Get_Viewport( this   : not null access Widget'Class;
                            x1, y1,
                            x2, y2 : out Integer ) is
    begin
        x1 := this.x1;
        y1 := this.y1;
        x2 := this.x2;
        y2 := this.y2;
    end Get_Viewport;

    ----------------------------------------------------------------------------

    function Get_Width( this : not null access Widget'Class ) return Natural is
    begin
        return this.x2 - this.x1 + 1;
    end Get_Width;

    ----------------------------------------------------------------------------

    function Get_Window( this : access Widget
                       ) return access Widgets.Containers.Windows.Window'Class is
    begin
        if this.parent /= null then
            return this.parent.Get_Window;
        end if;
        return null;
    end Get_Window;

    ----------------------------------------------------------------------------

    function Get_Zoom( this : not null access Widget'Class ) return Float is
    begin
        return this.zoom;
    end Get_Zoom;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Ancestor_Hidden( this : access Widget ) is
    begin
        this.Dispatch_Visibility_Action( visible => False );
    end Handle_Ancestor_Hidden;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Ancestor_Unhidden( this : access Widget ) is
    begin
        this.Dispatch_Visibility_Action( visible => True );
    end Handle_Ancestor_Unhidden;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Blur( this : access Widget ) is
        pragma Unreferenced( this );
    begin
        null;
    end Handle_Blur;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Click( this : access Widget;
                            evt  : not null A_Mouse_Button_Event ) is
        pragma Unreferenced( this, evt );
    begin
        null;
    end Handle_Click;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Descendant_Hidden( this       : access Widget;
                                        descendant : not null A_Widget ) is
    begin
        if this.parent /= null then
            this.parent.Handle_Descendant_Hidden( descendant );
        end if;
    end Handle_Descendant_Hidden;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Descendant_Unhidden( this       : access Widget;
                                          descendant : not null A_Widget ) is
    begin
        if this.parent /= null then
            this.parent.Handle_Descendant_Unhidden( descendant );
        end if;
    end Handle_Descendant_Unhidden;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Enter( this : access Widget ) is
    begin
        this.hover := True;
        A_Widget(this).Set_Dirty;
    end Handle_Enter;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Exit( this : access Widget ) is
    begin
        this.hover := False;
        A_Widget(this).Set_Dirty;
    end Handle_Exit;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Focus( this : access Widget ) is
        pragma Unreferenced( this );
    begin
        null;
    end Handle_Focus;

    ----------------------------------------------------------------------------

    not overriding
    function Handle_Key_Held( this : access Widget;
                              evt  : not null A_Key_Event ) return Boolean is
        modifiers : Modifiers_Array;
    begin
        evt.Get_Modifiers( modifiers );

        -- Try dispatching the key to an action handler
        if this.Dispatch_Key_Action( Key_Held, evt.Get_Key, modifiers ) then
            return True;
        end if;

        -- Check if the parent can handle the event
        if this.parent /= null then
            return this.parent.Handle_Key_Held( evt );
        end if;

        -- Unable to handle the event
        return False;
    end Handle_Key_Held;

    ----------------------------------------------------------------------------

    not overriding
    function Handle_Key_Press( this : access Widget;
                               evt  : not null A_Key_Event ) return Boolean is
        modifiers : Modifiers_Array;
        wdgt      : A_Widget := null;
    begin
        evt.Get_Modifiers( modifiers );

        -- Try dispatching the key to an action handler
        if this.Dispatch_Key_Action( Key_Press, evt.Get_Key, modifiers ) then
            return True;
        end if;

        -- Check if the widget uses the key internally
        if evt.Get_Key = KEY_TAB and then
           not modifiers(ALT) and then
           not modifiers(CTRL)
        then
            begin
                if modifiers(SHIFT) then
                    -- tab to previous widget
                    wdgt := this.Get_View.Get_Widget( To_String( this.prevId ) );
                    while wdgt /= A_Widget(this) loop
                        -- loop through the list of prev widgets until one of
                        -- them accepts focus or we get back to this widget.
                        if wdgt.Accepts_Focus then
                            this.Get_Window.Give_Focus( wdgt );
                            return True;
                        end if;
                        wdgt := this.Get_View.Get_Widget( To_String( wdgt.prevId ) );
                    end loop;
                    this.Get_Window.Give_Focus( wdgt );
                    return True;
                else
                    -- tab to next widget
                    wdgt := this.Get_View.Get_Widget( To_String( this.nextId ) );
                    while wdgt /= A_Widget(this) loop
                        -- loop through the list of next widgets until one of
                        -- them accepts focus or we get back to this widget.
                        if wdgt.Accepts_Focus then
                            this.Get_Window.Give_Focus( wdgt );
                            return True;
                        end if;
                        wdgt := this.Get_View.Get_Widget( To_String( wdgt.nextId ) );
                    end loop;
                end if;
            exception
                when ID_NOT_FOUND => null;
            end;
        end if;

        -- Check if the parent can handle the event
        if this.parent /= null then
            return this.parent.Handle_Key_Press( evt );
        end if;

        -- Unable to handle the event
        return False;
    end Handle_Key_Press;

    ----------------------------------------------------------------------------

    not overriding
    function Handle_Key_Release( this : access Widget;
                                 evt  : not null A_Key_Event ) return Boolean is
        modifiers : Modifiers_Array;
    begin
        evt.Get_Modifiers( modifiers );
        if this.Dispatch_Key_Action( Key_Release, evt.Get_Key, modifiers ) then
            return True;
        elsif this.parent /= null then
            return this.parent.Handle_Key_Release( evt );
        else
            return False;
        end if;
    end Handle_Key_Release;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Mouse_Held( this : access Widget;
                                 evt : not null A_Mouse_Button_Event ) is
        pragma Unreferenced( this, evt );
    begin
        null;
    end Handle_Mouse_Held;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Mouse_Move( this : access Widget;
                                 evt  : not null A_Mouse_Event ) is
        pragma Unreferenced( this, evt );
    begin
        null;
    end Handle_Mouse_Move;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Mouse_Press( this : access Widget;
                                  evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.pressed := True;
            A_Widget(this).Set_Dirty;
        end if;
    end Handle_Mouse_Press;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Mouse_Release( this : access Widget;
                                    evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.pressed := False;
            A_Widget(this).Set_Dirty;
        end if;
    end Handle_Mouse_Release;

    ----------------------------------------------------------------------------

    not overriding
    function Handle_Mouse_Scroll( this : access Widget;
                                  evt  : not null A_Mouse_Scroll_Event ) return Boolean is
    begin
        if this.parent /= null then
            return this.parent.Handle_Mouse_Scroll( evt );
        end if;
        return False;
    end Handle_Mouse_Scroll;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Resize( this : access Widget ) is
        w  : constant Integer := this.x2 - this.x1 + 1;
        h  : constant Integer := this.y2 - this.y1 + 1;
        vw : constant Integer := this.vx2 - this.vx1 + 1;
        vh : constant Integer := this.vy2 - this.vy1 + 1;
    begin
        Destroy_Bitmap( this.bmp );
        Destroy_Bitmap( this.drawbmp );
        if w > 0 and then h > 0 then
            this.bmp := Create_Bitmap( Integer(Float'Ceiling( vw * this.zoom )),
                                       Integer(Float'Ceiling( vh * this.zoom )) );
            if vw > 0 and then vh > 0 then
                this.drawbmp := Create_Bitmap( vw, vh );
            end if;
        end if;
        A_Widget(this).Set_Dirty;
    end Handle_Resize;

    ----------------------------------------------------------------------------

    function Is_Enabled( this : not null access Widget'Class ) return Boolean is
    begin
        return this.enabled and then (this.parent = null or else this.parent.Is_Enabled);
    end Is_Enabled;

    ----------------------------------------------------------------------------

    function Is_Visible( this : not null access Widget'Class ) return Boolean is
    begin
        return this.visible;
    end Is_Visible;

    ----------------------------------------------------------------------------

    function Is_Descendant_Of( this     : not null access Widget'Class;
                               ancestor : A_Widget ) return Boolean is
    begin
        if ancestor = null then
            return False;
        elsif ancestor = A_Widget(this) or else ancestor = this.parent then
            return True;
        elsif this.parent = null then
            return False;
        else
            return this.parent.Is_Descendant_Of( ancestor );
        end if;
    end Is_Descendant_Of;

    ----------------------------------------------------------------------------

    procedure Iterate_Listeners( this         : not null access Widget'Class;
                                 listenerType : String;
                                 examine      : not null access procedure( listener : A_Action_Listener ) ) is

        procedure Local_Examine( pos : Action_Listeners.Cursor ) is
            elem : constant A_Action_Listener := Element( pos );
        begin
            if elem /= null then
                examine.all( elem );
            end if;
        end Local_Examine;

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        if Has_Element( crsr ) then
            Element( crsr ).Iterate( Local_Examine'Access );
        end if;
    end Iterate_Listeners;

    ----------------------------------------------------------------------------

    procedure Move( this  : not null access Widget'Class;
                    xdist,
                    ydist : Integer ) is
    begin
        if this.parent /= null and then (xdist /= 0 or else ydist /= 0) then
            this.x1 := this.x1 + xdist;
            this.x2 := this.x2 + xdist;
            this.y1 := this.y1 + ydist;
            this.y2 := this.y2 + ydist;
        end if;
    end Move;

    ----------------------------------------------------------------------------

    procedure Move_To( this  : not null access Widget'Class; x, y : Integer ) is
    begin
        if this.parent /= null then
            this.x2 := x + (this.x2 - this.x1);
            this.y2 := y + (this.y2 - this.y1);
            this.x1 := x;
            this.y1 := y;
        end if;
    end Move_To;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pack( this : access Widget ) is
        oldCW : constant Integer := this.cwidth;
        oldCH : constant Integer := this.cheight;
        oldVW : constant Integer := this.vx2 - this.vx1 + 1;
        oldVH : constant Integer := this.vy2 - this.vy1 + 1;
        newW,
        newH  : Integer;
    begin
        if this.parent /= null then
            this.x1 := 0;
            this.y1 := 0;
            this.x2 := this.parent.cwidth - 1;
            this.y2 := this.parent.cheight - 1;

            A_Container(this.parent).Apply_Container_Layout( A_Widget(this) );
            this.layout.Apply( A_Widget(this) );

            newW := this.x2 - this.x1 + 1;
            newH := this.y2 - this.y1 + 1;
            this.cwidth := Integer'Max( A_Widget(this).Get_Min_Width, newW );
            this.cheight := Integer'Max( A_Widget(this).Get_Min_Height, newH );

            -- the following lines are commented out because the vx2, vy2
            -- corner of the widget view was being clamped to less than the
            -- vx1, vy1 corner if vx1, vy1 was off the bottom/right edge of the
            -- widget content at pack time. this happens when we change the
            -- contents of a scene widget and the viewport was in the lower right
            -- corner of a much bigger content area than the new map.
--            this.vx2 := Integer'Min( this.vx1 + Floor(newW / this.zoom) - 1, this.cwidth - 1 );
--            this.vy2 := Integer'Min( this.vy1 + Floor(newH / this.zoom) - 1, this.cheight - 1 );
            this.vx2 := this.vx1 + Floor( newW / this.zoom ) - 1;
            this.vy2 := this.vy1 + Floor( newH / this.zoom ) - 1;

            if oldVW /= this.vx2 - this.vx1 + 1 or else
               oldVH /= this.vy2 - this.vy1 + 1
            then
                A_Widget(this).Handle_Resize;
            end if;
            A_Widget(this).Set_Dirty;

            if oldCW /= this.cwidth or else oldCH /= this.cheight then
                this.Dispatch_Resize_Action;
            end if;
        end if;
    end Pack;

    ----------------------------------------------------------------------------

    procedure Prepend_Listener( this         : not null access Widget'Class;
                                listenerType : String;
                                listener     : not null A_Action_Listener ) is

        --------------------------------------------------------------------

        procedure Update( key     : String;
                          element : in out Action_Listeners.List ) is
            pragma Unreferenced( key );
            crsr : Action_Listeners.Cursor := element.Find( listener );
        begin
            -- Prepend the listener to the list. If it's already in the list,
            -- move it to the front.
            if Has_Element( crsr ) then
                element.Delete( crsr );
            end if;
            element.Prepend( listener );
        end Update;

        --------------------------------------------------------------------

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        if Listener_Maps.Has_Element( crsr ) then
            this.listeners.Update_Element( crsr, Update'Access );
        else
            -- Insert a new listener list of the given type with the listener
            -- as the first element.
            declare
                list : Action_Listeners.List;
            begin
                list.Prepend( listener );
                this.listeners.Insert( listenerType, list );
            end;
        end if;
    end Prepend_Listener;

    ----------------------------------------------------------------------------

    procedure Redraw( this : not null access Widget'Class ) is
    begin
        A_Widget(this).Set_Dirty;
    end Redraw;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Widget'Class;
                               listener : not null A_Key_Listener ) is
    begin
        this.Remove_Listener( "Key_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Widget'Class;
                               listener : not null A_Resize_Listener ) is
    begin
        this.Remove_Listener( "Resize_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this     : not null access Widget'Class;
                               listener : not null A_Visibility_Listener ) is
    begin
        this.Remove_Listener( "Visibility_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Remove_Listener( this         : not null access Widget'Class;
                               listenerType : String;
                               listener     : not null A_Action_Listener ) is

        procedure Remove_Listener( key  : String;
                                   list : in out Action_Listeners.List ) is
            pragma Unreferenced( key );
            crsr : Action_Listeners.Cursor := list.Find( listener );
        begin
            -- remove the specified listener
            if Action_Listeners.Has_Element( crsr ) then
                list.Delete( crsr );
            end if;
        end Remove_Listener;

        crsr : constant Listener_Maps.Cursor := this.listeners.Find( listenerType );
    begin
        -- check if we have any listeners registered as the given type
        if Listener_Maps.Has_Element( crsr ) then
            this.listeners.Update_Element( crsr, Remove_Listener'Access );
        end if;
    end Remove_Listener;

    ----------------------------------------------------------------------------

    procedure Scroll( this : not null access Widget'Class; x, y : Integer ) is
    begin
        if y /= 0 and then this.cheight > this.y2 - this.y1 + 1 then
            this.vy1 := this.vy1 + y;
            this.vy2 := this.vy2 + y;
            if this.vy1 < 0 then
                -- bound the viewport at the top of the content
                this.vy2 := this.vy2 + (0 - this.vy1);
                this.vy1 := 0;
            elsif this.vy2 >= this.cheight then
                -- bound the viewport at the bottom of the content
                this.vy1 := this.vy1 - (this.vy2 - (this.cheight - 1));
                this.vy2 := this.cheight - 1;
            end if;
            A_Widget(this).Set_Dirty;
        end if;
        if x /= 0 and then this.cwidth > this.x2 - this.x1 + 1 then
            this.vx1 := this.vx1 + x;
            this.vx2 := this.vx2 + x;
            if this.vx1 < 0 then
                -- bound the viewport at the left of the content
                this.vx2 := this.vx2 + (0 - this.vx1);
                this.vx1 := 0;
            elsif this.vx2 >= this.cwidth then
                -- bound the viewport at the right of the content
                this.vx1 := this.vx1 - (this.vx2 - (this.cwidth - 1));
                this.vx2 := this.cwidth - 1;
            end if;
            A_Widget(this).Set_Dirty;
        end if;
    end Scroll;

    ----------------------------------------------------------------------------

    procedure Scroll_To( this : not null access Widget'Class; x, y : Integer ) is
        vwidth  : Integer;
        vheight : Integer;
    begin
        if x /= this.vx1 then
            vwidth := this.vx2 - this.vx1 + 1;
            this.vx1 := x;
            if this.vx1 + vwidth >= this.cwidth then
                this.vx1 := this.cwidth - vwidth;
            end if;
            if this.vx1 < 0 then
                this.vx1 := 0;
            end if;
            this.vx2 := this.vx1 + vwidth - 1;
            A_Widget(this).Set_Dirty;
        end if;
        if y /= this.vy1 then
            vheight := this.vy2 - this.vy1 + 1;
            this.vy1 := y;
            if this.vy1 + vheight >= this.cheight then
                this.vy1 := this.cheight - vheight;
            end if;
            if this.vy1 < 0 then
                this.vy1 := 0;
            end if;
            this.vy2 := this.vy1 + vheight - 1;
            A_Widget(this).Set_Dirty;
        end if;
    end Scroll_To;

    ----------------------------------------------------------------------------

    procedure Set_Attribute( this : not null access Widget'Class;
                             name : String;
                             val  : Integer ) is
    begin
        this.attrs.Set_Value( name, val );
    end Set_Attribute;

    ----------------------------------------------------------------------------

    procedure Set_Attribute( this : not null access Widget'Class;
                             name : String;
                             val  : String ) is
    begin
        this.attrs.Set_Value( name, val );
    end Set_Attribute;

    ----------------------------------------------------------------------------

    procedure Set_Attribute( this  : not null access Widget'Class;
                             name  : String;
                             val   : in out A_Value ) is
    begin
        this.attrs.Set_Value( name, val );
    end Set_Attribute;

    ----------------------------------------------------------------------------

    procedure Set_Border( this   : not null access Widget'Class;
                          border : Border_Type ) is
    begin
        if this.border /= border then
            this.border := border;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Border;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Color( this    : access Widget;
                         purpose : Color_Purpose;
                         color   : Color_Type ) is
    begin
        if this.colors(purpose) /= color then
            this.colors(purpose) := color;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Color;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Dirty( this : access Widget ) is
    begin
        this.dirty := True;
    end Set_Dirty;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Enabled( this : access Widget; enabled : Boolean ) is
    begin
        if enabled = not this.enabled then
            this.enabled := enabled;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Enabled;

    ----------------------------------------------------------------------------

    procedure Set_Focusable( this      : not null access Widget'Class;
                             focusable : Boolean ) is
    begin
        this.acceptFocus := focusable;
    end Set_Focusable;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Focused( this : access Widget; focused : Boolean ) is
    begin
        if this.focused /= focused then
            this.focused := focused;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Focused;

    ----------------------------------------------------------------------------

    procedure Set_Font( this     : not null access Widget'Class;
                        name : String := "";
                        size : Natural := 0 ) is
    begin
        if name'Length > 0 then
            this.fontName := To_Unbounded_String( name );
            this.Set_Dirty;
        end if;
        if size > 0 then
            this.fontSize := size;
            this.Set_Dirty;
        end if;
    end Set_Font;

    ----------------------------------------------------------------------------

    procedure Set_Layout( this   : not null access Widget'Class;
                          layout : in out A_Layout ) is
    begin
        Delete( this.layout );
        this.layout := layout;
        A_Widget(this).Set_Dirty;
        if this.parent /= null then
            A_Widget(this).Pack;
        end if;
        layout := null;
    end Set_Layout;

    ----------------------------------------------------------------------------

    procedure Set_Layout_Center( this   : not null access Widget'Class;
                                 width,
                                 height : Natural ) is
        layout : A_Layout := Create_Layout_Center( width, height );
    begin
        this.Set_Layout( layout );
    end Set_Layout_Center;

    ----------------------------------------------------------------------------

    procedure Set_Layout_CenterH( this   : not null access Widget'Class;
                                  width  : Natural;
                                  top,
                                  bottom : Integer ) is
        layout : A_Layout := Create_Layout_CenterH( width, top, bottom );
    begin
        this.Set_Layout( layout );
    end Set_Layout_CenterH;

    ----------------------------------------------------------------------------

    procedure Set_Layout_CenterHY( this   : not null access Widget'Class;
                                   width  : Natural;
                                   y1, y2 : Integer ) is
        layout : A_Layout := Create_Layout_CenterHY( width, y1, y2 );
    begin
        this.Set_Layout( layout );
    end Set_Layout_CenterHY;

    ----------------------------------------------------------------------------

    procedure Set_Layout_LTRB( this   : not null access Widget'Class;
                               left,
                               top,
                               right,
                               bottom : Integer ) is
        layout : A_Layout := Create_Layout_LTRB( left, top, right, bottom );
    begin
        this.Set_Layout( layout );
    end Set_Layout_LTRB;

    ----------------------------------------------------------------------------

    procedure Set_Layout_LTWH( this   : not null access Widget'Class;
                               left,
                               top    : Integer;
                               width,
                               height : Natural ) is
        layout : A_Layout := Create_Layout_LTWH( left, top, width, height );
    begin
        this.Set_Layout( layout );
    end Set_Layout_LTWH;

    ----------------------------------------------------------------------------

    procedure Set_Layout_XYWH( this   : not null access Widget'Class;
                               x, y   : Integer;
                               width,
                               height : Natural ) is
        layout : A_Layout := Create_Layout_XYWH( x, y, width, height );
    begin
        this.Set_Layout( layout );
    end Set_Layout_XYWH;

    ----------------------------------------------------------------------------

    procedure Set_Min_Height( this   : not null access Widget'Class;
                              height : Natural ) is
    begin
        this.minHeight := height;
        A_Widget(this).Pack;
    end Set_Min_Height;

    ----------------------------------------------------------------------------

    -- Sets the minimum width value to override the computed minimum width
    -- for the widget. If width is set to zero, the computed value is returned
    -- by Get_Min_Width.
    procedure Set_Min_Width( this  : not null access Widget'Class;
                             width : Natural ) is
    begin
        this.minWidth := width;
        A_Widget(this).Pack;
    end Set_Min_Width;

    ----------------------------------------------------------------------------

    procedure Set_Next( this : not null access Widget'Class; id : String ) is
    begin
        this.nextId := To_Unbounded_String( id );
    end Set_Next;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Parent( this : access Widget; parent : A_Widget ) is
    begin
        -- NOTE: This procedure only sets the new parent of the widget, it does
        --       not break any current parent's association with this widget as
        --       a child!
        if this.parent /= parent then
            this.parent := parent;
            A_Widget(this).Set_Dirty;
            if this.parent /= null then
                A_Widget(this).Pack;
                -- this widget may have become visible because it is visible and
                -- was just added to a parent widget, so we notify the parent
                -- that a descendent has become unhidden.
                if this.visible then
                    this.parent.Handle_Descendant_Unhidden( A_Widget(this) );
                end if;
            end if;
        end if;
    end Set_Parent;

    ----------------------------------------------------------------------------

    procedure Set_Prev( this : not null access Widget'Class; id : String ) is
    begin
        this.prevId := To_Unbounded_String( id );
    end Set_Prev;

    ----------------------------------------------------------------------------

    procedure Set_Transparent( this        : not null access Widget'Class;
                               transparent : Boolean ) is
    begin
        if this.transparent /= transparent then
            this.transparent := transparent;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Transparent;

    ----------------------------------------------------------------------------

    procedure Set_Visible( this : access Widget; visible : Boolean ) is
    begin
        if this.visible /= visible then
            this.visible := visible;
            if this.visible then
                -- notify the ancestors that this widget appeared
                if this.parent /= null then
                    this.parent.Handle_Descendant_Unhidden( A_Widget(this) );
                end if;
                A_Widget(this).Handle_Ancestor_Unhidden;
                A_Widget(this).Set_Dirty;
            else
                -- notify the ancestors that this widget disappeared
                if this.parent /= null then
                    this.parent.Handle_Descendant_Hidden( A_Widget(this) );
                end if;
                A_Widget(this).Handle_Ancestor_Hidden;
            end if;
        end if;
    end Set_Visible;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Zoom( this : access Widget; zoom : Float ) is
        vw, vh : Integer;    -- view width, height
    begin
        if this.zoomable and then this.zoom /= zoom then
            Destroy_Bitmap( this.drawbmp );
            vw := Floor((this.x2 - this.x1 + 1) / zoom);
            vh := Floor((this.y2 - this.y1 + 1) / zoom);

            this.vx2 := Integer'Min( this.vx1 + vw, this.cwidth ) - 1;
            this.vy2 := Integer'Min( this.vy1 + vh, this.cheight ) - 1;
            this.vx1 := this.vx2 - vw + 1;
            this.vy1 := this.vy2 - vh + 1;
            if this.vx1 < 0 then
                this.vx1 := 0;
                this.vx2 := vw - 1;
            end if;
            if this.vy1 < 0 then
                this.vy1 := 0;
                this.vy2 := vh - 1;
            end if;

            if vw > 0 and then vh > 0 then
                this.drawbmp := Create_Bitmap( vw, vh );
            end if;
            this.zoom := zoom;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Zoom;

    ----------------------------------------------------------------------------

    procedure Smooth_Text( this : not null access Widget'Class; enabled : Boolean ) is
    begin
        if enabled /= this.fontSmooth then
            this.fontSmooth := enabled;
            this.Set_Dirty;
        end if;
    end Smooth_Text;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Widget ) return String is
    begin
        return '<' & this.Get_Class_Name & " : id=" & To_String( this.id ) & '>';
    end To_String;

    ----------------------------------------------------------------------------

    not overriding
    procedure Translate_To_Content( this   : access Widget;
                                    sx, sy : Integer;
                                    cx, cy : out Integer ) is
    begin
        if this.parent /= null then
            this.parent.Translate_To_Content( sx, sy, cx, cy );
        else
            cx := sx;
            cy := sy;
        end if;
        cx := Integer((cx - this.x1) / this.zoom) + this.vx1;
        cy := Integer((cy - this.y1) / this.zoom) + this.vy1;
    end Translate_To_Content;

    ----------------------------------------------------------------------------

    procedure Translate_To_Window( this   : access Widget;
                                   cx, cy : Integer;
                                   wx, wy : out Integer ) is
        parentX, parentY : Integer;
    begin
        -- translate the content coordinates of this widget to the content
        -- coordinates of the parent
        wx := Integer(this.zoom * (cx - this.vx1)) + this.x1;
        wy := Integer(this.zoom * (cy - this.vy1)) + this.y1;

        -- translate the content coordinates of the parent to screen coordinates
        if this.parent /= null then
            parentX := wx;
            parentY := wy;
            this.parent.Translate_To_Window( parentX, parentY, wx, wy );
        end if;
    end Translate_To_Window;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Widget ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    function Lt( l, r : A_Widget ) return Boolean is
    begin
        if l = null then
            return r /= null;
        elsif r /= null then
            return l.id < r.id;
        end if;

        -- l /= null and r = null
        return False;
    end Lt;

    --==========================================================================

    not overriding
    procedure Apply( this : access Layout; widget : not null A_Widget ) is
        pragma Unreferenced( this, widget );
    begin
        raise Constraint_Error with "Invoked abstract method Layout.Apply";
    end Apply;
    pragma No_Return( Apply );

    --==========================================================================

    function Create_Listener( key       : Integer;
                              modifiers : Modifiers_Pattern;
                              handler   : not null A_Key_Handler
                            ) return A_Key_Listener is
        this : constant A_Simple_Key_Listener := new Simple_Key_Listener;
    begin
        this.Construct( key, modifiers, handler );
        return A_Key_Listener(this);
    end Create_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this      : access Simple_Key_Listener;
                         key       : Integer;
                         modifiers : Modifiers_Pattern;
                         handler   : not null A_Key_Handler ) is
    begin
        Simple_Action_Listener(this.all).Construct;
        this.key := key;
        this.modifiers := modifiers;
        this.handler := handler;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this    : access Simple_Key_Listener;
                             action  : A_Key_Action;
                             handled : out Boolean ) is
    begin
        if action.key = this.key and then action.modifiers = this.modifiers then
            this.handler( action, handled );
        else
            handled := False;
        end if;
    end Handle_Action;

    --==========================================================================

    not overriding
    procedure Construct( this      : access Key_Action;
                         id        : Action_Id;
                         source    : not null A_Widget;
                         key       : Integer;
                         modifiers : Modifiers_Array ) is
    begin
        Action(this.all).Construct( id, source );
        this.key := key;
        this.modifiers := modifiers;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Key( this : access Key_Action ) return Integer is
    begin
        return this.key;
    end Get_Key;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Modifiers( this : access Key_Action ) return Modifiers_Array is
    begin
        return this.modifiers;
    end Get_Modifiers;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Key_Action ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    --==========================================================================

    procedure Delete( this : in out A_Resize_Action ) is
    begin
        Delete( A_Object(this) );
    end Delete;


    --==========================================================================

    not overriding
    procedure Construct( this    : access Visibility_Action;
                         id      : Action_Id;
                         source  : not null A_Widget;
                         visible : Boolean ) is
    begin
        Action(this.all).Construct( id, source );
        this.visible := visible;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Visible( this : access Visibility_Action ) return Boolean is
    begin
        return this.visible;
    end Is_Visible;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Visibility_Action ) is
    begin
        Delete( A_Object(this) );
    end Delete;

end Widgets;
