
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Text_IO;
with Applications;
with Content_Signing;                   use Content_Signing;
with GNAT.Strings;                      use GNAT.Strings;
with Preferences;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;

pragma Elaborate_All( Applications );
pragma Elaborate_All( Preferences );
pragma Elaborate_All( Support.Paths );

package body Applications.Shell.Fingerprint is

    function Create_Application return A_Application is
        this : constant A_Application := new Fingerprint_Application;
    begin
        this.Construct;
        return this;
    end Create_Application;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Fingerprint_Application ) is
    begin
        Shell_Application(this.all).Construct( "Kevin Wellwood",
                                               "fingerprint" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Run( this : access Fingerprint_Application; returnCode : in out Integer ) is
        pragma Unreferenced( returnCode );

        quiet : Boolean := False;
        pragma Warnings( Off, quiet );

        ------------------------------------------------------------------------

        procedure Output_Line( str : String ) is
        begin
            if not quiet then
                Ada.Text_IO.Put_Line( str );
            end if;
        end Output_Line;

        ------------------------------------------------------------------------

        procedure Display_Title is
        begin
            Output_Line( "" );
            Output_Line( this.Get_Announce );
            Output_Line( "" );
        end Display_Title;

        ------------------------------------------------------------------------

        procedure Display_Usage is
        begin
            Display_Title;
            Output_Line( "Usage:" );
            Output_Line( "  " & this.Get_Name & " [-q] <output_file> <input_file>*" );
            Output_Line( "" );
            Output_Line( "Where:" );
            Output_Line( "  <output_file> : Name of the digest file to create" );
            Output_Line( "  <input_file>  : Path of a file to add to the digest" );
            Output_Line( "" );
            Output_Line( "Switches:" );
            Output_Line( "  -q : Quiet mode" );
            Output_Line( "" );
            returnCode := 1;
        end Display_Usage;

        ------------------------------------------------------------------------

        procedure Display_Progress( filename : String; completion : Natural ) is
        begin
            Output_Line( Image( completion ) & "% " & filename );
        end Display_Progress;

        ------------------------------------------------------------------------

        carg        : Positive := 1;
        outFilename : Unbounded_String;
    begin
        if Argument_Count < 2 then
            Display_Usage;
            return;
        end if;

        if Argument( carg ) = "-q" then
            quiet := True;
            carg := carg + 1;

            if carg = Argument_Count then
                -- missing input file list
                Display_Usage;
                Output_Line( "Error: Expected input file list" );
                returnCode := 1;
                return;
            end if;
        end if;

        outFilename := To_Unbounded_String( Argument( carg ) );
        carg := carg + 1;

        Display_Title;

        declare
            filenames : String_List(carg..Argument_Count);
        begin
            for i in filenames'Range loop
                filenames(i) := new String'(Argument( i ));
            end loop;

            Output_Line( "Writing " & To_String( outFilename ) & "..." );
            Output_Line( "" );

            Content_Signing.Make_Dictionary( To_String( outFilename ),
                                             filenames,
                                             this.Get_Company,
                                             Display_Progress'Unrestricted_Access );

            for i in filenames'Range loop
                Free( filenames(i) );
            end loop;
        exception
            when e : others =>
                for i in filenames'Range loop
                    Free( filenames(i) );
                end loop;
                Output_Line( "Error: " & Exception_Message( e ) );
                returnCode := 3;
        end;
    end Run;

    ----------------------------------------------------------------------------

begin

    Register_Allocator( Create_Application'Access );

    Preferences.Set_Default( "application", "media", "media" & Slash );

end Applications.Shell.Fingerprint;
