
with Ada.Strings;                       use Ada.Strings;
with Applications;                      use Applications;
--with Debugging;                         use Debugging;
with Events.Audio;                      use Events.Audio;
with Game_Views;                        use Game_Views;
with Interfaces;                        use Interfaces;
with Resources;                         use Resources;
with Support;                           use Support;

package body Widgets.Progress_Boards is

    DIGIT_WIDTH : constant := 8;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Listener( this     : access Progress_Board;
                            listener : not null A_PBoard_Listener ) is
    begin
        this.Add_Listener( "PBoard_Listener", A_Action_Listener(listener) );
    end Add_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Listener( this    : access Progress_Board;
                            handler : not null A_PBoard_Handler ) is
        listener : constant A_PBoard_Listener := Create_Listener( handler );
    begin
        this.Add_Listener( listener );
    end Add_Listener;

    ----------------------------------------------------------------------------

    function Create_Progress_Board( view    : not null access Game_Views.Game_View'Class;
                                    id      : String;
                                    libName : String ) return A_Progress_Board is
        this : A_Progress_Board := new Progress_Board;
    begin
        this.Construct( view, id, libName );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Progress_Board;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Progress_Board;
                         view    : not null access Game_Views.Game_View'Class;
                         id      : String;
                         libName : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.Set_Font( "dos", 10 );
        this.acceptFocus := False;
        this.transparent := True;
        this.visible := False;
        this.lib := Load_Library( libName );
        if this.lib /= null then
            this.backgroundId := this.lib.Get_Id( "dropdown.png" );
            for i in 0..9 loop
                this.digitId(i) := this.lib.Get_Id( "digit" & Image( i ) & ".png" );
            end loop;
            this.ancientId := this.lib.Get_Id( "ancient_icon.png" );
            this.redKeyId := this.lib.Get_Id( "gem_icon-red.png" );
            this.blueKeyId := this.lib.Get_Id( "gem_icon-blue.png" );
            this.yellowKeyId := this.lib.Get_Id( "gem_icon-yellow.png" );
            this.greenKeyId := this.lib.Get_Id( "gem_icon-green.png" );
            if this.backgroundId > 0 then
                -- center the board horizontally and off the top edge of the
                -- parent's content region. when the widget is shown, it will
                -- move itself down into view.
                this.Set_Layout_CenterHY( this.Get_Min_Width, -this.Get_Min_Height, -1 );
            end if;
        else
            raise RESOURCE_ERROR with "Error loading tile library <" & libName & ">";
        end if;
        this.Get_View.Attach( A_Process(this) );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Progress_Board ) is
    begin
        this.Get_View.Detach( A_Process(this) );
        Unload_Library( this.lib );
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Dispatch_Action( this : access Progress_Board; id : Action_Id ) is
        action : A_PBoard_Action := new PBoard_Action;

        procedure Dispatch( listener : A_Action_Listener ) is
        begin
            A_PBoard_Listener(listener).Handle_Action( action );
        end Dispatch;

    begin
        action.Construct( id, A_Widget(this) );
        this.Iterate_Listeners( "PBoard_Listener", Dispatch'Access );
        Delete( action );
    end Dispatch_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Progress_Board; dc : Drawing_Context ) is

        ------------------------------------------------------------------------

        -- draw a number on the board, given the location of the upper left
        -- corner of the most significant digit and the maximum number of digits
        -- to draw. if the number needs more than the maximum number of digits,
        -- the largest number possible will be drawn using the maximum number of
        -- digits. (ex: digits = 2, n = 100, draws "99")
        procedure Draw_Number( number : Natural;
                               x, y   : Integer;
                               dgts   : Positive ) is
            number2 : Unsigned_64 := Unsigned_64(number);
            n       : Unsigned_64;
            exp     : Unsigned_64;
            digit   : Natural;
        begin
            if number2 >= Unsigned_64(dgts) ** 10 then
                number2 := Unsigned_64(dgts) ** 10 - 1;
            end if;
            n := number2;

            for i in reverse 0..dgts-1 loop
                exp := 10 ** i;
                if exp <= number2 or else i = 0 then
                    digit := Natural((n - (n mod exp)) / exp);
                    Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.digitId( digit ) ),
                                       x + ((dgts - 1) - i) * DIGIT_WIDTH, y );
                    n := n - Unsigned_64(digit) * exp;
                end if;
            end loop;
        end Draw_Number;

        ------------------------------------------------------------------------

        theme    : constant A_Theme := Get_Application.Get_Theme;
        bmp      : A_Bitmap;
        locSplit : Integer;
    begin
        if this.lib = null then
            return;
        end if;

        -- draw the background
        bmp := this.lib.Get_Bitmap( this.backgroundId );
        if this.bmp /= null then
            Draw_Alpha_Sprite( dc, bmp, 0, 0 );
        end if;

        -- draw the location text
        locSplit := Length( this.location ) + 1;
        while theme.Get_Text_Width( Slice( this.location, 1, locSplit - 1 ), To_String( this.fontName ), this.fontSize ) > 150 loop
            locSplit := Index( this.location, " ", locSplit - 1, Backward );
             if locSplit < 1 then
                 locSplit := Length( this.location ) + 1;
                 exit;
             end if;
        end loop;
        if locSplit > Length( this.location ) then
            theme.Draw_Label( dc,
                              40, 40, 191, 57,
                              null, To_String( this.location ),
                              To_String( this.fontName ), this.fontSize,
                              this.colors(Text), Align_Center,
                              smooth => this.fontSmooth and then not this.transparent );
        else
            theme.Draw_Label( dc,
                              40, 40, 191, 48,
                              null, Slice( this.location, 1, locSplit - 1),
                              To_String( this.fontName ), this.fontSize,
                              this.colors(Text), Align_Center,
                              smooth => this.fontSmooth and then not this.transparent );
            theme.Draw_Label( dc,
                              40, 49, 191, 57,
                              null, Slice( this.location, locSplit + 1, Length( this.location ) ),
                              To_String( this.fontName ), this.fontSize,
                              this.colors(Text), Align_Center,
                              smooth => this.fontSmooth and then not this.transparent );
        end if;

        Draw_Number( this.points, 37, 72, dgts => 8 );
        Draw_Number( this.nextLife, 133, 72, dgts => 8 );
        Draw_Number( this.ammo, 173, 112, dgts => 3 );
        Draw_Number( this.lives, 85, 128, dgts => 2 );
        Draw_Number( this.drops, 181, 128, dgts => 2 );

        for i in 0..this.ancients-1 loop
            Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.ancientId ), 36 + i*8, 96 );
        end loop;

        if this.redKey then
            Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.redKeyId ), 76, 114 );
        end if;
        if this.blueKey then
            Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.blueKeyId ), 84, 114 );
        end if;
        if this.yellowKey then
            Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.yellowKeyId ), 92, 114 );
        end if;
        if this.greenKey then
            Draw_Alpha_Sprite( dc, this.lib.Get_Bitmap( this.greenKeyId ), 100, 114 );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Progress_Board ) return Natural is
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        elsif this.lib /= null then
            return Get_Height( this.lib.Get_Bitmap_Sync( this.backgroundId ) );
        else
            return 0;
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Progress_Board ) return Natural is
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        elsif this.lib /= null then
            return Get_Width( this.lib.Get_Bitmap_Sync( this.backgroundId ) );
        else
            return 0;
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Progress_Board ) return String is
        pragma Unreferenced( this );
    begin
        return "Progress Board";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Open( this : access Progress_Board ) return Boolean is
    begin
        return this.slidePos > -1.0 and then this.Is_Visible;
    end Is_Open;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Listener( this     : access Progress_Board;
                               listener : not null A_PBoard_Listener ) is
    begin
        this.Remove_Listener( "PBoard_Listener", A_Action_Listener(listener) );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Ammo( this : access Progress_Board; ammo : Natural ) is
    begin
        if this.ammo /= ammo then
            this.ammo := ammo;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Ammo;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Ancients( this : access Progress_Board; ancients : Natural ) is
    begin
        if this.ancients /= ancients then
            this.ancients := ancients;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Ancients;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Blue_Key( this : access Progress_Board; blueKey : Boolean ) is
    begin
        if this.blueKey /= blueKey then
            this.blueKey := blueKey;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Blue_Key;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Drops( this : access Progress_Board; drops : Natural ) is
    begin
        if this.drops /= drops then
            this.drops := drops;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Drops;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Green_Key( this : access Progress_Board; greenKey : Boolean ) is
    begin
        if this.greenKey /= greenKey then
            this.greenKey := greenKey;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Green_Key;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Lives( this : access Progress_Board; lives : Natural ) is
    begin
        if this.lives /= lives then
            this.lives := lives;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Lives;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Location( this : access Progress_Board; location : String ) is
    begin
        if To_String( this.location ) /= location then
            this.location := To_Unbounded_String( location );
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Location;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Next_Life( this : access Progress_Board; nextLife : Natural ) is
    begin
        if this.nextLife /= nextLife then
            this.nextLife := nextLife;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Next_Life;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Points( this : access Progress_Board; points : Natural ) is
    begin
        if this.points /= points then
            this.points := points;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Points;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Red_Key( this : access Progress_Board; redKey : Boolean ) is
    begin
        if this.redKey /= redKey then
            this.redKey := redKey;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Red_Key;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Scuba( this : access Progress_Board; scuba : Boolean ) is
    begin
        if this.scuba /= scuba then
            this.scuba := scuba;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Scuba;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Yellow_Key( this : access Progress_Board; yellowKey : Boolean ) is
    begin
        if this.yellowKey /= yellowKey then
            this.yellowKey := yellowKey;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Yellow_Key;

    ----------------------------------------------------------------------------

    not overriding
    procedure Show( this : access Progress_Board; enabled : Boolean ) is
    begin
        if enabled then
            -- slide down
            if not this.Is_Open then
                this.Set_Visible( True );
                this.Dispatch_Action( Opened );
            end if;
            this.slideDir := 1;
            Queue_Play_Sound( "menu_open" );
        else
            -- slide up
            if this.Is_Open then
                this.slideDir := -1;
            end if;
            Queue_Play_Sound( "menu_close" );
        end if;
    end Show;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Progress_Board; upTime, dt : Time_Span ) is
        dtf  : constant Float := Float(To_Duration( dt ));
        minH : constant Float := Float(this.Get_Min_Height);
        pos  : Integer;
        pragma Unreferenced( upTime );
    begin
        if this.slideDir < 0 then
            -- slide the widget up if we can
            if this.backgroundId > 0 and then this.slidePos > -1.0 then
                this.slidePos := Float'Max( -1.0, this.slidePos - dtf * minH / this.slideDelay );
                pos := Integer(this.slidePos);
                if this.y2 /= pos then
                    this.Set_Layout_CenterHY( this.Get_Min_Width, pos - this.Get_Min_Height + 1, pos );
                end if;
            end if;
            -- check if the slide is complete
            if this.slidePos <= -1.0 then
                this.slideDir := 0;
                this.Set_Visible( False );
                this.Dispatch_Action( Closed );
            end if;

        elsif this.slideDir > 0 then
            -- slide the widget down if we can
            if this.backgroundId > 0 and then this.slidePos < minH then
                this.slidePos := Float'Min( minH - 1.0, this.slidePos + dtf * minH / this.slideDelay );
                pos := Integer(this.slidePos);
                if this.y2 /= pos then
                    this.Set_Layout_CenterHY( this.Get_Min_Width, pos - this.Get_Min_Height + 1, pos );
                end if;
            end if;
            -- check if the slide is complete
            if this.slidePos >= minH - 1.0 then
                this.slideDir := 0;
            end if;

        end if;
    end Tick;

    ----------------------------------------------------------------------------

    not overriding
    procedure Toggle( this : access Progress_Board ) is
    begin
        this.Show( not this.Is_Open );
    end Toggle;

    --==========================================================================

    function Create_Listener( handler : not null A_PBoard_Handler ) return A_PBoard_Listener is
        this : constant A_Simple_PBoard_Listener := new Simple_PBoard_Listener;
    begin
        this.Construct( handler );
        return A_PBoard_Listener(this);
    end Create_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Simple_PBoard_Listener;
                         handler : not null A_PBoard_Handler ) is
    begin
        Simple_Action_Listener(this.all).Construct;
        this.handler := handler;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Action( this   : access Simple_PBoard_Listener;
                             action : A_PBoard_Action ) is
    begin
        this.handler( action );
    end Handle_Action;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_PBoard_Action ) is
    begin
        Delete( A_Object(this) );
    end Delete;

end Widgets.Progress_Boards;
