
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Text_IO;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Applications;
with GNAT.Directory_Operations.Iteration;
 use GNAT.Directory_Operations.Iteration;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Resources.Images;                  use Resources.Images;
with Support.Paths;                     use Support.Paths;
with Tile_Operations;                   use Tile_Operations;

pragma Elaborate_All( Applications );

package body Applications.Shell.TileTool is

    function Create_Application return A_Application is
        this : constant A_Application := new TileTool_Application;
    begin
        this.Construct;
        return this;
    end Create_Application;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access TileTool_Application ) is
    begin
        Shell_Application(this.all).Construct( company => "Kevin Wellwood",
                                               name    => "tiletool" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Run( this       : access TileTool_Application;
                   returnCode : in out Integer ) is

        ------------------------------------------------------------------------

        procedure Display_Version is
        begin
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put_Line( this.Get_Announce );
            Ada.Text_IO.New_Line;
        end Display_Version;

        ------------------------------------------------------------------------

        -- 'command' should be lower case.
        procedure Display_Help( command : String ) is
        begin
            if command = "convert" then
                Ada.Text_IO.Put_Line( "convert: Converts BMP files to 24-bit" );
                Ada.Text_IO.New_Line;
                Ada.Text_IO.Put_Line( "Usage: convert <bmp_dir>" );
                Ada.Text_IO.New_Line;
                Ada.Text_IO.Put_Line( "Where:" );
                Ada.Text_IO.Put_Line( "  bmp_dir : The directory containing the BMP files. All the BMP files will be" );
                Ada.Text_IO.Put_Line( "            converted, non-recursively." );
            elsif command = "cut" then
                Ada.Text_IO.Put_Line( "cut: Cuts and saves tiles from a sheet" );
                Ada.Text_IO.New_Line;
                Ada.Text_IO.Put_Line( "Usage: cut [switches] <image_file>" );
                Ada.Text_IO.New_Line;
                Ada.Text_IO.Put_Line( "Switches:" );
                Ada.Text_IO.Put_Line( "  -width=?   : Tile width in pixels (required)" );
                Ada.Text_IO.Put_Line( "  -spacing=? : Number of pixels between tiles" );
                Ada.Text_IO.Put_Line( "  -skip      : Skip completely transparent tiles" );
                Ada.Text_IO.Put_Line( "  -prefix=?  : Path prefix for output images" );
                Ada.Text_IO.Put_Line( "  -format=xx : Output images format, xx=bmp|png" );
                Ada.Text_IO.Put_Line( "  -index=?   : Path of output tile index" );
                Ada.Text_IO.Put_Line( "  -id=?      : First id for generated tile index" );
            else
                Ada.Text_IO.Put_Line( "help: Unknown command: " & command );
            end if;
        end Display_Help;

        ------------------------------------------------------------------------

        procedure Display_Usage is
        begin
            Ada.Text_IO.Put_Line( "Performs operations on tile and tilesheet images" );
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put_Line( "Usage: " & this.Get_Name & " <command> [...]" );
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put_Line( "Commands:" );
            Ada.Text_IO.Put_Line( "  convert : Converts BMP files to 24-bit" );
            Ada.Text_IO.Put_Line( "  cut     : Cuts and saves tiles from a sheet" );
            Ada.Text_IO.Put_Line( "  help    : Shows help about a specific command" );
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put_Line( "Type '" & this.Get_Name & " help <command>' for help with a specific command." );
            returnCode := 1;
        end Display_Usage;

        ------------------------------------------------------------------------

        procedure Convert_BMPs( dir : String ) is

            procedure Convert( item : String; index : Positive; quit : in out Boolean ) is
                pragma Unreferenced( index, quit );
                filename : constant String := Normalize_Pathname( item );
            begin
                Ada.Text_IO.Put( "  " & filename );
                if Convert_BMP_To_24( filename ) then
                    Ada.Text_IO.Put_Line( " : Ok" );
                else
                    Ada.Text_IO.Put_Line( " : Failed" );
                end if;
            end Convert;

            procedure Convert_All is new Wildcard_Iterator( Convert );
        begin
            Ada.Text_IO.Put_Line( "Converting " & dir & Slash & "*.bmp ..." );
            Convert_All( dir & Slash & "*.[bB][mM][pP]" );
        end Convert_BMPs;

        ------------------------------------------------------------------------

        function Arg_Value( arg : String ) return String is
            eq : constant Integer := Index( arg, "=" );
        begin
            if eq >= arg'First and then eq < arg'Last then
                return arg(eq+1..arg'Last);
            else
                return "";
            end if;
        end Arg_Value;

        ------------------------------------------------------------------------

        function Match_Arg( arg : String; head : String ) return Boolean is
            larg : constant String := To_Lower( arg );
        begin
            return larg'Length >= head'Length and then
                   larg(larg'First..larg'First+head'Length-1) = To_Lower( head );
        end Match_Arg;

        ------------------------------------------------------------------------

        function Is_Integer( str : String ) return Boolean is
        begin
             if str'Length = 0 then
                 return False;
             end if;
             for i in str'Range loop
                 if not Is_Digit( str(i) ) then
                     return False;
                 end if;
             end loop;
             return True;
        end Is_Integer;

        ------------------------------------------------------------------------

    begin
        if Argument_Count = 0 then
            Ada.Text_IO.Put_Line( "Type '" & this.Get_Window_Title & " help' for usage." );
            returnCode := 1;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        elsif To_Lower( Argument( 1 ) ) = "help" then
            if Argument_Count > 1 then
                Display_Help( To_Lower( Argument( 2 ) ) );
            else
                Display_Version;
                Display_Usage;
            end if;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        elsif To_Lower( Argument( 1 ) ) = "convert" then
            if Argument_Count < 2 then
                -- Error on missing image file
                Ada.Text_IO.Put_Line( "convert: Missing argument 'bmp_dir'" );
                returnCode := 1;
            else
                -- Match arguments
                for i in 2..Argument_Count-1 loop
                    Ada.Text_IO.Put_Line( "convert: Unknown switch: " & Argument( i ) );
                    returnCode := 1;

                    exit when returnCode /= 0;
                end loop;

                if not Is_Directory( Normalize_Pathname( Argument( Argument_Count ) ) ) then
                    Ada.Text_IO.Put_Line( "convert: Directory not found: " & Argument( Argument_Count ) );
                    returnCode := 1;
                end if;

                if returnCode = 0 then
                    Convert_BMPs( Normalize_Pathname( Argument( Argument_Count ) ) );
                end if;
            end if;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        elsif To_Lower( Argument( 1 ) ) = "cut" then
            declare
                width   : Integer := 0;
                spacing : Integer := 0;
                skip    : Boolean := False;
                prefix  : Unbounded_String;
                format  : Unbounded_String;
                idxPath : Unbounded_String;
                startId : Positive := 1;
            begin
                if Argument_Count < 2 then
                    -- Error on missing image file
                    Ada.Text_IO.Put_Line( "cut: Missing argument 'image_file'" );
                    returnCode := 1;
                else
                    -- Match arguments
                    for i in 2..Argument_Count-1 loop
                        if Match_Arg( Argument( i ), "-width=" ) then
                            if Is_Integer( Arg_Value( Argument( i ) ) ) then
                                width := Integer'Max( 1, Integer'Value( Arg_Value( Argument( i ) ) ) );
                            else
                                Ada.Text_IO.Put_Line( "cut: Expected integer for '-width'" );
                                returnCode := 1;
                            end if;

                        elsif Match_Arg( Argument( i ), "-spacing=" ) then
                            if Is_Integer( Arg_Value( Argument( i ) ) ) then
                                spacing := Integer'Max( 0, Integer'Value( Arg_Value( Argument( i ) ) ) );
                            else
                                Ada.Text_IO.Put_Line( "cut: Expected integer for '-spacing'" );
                                returnCode := 1;
                            end if;

                        elsif Match_Arg( Argument( i ), "-skip" ) then
                            skip := True;

                        elsif Match_Arg( Argument( i ), "-prefix=" ) then
                            prefix := To_Unbounded_String( Arg_Value( Argument( i ) ) );

                        elsif Match_Arg( Argument( i ), "-format=" ) then
                            format := To_Unbounded_String( To_Lower( Arg_Value( Argument( i ) ) ) );
                            if Length( format ) > 0 and then not Is_Format_Supported( To_String( format ) ) then
                                Ada.Text_IO.Put_Line( "cut: Unsupported output format: " &
                                                      Arg_Value( Argument( i ) ) );
                                returnCode := 1;
                            end if;

                        elsif Match_Arg( Argument( i ), "-index=" ) then
                            idxPath := To_Unbounded_String( Arg_Value( Argument( i ) ) );

                        elsif Match_Arg( Argument( i ), "-id=" ) then
                            if Is_Integer( Arg_Value( Argument( i ) ) ) then
                                startId := Integer'Max( 1, Integer'Value( Arg_Value( Argument( i ) ) ) );
                            else
                                Ada.Text_IO.Put_Line( "cut: Expected integer > 0 for '-id'" );
                                returnCode := 1;
                            end if;

                        else
                            Ada.Text_IO.Put_Line( "cut: Unknown switch: " & Argument( i ) );
                            returnCode := 1;

                        end if;

                        exit when returnCode /= 0;
                    end loop;

                    -- Error on missing arguments
                    if width = 0 then
                        Ada.Text_IO.Put_Line( "cut: Missing argument '-width'" );
                        returnCode := 1;
                    end if;
                end if;

                -- Execute command if everything is ok
                if returnCode = 0 then
                    Tile_Operations.Cut( filename  => Argument( Argument_Count ),
                                         width     => width,
                                         spacing   => spacing,
                                         skipEmpty => skip,
                                         prefix    => To_String( prefix ),
                                         format    => To_String( format ),
                                         indexPath => To_String( idxPath ),
                                         startId   => startId,
                                         textOut   => Ada.Text_IO.Put_Line'Access );
                end if;
            end;

        -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        else
            Ada.Text_IO.Put_Line( "Unknown command: " & To_Lower( Argument( 1 ) ) );
            Ada.Text_IO.Put_Line( "Type '" & this.Get_Window_Title & " help' for help." );
            returnCode := 1;
        end if;

        Ada.Text_IO.New_Line;
    end Run;

    ----------------------------------------------------------------------------

begin

    Register_Allocator( Create_Application'Access );

end Applications.Shell.TileTool;
